/*
 * $RCSfile: TIFFJPEGDecompressor.java,v $
 *
 *
 * Copyright (c) 2005 Sun Microsystems, Inc. All  Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistribution of source code must retain the above copyright
 *   notice, this  list of conditions and the following disclaimer.
 *
 * - Redistribution in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * This software is provided "AS IS," without a warranty of any
 * kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
 * EXCLUDED. SUN MIDROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL
 * NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF
 * USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR
 * ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
 * CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND
 * REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * You acknowledge that this software is not designed or intended for
 * use in the design, construction, operation or maintenance of any
 * nuclear facility.
 *
 * $Revision: 1.2 $
 * $Date: 2006/04/11 22:10:36 $
 * $State: Exp $
 */
package com.sun.media.imageioimpl.plugins.tiff;

import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.util.Iterator;
import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.ImageReadParam;
import javax.imageio.stream.MemoryCacheImageInputStream;
import javax.imageio.stream.ImageInputStream;
import com.sun.media.imageio.plugins.tiff.BaselineTIFFTagSet;
import com.sun.media.imageio.plugins.tiff.TIFFDecompressor;
import com.sun.media.imageio.plugins.tiff.TIFFField;

public class TIFFJPEGDecompressor extends TIFFDecompressor {
   //CHECKSTYLE:OFF

   // Start of Image
   protected static final int SOI = 0xD8;

   // End of Image
   protected static final int EOI = 0xD9;

   protected ImageReader jPEGReader;
   protected ImageReadParam jPEGParam;

   protected boolean hasJPEGTables;
   protected byte[] tables;

   private byte[] _data = new byte[0];

   @Override
   public final void beginDecoding() {
      // Initialize the JPEG reader if needed.
      if (this.jPEGReader == null) {
         // Get all JPEG readers.
         Iterator<ImageReader> iter = ImageIO.getImageReadersByFormatName("jpeg");

         if (!iter.hasNext()) {
            throw new IllegalStateException("No JPEG readers found!");
         }

         // Initialize reader to the first one.
         this.jPEGReader = iter.next();

         this.jPEGParam = jPEGReader.getDefaultReadParam();
      }

      // Get the JPEGTables field.
      TIFFImageMetadata tmetadata = (TIFFImageMetadata) metadata;
      TIFFField field
          = tmetadata.getTIFFField(BaselineTIFFTagSet.TAG_JPEG_TABLES);

      if (field != null) {
         this.hasJPEGTables = true;
         this.tables = field.getAsBytes();
      } else {
         this.hasJPEGTables = false;
      }
   }

   @Override
   public void decodeRaw( byte[] bytes,
       int dstOffset,
       int bitsPerPixel,
       int scanlineStride ) throws IOException {
      // Seek to the data position for this segment.
      stream.seek(offset);

      // Set the stream variable depending on presence of JPEGTables.
      ImageInputStream is;
      if (this.hasJPEGTables) {
         // The current strip or tile is an abbreviated JPEG stream.

         // Reallocate memory if there is not enough already.
         int dataLength = tables.length + byteCount;
         if (_data.length < dataLength) {
            _data = new byte[dataLength];
         }

         // Copy the tables ignoring any EOI and subsequent bytes.
         int dataOffset = tables.length;
         for (int i = tables.length - 2; i > 0; i--) {
            if ((tables[i] & 0xff) == 0xff
                && (tables[i + 1] & 0xff) == EOI) {
               dataOffset = i;
               break;
            }
         }
         System.arraycopy(tables, 0, _data, 0, dataOffset);

         // Check for SOI and skip it if present.
         byte byte1 = (byte) stream.read();
         byte byte2 = (byte) stream.read();
         if (!((byte1 & 0xff) == 0xff && (byte2 & 0xff) == SOI)) {
            _data[dataOffset++] = byte1;
            _data[dataOffset++] = byte2;
         }

         // Read remaining data.
         stream.readFully(_data, dataOffset, byteCount - 2);

         // Create ImageInputStream.
         ByteArrayInputStream bais = new ByteArrayInputStream(_data);
         is = new MemoryCacheImageInputStream(bais);
      } else {
         // The current strip or tile is a complete JPEG stream.
         is = stream;
      }

      // Set the stream on the reader.
      jPEGReader.setInput(is, false, true);

      // Set the destination to the raw image ignoring the parameters.
      jPEGParam.setDestination(rawImage);

      // Read the strip or tile.
      jPEGReader.read(0, jPEGParam);
   }

   @Override
   protected void finalize() throws Throwable {
      super.finalize();
      jPEGReader.dispose();
   }
   //CHECKSTYLE:ON
}
