/*
 * $RCSfile: TIFFOldJPEGDecompressor.java,v $
 *
 * 
 * Copyright (c) 2005 Sun Microsystems, Inc. All  Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met: 
 * 
 * - Redistribution of source code must retain the above copyright 
 *   notice, this  list of conditions and the following disclaimer.
 * 
 * - Redistribution in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in 
 *   the documentation and/or other materials provided with the
 *   distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of 
 * contributors may be used to endorse or promote products derived 
 * from this software without specific prior written permission.
 * 
 * This software is provided "AS IS," without a warranty of any 
 * kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND 
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
 * EXCLUDED. SUN MIDROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL 
 * NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF 
 * USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR 
 * ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
 * CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND
 * REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES. 
 * 
 * You acknowledge that this software is not designed or intended for 
 * use in the design, construction, operation or maintenance of any 
 * nuclear facility. 
 *
 * $Revision: 1.3 $
 * $Date: 2006/04/11 22:10:37 $
 * $State: Exp $
 */
package com.sun.media.imageioimpl.plugins.tiff;

import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import javax.imageio.IIOException;
import javax.imageio.stream.MemoryCacheImageInputStream;
import javax.imageio.stream.ImageInputStream;
import com.sun.media.imageio.plugins.tiff.BaselineTIFFTagSet;
import com.sun.media.imageio.plugins.tiff.TIFFField;

/**
 * <code>TIFFDecompressor</code> for "Old JPEG" compression.
 */
public final class TIFFOldJPEGDecompressor extends TIFFJPEGDecompressor {
	//CHECKSTYLE:OFF
	// Start of Image
	// private static final int SOI = 0xD8; // now defined in superclass

	// Define Huffman Tables
	private static final int DHT = 0xC4;

	// Define Quantisation Tables
	private static final int DQT = 0xDB;

	// Define Restart Interval
	private static final int DRI = 0xDD;

	// Baseline DCT
	private static final int SOF0 = 0xC0;

	// Start of Scan
	private static final int SOS = 0xDA;

	// End of Image
	// private static final int EOI = 0xD9; // now defined in superclass

	// Whether the decompressor has been initialized.
	private boolean _isInitialized;

	//
	// Instance variables set by the initialize() method.
	//
	// Offset to a complete, contiguous JPEG stream.
	private Long _JPEGStreamOffset;
	// Offset to the SOF marker.
	private int _SOFPosition = -1;
	// Value of the SOS marker.
	private byte[] _SOSMarker;

	// Horizontal chroma subsampling factor.
	private int _subsamplingX = 2;

	// Vertical chroma subsampling factor.
	private int _subsamplingY = 2;

	private synchronized void initialize() throws IOException {
		if (_isInitialized) {
			return;
		}

		// Get the TIFF metadata object.
		TIFFImageMetadata tim = (TIFFImageMetadata)metadata;

		// Get the JPEGInterchangeFormat field.
		TIFFField jPEGInterchangeFormatField = tim.getTIFFField(BaselineTIFFTagSet.TAG_JPEG_INTERCHANGE_FORMAT);

		// Get the tile or strip offsets.
		TIFFField segmentOffsetField =
				tim.getTIFFField(BaselineTIFFTagSet.TAG_TILE_OFFSETS);
		if (segmentOffsetField == null) {
			segmentOffsetField =
					tim.getTIFFField(BaselineTIFFTagSet.TAG_STRIP_OFFSETS);
			if (segmentOffsetField == null) {
				segmentOffsetField = jPEGInterchangeFormatField;
			}
		}
		long[] segmentOffsets = segmentOffsetField.getAsLongs();

		// Determine whether the image has more than one strip or tile.
		boolean isTiled = segmentOffsets.length > 1;

		if (!isTiled) {
			//
			// If the image has only a single strip or tile and it looks
			// as if a complete JPEG stream is present then set the value
			// of JPEGStreamOffset to the offset of the JPEG stream;
			// otherwise leave JPEGStreamOffset set to null.
			//

			stream.seek(offset);
			stream.mark();
			if (stream.read() == 0xff && stream.read() == SOI) {
				// Tile or strip offset points to SOI.
				_JPEGStreamOffset = offset;
				// Set initialization flag and return.
				((TIFFImageReader)reader).forwardWarningMessage("SOI marker detected at start of strip or tile.");
				_isInitialized = true;
				return;
			}
			stream.reset();

			if (jPEGInterchangeFormatField != null) {
				// Get the value of JPEGInterchangeFormat.
				long jpegInterchangeOffset = jPEGInterchangeFormatField.getAsLong(0);

				// Get the JPEGInterchangeFormatLength field.
				TIFFField jPEGInterchangeFormatLengthField = tim.getTIFFField(BaselineTIFFTagSet.TAG_JPEG_INTERCHANGE_FORMAT_LENGTH);

				if (jPEGInterchangeFormatLengthField == null) {
					// JPEGInterchangeFormat stream is of indeterminate
					// length so use it as a complete JPEG stream.
					_JPEGStreamOffset = jpegInterchangeOffset;

					// Set initialization flag and return.
					((TIFFImageReader)reader).forwardWarningMessage("JPEGInterchangeFormatLength field is missing");
					_isInitialized = true;
					return;
				} else {
					// Get the JPEGInterchangeFormatLength field's value.
					long jpegInterchangeLength =
							jPEGInterchangeFormatLengthField.getAsLong(0);

					if (jpegInterchangeOffset < segmentOffsets[0]
							&& (jpegInterchangeOffset + jpegInterchangeLength) > segmentOffsets[0]) {
						// JPEGInterchangeFormat points to a position
						// below the segment start position and ends at
						// a position after the segment start position.
						_JPEGStreamOffset = jpegInterchangeOffset;

						// Set initialization flag and return.
						_isInitialized = true;
						return;
					}
				}
			}
		}

		// Get the subsampling factors.
		TIFFField yCbCrSubsamplingField = tim.getTIFFField(BaselineTIFFTagSet.TAG_Y_CB_CR_SUBSAMPLING);
		if (yCbCrSubsamplingField != null) {
			_subsamplingX = yCbCrSubsamplingField.getAsChars()[0];
			_subsamplingY = yCbCrSubsamplingField.getAsChars()[1];
		}

		//
		// Initialize the 'tables' instance variable either for later
		// use in prepending to individual abbreviated strips or tiles.
		//
		if (jPEGInterchangeFormatField != null) {
			// Get the value of JPEGInterchangeFormat.
			long jpegInterchangeOffset =
					jPEGInterchangeFormatField.getAsLong(0);

			// Get the JPEGInterchangeFormatLength field.
			TIFFField jPEGInterchangeFormatLengthField = tim.getTIFFField(BaselineTIFFTagSet.TAG_JPEG_INTERCHANGE_FORMAT_LENGTH);

			if (jPEGInterchangeFormatLengthField != null) {
				// Get the JPEGInterchangeFormatLength field's value.
				long jpegInterchangeLength = jPEGInterchangeFormatLengthField.getAsLong(0);

				if (jpegInterchangeLength >= 2
						&& jpegInterchangeOffset + jpegInterchangeLength <= segmentOffsets[0]) {
					// Determine the length excluding any terminal EOI marker
					// and allocate table memory.
					stream.mark();
					stream.seek(jpegInterchangeOffset + jpegInterchangeLength - 2);
					if (stream.read() == 0xff && stream.read() == EOI) {
						this.tables = new byte[(int)(jpegInterchangeLength - 2)];
					} else {
						this.tables = new byte[(int)jpegInterchangeLength];
					}
					stream.reset();

					// Read the tables.
					stream.mark();
					stream.seek(jpegInterchangeOffset);
					stream.readFully(tables);
					stream.reset();
					((TIFFImageReader)reader).forwardWarningMessage("Incorrect JPEG interchange format: using JPEGInterchangeFormat offset to derive tables.");
				} else {
					((TIFFImageReader)reader).forwardWarningMessage("JPEGInterchangeFormat+JPEGInterchangeFormatLength > offset to first strip or tile.");
				}
			}
		}

		if (this.tables == null) {
			//
			// Create tables-only stream in tables[] consisting of
			// SOI+DQTs+DHTs
			//

			ByteArrayOutputStream baos =
					new ByteArrayOutputStream();


			// SOI
			baos.write(0xff);
			baos.write(SOI);

			// Quantization Tables
			TIFFField field =
					tim.getTIFFField(BaselineTIFFTagSet.TAG_JPEG_Q_TABLES);
			if (field == null) {
				throw new IIOException("JPEGQTables field missing!");
			}

			long[] off = field.getAsLongs();
			// Save stream length;
			long streamLength = stream.length();

			for (int i = 0; i < off.length; i++) {
				baos.write(0xff); // Marker ID
				baos.write(DQT);

				char markerLength = (char)67;
				baos.write((markerLength >>> 8) & 0xff); // Length
				baos.write(markerLength & 0xff);

				baos.write(i); // Table ID and precision

				if (streamLength != -1 && off[i] > streamLength) {
					throw new IIOException("JPEGQTables offset for index "
							+ i + " is not in the stream!");
				}
				byte[] qtable = new byte[64];
				stream.seek(off[i]);
				stream.readFully(qtable);

				baos.write(qtable); // Table data
			}

			// Huffman Tables (k == 0 ? DC : AC).
			for (int k = 0; k < 2; k++) {
				int tableTagNumber = k == 0 ? BaselineTIFFTagSet.TAG_JPEG_DC_TABLES : BaselineTIFFTagSet.TAG_JPEG_AC_TABLES;
				field = tim.getTIFFField(tableTagNumber);
				String fieldName = tableTagNumber == BaselineTIFFTagSet.TAG_JPEG_DC_TABLES ? "JPEGDCTables" : "JPEGACTables";

				if (field == null) {
					throw new IIOException(fieldName + " field missing!");
				}
				off = field.getAsLongs();

				for (int i = 0; i < off.length; i++) {
					baos.write(0xff); // Marker ID
					baos.write(DHT);

					if (streamLength != -1 && off[i] > streamLength) {
						throw new IIOException(fieldName
								+ " offset for index "
								+ i + " is not in the stream!");
					}
					byte[] blengths = new byte[16];
					stream.seek(off[i]);
					stream.readFully(blengths);
					int numCodes = 0;
					for (int j = 0; j < 16; j++) {
						numCodes += blengths[j] & 0xff;
					}

					char markerLength = (char)(19 + numCodes);

					baos.write((markerLength >>> 8) & 0xff); // Length
					baos.write(markerLength & 0xff);

					baos.write(i | (k << 4)); // Table ID and type

					baos.write(blengths); // Number of codes

					byte[] bcodes = new byte[numCodes];
					stream.readFully(bcodes);
					baos.write(bcodes); // Codes
				}
			}

			// SOF0
			baos.write((byte)0xff); // Marker identifier
			baos.write((byte)SOF0);
			short sval = (short)(8 + 3 * samplesPerPixel); // Length
			baos.write((byte)((sval >>> 8) & 0xff));
			baos.write((byte)(sval & 0xff));
			baos.write((byte)8); // Data precision
			sval = (short)srcHeight; // Tile/strip height
			baos.write((byte)((sval >>> 8) & 0xff));
			baos.write((byte)(sval & 0xff));
			sval = (short)srcWidth; // Tile/strip width
			baos.write((byte)((sval >>> 8) & 0xff));
			baos.write((byte)(sval & 0xff));
			baos.write((byte)samplesPerPixel); // Number of components
			if (samplesPerPixel == 1) {
				baos.write((byte)1); // Component ID
				baos.write((byte)0x11); // Subsampling factor
				baos.write((byte)0); // Quantization table ID
			} else { // 3
				for (int i = 0; i < 3; i++) {
					baos.write((byte)(i + 1)); // Component ID
					baos.write((i != 0) ? (byte)0x11 : (byte)(((_subsamplingX & 0x0f) << 4)
							| (_subsamplingY & 0x0f)));

					baos.write((byte)i); // Quantization table ID
				}
			}

			// DRI (optional).
			field = tim.getTIFFField(BaselineTIFFTagSet.TAG_JPEG_RESTART_INTERVAL);
			if (field != null) {
				char restartInterval = field.getAsChars()[0];

				if (restartInterval != 0) {
					baos.write((byte)0xff); // Marker identifier
					baos.write((byte)DRI);

					sval = 4;
					baos.write((byte)((sval >>> 8) & 0xff)); // Length
					baos.write((byte)(sval & 0xff));

					// RestartInterval
					baos.write((byte)((restartInterval >>> 8) & 0xff));
					baos.write((byte)(restartInterval & 0xff));
				}
			}

			tables = baos.toByteArray();
		}

		//
		// Check for presence of SOF marker and save its position.
		//
		int idx = 0;
		int idxMax = tables.length - 1;
		while (idx < idxMax) {
			if ((tables[idx] & 0xff) == 0xff && (tables[idx + 1] & 0xff) == SOF0) {
				_SOFPosition = idx;
				break;
			}
			idx++;
		}

		//
		// If no SOF marker, add one.
		//
		if (_SOFPosition == -1) {
			byte[] tmpTables =
					new byte[tables.length + 10 + 3 * samplesPerPixel];
			System.arraycopy(tables, 0, tmpTables, 0, tables.length);
			int tmpOffset = tables.length;
			_SOFPosition = tables.length;
			tables = tmpTables;

			tables[tmpOffset++] = (byte)0xff; // Marker identifier
			tables[tmpOffset++] = (byte)SOF0;
			short sval = (short)(8 + 3 * samplesPerPixel); // Length
			tables[tmpOffset++] = (byte)((sval >>> 8) & 0xff);
			tables[tmpOffset++] = (byte)(sval & 0xff);
			tables[tmpOffset++] = (byte)8; // Data precision
			sval = (short)srcHeight; // Tile/strip height
			tables[tmpOffset++] = (byte)((sval >>> 8) & 0xff);
			tables[tmpOffset++] = (byte)(sval & 0xff);
			sval = (short)srcWidth; // Tile/strip width
			tables[tmpOffset++] = (byte)((sval >>> 8) & 0xff);
			tables[tmpOffset++] = (byte)(sval & 0xff);
			tables[tmpOffset++] = (byte)samplesPerPixel; // Number of components
			if (samplesPerPixel == 1) {
				tables[tmpOffset++] = (byte)1; // Component ID
				tables[tmpOffset++] = (byte)0x11; // Subsampling factor
				tables[tmpOffset++] = (byte)0; // Quantization table ID
			} else { // 3
				for (int i = 0; i < 3; i++) {
					tables[tmpOffset++] = (byte)(i + 1); // Component ID
					tables[tmpOffset++] = (i != 0) ? (byte)0x11 : (byte)(((_subsamplingX & 0x0f) << 4)
							| (_subsamplingY & 0x0f));

							tables[tmpOffset++] = (byte)i; // Quantization table ID
				}
			}
		}

		//
		// Initialize SOSMarker.
		//
		stream.mark();
		stream.seek(segmentOffsets[0]);
		if (stream.read() == 0xff && stream.read() == SOS) {
			//
			// If the first segment starts with an SOS marker save it.
			//
			int sOSLength = (stream.read() << 8) | stream.read();
			_SOSMarker = new byte[sOSLength + 2];
			_SOSMarker[0] = (byte)0xff;
			_SOSMarker[1] = (byte)SOS;
			_SOSMarker[2] = (byte)((sOSLength & 0xff00) >> 8);
			_SOSMarker[3] = (byte)(sOSLength & 0xff);
			stream.readFully(_SOSMarker, 4, sOSLength - 2);
		} else {
			//
			// Manufacture an SOS marker.
			//
			_SOSMarker = new byte[2 + 6 + 2 * samplesPerPixel];
			int sOSMarkerIndex = 0;
			_SOSMarker[sOSMarkerIndex++] = (byte)0xff; // Marker identifier
			_SOSMarker[sOSMarkerIndex++] = (byte)SOS;
			short sval = (short)(6 + 2 * samplesPerPixel); // Length
			_SOSMarker[sOSMarkerIndex++] = (byte)((sval >>> 8) & 0xff);
			_SOSMarker[sOSMarkerIndex++] = (byte)(sval & 0xff);
			// Number of components in scan
			_SOSMarker[sOSMarkerIndex++] = (byte)samplesPerPixel;
			if (samplesPerPixel == 1) {
				_SOSMarker[sOSMarkerIndex++] = (byte)1; // Component ID
				_SOSMarker[sOSMarkerIndex++] = (byte)0; // Huffman table ID
			} else { // 3
				for (int i = 0; i < 3; i++) {
					_SOSMarker[sOSMarkerIndex++] =
							(byte)(i + 1); // Component ID
					_SOSMarker[sOSMarkerIndex++] =
							(byte)((i << 4) | i); // Huffman table IDs
				}
			}
			_SOSMarker[sOSMarkerIndex++] = (byte)0;
			_SOSMarker[sOSMarkerIndex++] = (byte)0x3f;
			_SOSMarker[sOSMarkerIndex++] = (byte)0;
		}
		stream.reset();

		// Set initialization flag.
		_isInitialized = true;
	}

	//
	// The strategy for reading cases 1-3 is to treat the data as a complete
	// JPEG interchange stream located at JPEGStreamOffset.
	//
	// The strategy for cases 4-5 is to concatenate a tables stream created
	// in initialize() with the entropy coded data in each strip or tile.
	//
	@Override
	public void decodeRaw(byte[] bytes,
			int dstOffset,
			int bitsPerPixel,
			int scanlineStride) throws IOException {

		initialize();

//		TIFFImageMetadata tim = (TIFFImageMetadata)metadata;

		if (_JPEGStreamOffset != null) {
			stream.seek(_JPEGStreamOffset);
			jPEGReader.setInput(stream, false, true);
		} else if (tables != null) {
			// Determine buffer length and allocate.
			int tableLength = tables.length;
			int bufLength = tableLength + _SOSMarker.length + byteCount + 2; // 2 for EOI.
			byte[] buf = new byte[bufLength];
			System.arraycopy(tables, 0, buf, 0, tableLength);
			int bufOffset = tableLength;

			// Update the SOF dimensions.
			short sval = (short)srcHeight; // Tile/strip height
			buf[_SOFPosition + 5] = (byte)((sval >>> 8) & 0xff);
			buf[_SOFPosition + 6] = (byte)(sval & 0xff);
			sval = (short)srcWidth; // Tile/strip width
			buf[_SOFPosition + 7] = (byte)((sval >>> 8) & 0xff);
			buf[_SOFPosition + 8] = (byte)(sval & 0xff);

			// Seek to data.
			stream.seek(offset);

			// Add SOS marker if data segment does not start with one.
			byte[] twoBytes = new byte[2];
			stream.readFully(twoBytes);
			if (!((twoBytes[0] & 0xff) == 0xff && (twoBytes[1] & 0xff) == SOS)) {
				// Segment does not start with SOS marker;
				// use the pre-calculated SOS marker.
				System.arraycopy(_SOSMarker, 0, buf, bufOffset,
						_SOSMarker.length);
				bufOffset += _SOSMarker.length;
			}

			// Copy the segment data into the buffer.
			buf[bufOffset++] = twoBytes[0];
			buf[bufOffset++] = twoBytes[1];
			stream.readFully(buf, bufOffset, byteCount - 2);
			bufOffset += byteCount - 2;

			// EOI.
			buf[bufOffset++] = (byte)0xff; // Marker identifier
			buf[bufOffset++] = (byte)EOI;

			ByteArrayInputStream bais =
					new ByteArrayInputStream(buf, 0, bufOffset);
			ImageInputStream is = new MemoryCacheImageInputStream(bais);

			jPEGReader.setInput(is, true, true);
		}

		// Read real image
		jPEGParam.setDestination(rawImage);
		jPEGReader.read(0, jPEGParam);
	}

	@Override
	protected void finalize() throws Throwable {
		super.finalize();
		jPEGReader.dispose();
	}
	//CHECKSTYLE:ON
}
