/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane;

import java.awt.Image;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.imageio.spi.IIORegistry;
import javax.swing.ImageIcon;

import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.tree.ItemTreeModel;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Presenter;
import fr.greyc.pantheon.ariane.view.ArianeFrame;
import fr.greyc.pantheon.ariane.view.SplashScreen;

/**
 * <code>Ariane</code> is the main class of the software.
 * It reads the configuration file and initializes the application.
 * It uses a Splash Screen to pass the time.
 */
public final class Ariane {

   private static Presenter _presenter;
   private static ArianeFrame _mainFrame;

   public static void main( String[] args ) {
      installImageIOLibrary();
      new Ariane().launch(args);
   }

   public void initForTest() {
      UserPreferences.initializePreferencesWithDefaultValues();
      _presenter = new Presenter(new ItemTreeModel());
   }

   public void launch( String[] args ) {
      final Map<String, Object> options = new HashMap<>();
      options.put("nosplash", false);

      final List<String> graphFilenames = CommandLineParser.parse(args, options);

      boolean noSplashOption = (Boolean) options.get("nosplash");
      final String configDirName = (String) options.get("config");
      final String operatorLibraryFilename = (String) options.get("lib");
      final String scriptFilename = (String) options.get("generate");
      initializeInterface(noSplashOption, operatorLibraryFilename, graphFilenames, configDirName, scriptFilename);
   }

   public static ArianeFrame getFrame() {
      return _mainFrame;
   }

   public static Presenter getPresenter() {
      return _presenter;
   }

   protected List<XMLElement> readConfigurationFile( File configFile ) {
      if (configFile != null) {
         try {
            return ArianeSettings.readXMLFile(configFile);
         } catch (IOException e) {
            System.err.println(ArianeSettings.getString("error") + e.getLocalizedMessage());
         }
      }
      return null;
   }

   protected String locateLibraryPath( List<XMLElement> preferenceHandler ) {
      if (preferenceHandler != null) {
         UserPreferences.load(preferenceHandler);
         String libraryFilename = ArianeSettings.getLibraryPath() + ArianeSettings.getLibraryConfigFilename();
         File file = new File(libraryFilename);
         if (file.exists()) {
            return libraryFilename;
         }
         // Kept for the sake of compatibility
         libraryFilename = ArianeSettings.getLibraryPath() + File.separator + "lib" + File.separator + "pandore.cfg";
         File file1 = new File(libraryFilename);
         if (file1.exists()) {
            return libraryFilename;
         }
      }
      return null;
   }

   private static void installImageIOLibrary() {
      IIORegistry registry = IIORegistry.getDefaultInstance();
      registry.registerServiceProvider(new com.sun.media.imageioimpl.plugins.tiff.TIFFImageReaderSpi());
   }

   private void initializeInterface( boolean noSplashOption, String operatorLibraryFilename, List<String> graphFilenames, String configDirName, String scriptFilename ) {
      boolean graphicalInstance;
      boolean doSplash;
      if (scriptFilename != null) {
         doSplash = false;
         graphicalInstance = false;
      } else {
         graphicalInstance = true;
         doSplash = !noSplashOption;
      }

      _mainFrame = new ArianeFrame();
      _mainFrame.installUI();
      SplashScreen splash = new SplashScreen(ArianeSettings.getIcon("splash.png"));
      splash.setVisible(doSplash);

      splash.showProgress("Reading preferences file ...", 0);
      UserPreferences.initializePreferencesWithDefaultValues();
      File preferencesFile = ArianeSettings.locateConfigureFileIn(configDirName);
      List<XMLElement> preferenceHandler = readConfigurationFile(preferencesFile);
      String defaultLibraryFilename = locateLibraryPath(preferenceHandler);

      splash.showProgress("Reading common operators...", 20);
      ItemTreeModel itemTree = readCommonOperatorLibrary(operatorLibraryFilename, defaultLibraryFilename);

      splash.showProgress("Reading personal operators ...", 40);
      readPersonalOperatorLibrary(preferencesFile, preferenceHandler, itemTree);

      splash.showProgress("Reading subroutines ...", 50);
      readSubroutineLibrary(itemTree);

      splash.showProgress("Building graphical interfaces ...", 60);
      _presenter = new Presenter(itemTree);

      if (graphicalInstance) {
         splash.showProgress("Initializing windows contents ...", 80);
         createWorkscape();
         if (!graphFilenames.isEmpty()) {
            splash.showProgress("Reading input workspace(s) ...", 90);
            readInputWorkspaces(graphFilenames);
         }

         splash.showProgress("Starting session ...", 100);
         startSession();
         splash.setVisible(false);
         splash.dispose();

         if (isFirstUsed()) {
            _presenter.chooseOperatorLibrary();
         }
      } else {
         if (_presenter.generatesScriptCodeFromCurrentGraph(scriptFilename, graphFilenames)) {
            System.err.println("Script file " + scriptFilename + " generated.");
         }
         System.exit(0);
      }
   }

   private void startSession() {
      _mainFrame.restoreFrameSize();
      _presenter.restoreSplitPanesLocation();
      _mainFrame.setVisible(true);
   }

   private void readInputWorkspaces( List<String> graphFilenames ) {
      _presenter.display(graphFilenames);
      File file = new File(graphFilenames.get(0));
      UserPreferences.setCurrentDirectory(file.getAbsoluteFile().getParent());
   }

   private void createWorkscape() {
      ImageIcon logo64 = ArianeSettings.getIcon("logo64.png");
      ImageIcon logo32 = ArianeSettings.getIcon("logo32.png");
      _mainFrame.setIconImages(Arrays.asList(new Image[]{logo64.getImage(), logo32.getImage()}));
      _mainFrame.setJMenuBar(_presenter.buildMenuBar());
      _presenter.initContent();
      _mainFrame.initFrame(_presenter);
   }

   private void readSubroutineLibrary( ItemTreeModel itemTree ) {
      File subRoutineFile = new File(ArianeSettings.getSubroutineFilename());
      if (subRoutineFile.exists()) {
         List<XMLElement> subRoutineHandler = readConfigurationFile(subRoutineFile);
         itemTree.addInSubroutines(subRoutineHandler);
      }
   }

   private void readPersonalOperatorLibrary( File preferencesFile, List<XMLElement> preferenceHandler, ItemTreeModel itemTree ) {
      File personalOperatorFile = new File(ArianeSettings.getPersonalOperatorFilename());
      if (personalOperatorFile.exists()) {
         itemTree.addInUserOperators(readConfigurationFile(personalOperatorFile));
      } else if (preferenceHandler != null) { // Keep this for sake of compatibility
         itemTree.addInUserOperators(preferenceHandler);
         File directory = personalOperatorFile.getParentFile();
         directory.mkdirs();
         preferencesFile.renameTo(personalOperatorFile);
      }
   }

   private ItemTreeModel readCommonOperatorLibrary( String operatorLibraryFilename, String defaultLibraryFilename ) {
      List<XMLElement> libraryHandler;
      if (operatorLibraryFilename != null) {
         libraryHandler = readConfigurationFile(new File(operatorLibraryFilename));
      } else if (defaultLibraryFilename != null && UserPreferences.isOperatorLibraryUsed()) {
         libraryHandler = readConfigurationFile(new File(defaultLibraryFilename));
      } else {
         libraryHandler = new ArrayList<>();
      }
      ItemTreeModel itemTree = new ItemTreeModel();
      itemTree.setCommonOperators(libraryHandler);
      return itemTree;
   }

   private boolean isFirstUsed() {
      return ".".equalsIgnoreCase(ArianeSettings.getLibraryPath()) && UserPreferences.isOperatorLibraryUsed();
   }
}
