/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane;

import java.awt.Toolkit;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.swing.ImageIcon;
import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

import fr.greyc.pantheon.ariane.model.PlatformSettings;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.model.xml.XMLFileInputStream;
import java.util.Arrays;
import java.util.Collections;

public final class ArianeSettings {
   public static final String FRAME_TITLE = "Au fil d'Ariane ... @ GREYC (ENSICAEN & Université de Caen Basse-Normandie)";
   public static final String VERSION_ID = "3.4.2 (Apr 14, 2016)";
   public static final List<String> AUTHORS = Collections.unmodifiableList(Arrays.asList(
      "Régis CLOUARD",
      "François RIOULT",
      "Frédéric LAIR (2007)",
      "Guillaume CHATELET (2002)",
      "Alban DOUBLET (2002)",
      "Mathieu JOETS (2001)",
      "Adrien SCHRÈQUE (2000)",
      "Hicham AMADI (2000)",
      "Sébastien LEFÈVRE (1999)",
      "Richard LOBRÉAU (1999)"));
 
   public static final String DEFAULT_ARIANE_FILE_EXT = ".aws";
   public static final String CONFIG_STATIC_DIRNAME = ".greyc" + File.separator + "ariane";
   public static final String CONFIG_DYNAMIC_DIRNAME = ".ariane";

   public static final String PREFERENCES_FILENAME = "ariane.ini";

   private static final String CONFIG_SUBROUTINE_FILENAME = "subroutines.cfg";
   private static final String CONFIG_OPERATOR_FILENAME = "operators.cfg";
   private static final String STRING_RESOURCE_DIRECTORY = "fr/greyc/pantheon/ariane/resources/locale/";
   private static final String STRING_RESOURCE_FILENAME = "messages";

   private static final String ICON_RESOURCE_DIRECTORY = "fr/greyc/pantheon/ariane/resources/images/";
   private static final String LIBRARY_CONFIG_FILENAME = "ariane.cfg";

   public enum ErrorCode {
      NOERROR,
      MISSINGPARAMETER,
      MISSINGRESULT,
      MISSINGINPUT,
      RUNTIME,
      INPUTOUTPUT,
      INTERRUPTEDEXECUTION,
      MISSINGFILE,
      UNKNOWNFILE,
      MISSINGFORMULA,
      ARITHMETICS,
   }

   private static String _operatorLibraryPath = ".";
   private static String _configurationDirectoryPath;
   private static ResourceBundle _language;
   private static String _arianeManualURL;
   private static String _libraryDocumentationPath = "";

   private ArianeSettings() {
   }

   public static String getLibraryConfigFilename() {
      return File.separator + "lib" + File.separator + LIBRARY_CONFIG_FILENAME;
   }

   public static ImageIcon getIcon( String filename ) {
      URL location = Ariane.class.getClassLoader().getResource(ICON_RESOURCE_DIRECTORY + filename);
      return new ImageIcon(Toolkit.getDefaultToolkit().getImage(location));
   }

   public static URL getDocumentationURI( String page ) {
      return Ariane.class.getClassLoader().getResource(ArianeSettings._arianeManualURL + page);
   }

   public static String getString( final String name ) {
      return ArianeSettings._language.getString(name);
   }

   public static String getSubroutineFilename() {
      return ArianeSettings._configurationDirectoryPath + File.separator + CONFIG_SUBROUTINE_FILENAME;
   }

   public static ResourceBundle getLanguage() {
      return ArianeSettings._language;
   }

   public static String getPreferencesFilename() {
      return ArianeSettings._configurationDirectoryPath + File.separator + PREFERENCES_FILENAME;
   }

   public static String getPersonalOperatorFilename() {
      return ArianeSettings._configurationDirectoryPath + File.separator + CONFIG_OPERATOR_FILENAME;
   }

   public static void setLibraryPath( String path ) {
      ArianeSettings._operatorLibraryPath = path;
      System.setProperty("java.library.path", PlatformSettings.getJavaLibraryPath() + File.pathSeparatorChar + ArianeSettings._operatorLibraryPath + File.separatorChar + "lib");
   }

   public static String getLibraryPath() {
      return ArianeSettings._operatorLibraryPath;
   }

   public static void setLibraryDocumentationPath( final String path ) {
      ArianeSettings._libraryDocumentationPath = path;
   }

   public static String getLibraryDocumentationPath() {
      return ArianeSettings._libraryDocumentationPath + "/" + UserPreferences.getCurrentLocale().getLanguage();
   }

   public static void setCurrentLocale( Locale locale ) {
      final String url = STRING_RESOURCE_DIRECTORY + locale.getLanguage() + "/";
      ArianeSettings._language = ResourceBundle.getBundle(url + STRING_RESOURCE_FILENAME, locale);
      ArianeSettings._arianeManualURL = url + "manual/";
   }

   public static File locateConfigureFileIn( String directoryName ) {
      // 1. use the specified directory name
      if (directoryName != null) {
         File configurationFile = new File(directoryName + File.separator + ArianeSettings.PREFERENCES_FILENAME);
         if (configurationFile.exists()) {
            ArianeSettings._configurationDirectoryPath = configurationFile.getAbsoluteFile().getParent();
            return configurationFile;
         }
      }
      // 2. The config directory is in the  same directory than the jar file.
      final String[] possibleConfigJarFiles = {
         File.separator + CONFIG_DYNAMIC_DIRNAME + File.separator + PREFERENCES_FILENAME,
         File.separator + "ariane" + File.separator + PREFERENCES_FILENAME, // For compatibility
         File.separator + "ariane.cfg" // For compatibility
      };
      final String currentJarPath = PlatformSettings.getCurrentJarPath();
      for (String configFile : possibleConfigJarFiles) {
         File configurationFile = new File(currentJarPath + configFile);
         if (configurationFile.exists()) {
            _configurationDirectoryPath = configurationFile.getParent();
            return configurationFile;
         }
      }
      // 3. The config directory is in user's config directory
      final String[] possibleConfigFiles = {
         File.separator + ArianeSettings.CONFIG_STATIC_DIRNAME + File.separator + ArianeSettings.PREFERENCES_FILENAME,
         File.separator + ".greyc" + File.separator + "ariane.cfg", // For compatibility
         File.separator + ".ariane" // For compatibility
      };
      for (String configurationFileName : possibleConfigFiles) {
         File configurationFile1 = new File(PlatformSettings.getHomePath() + configurationFileName);
         if (configurationFile1.exists()) {
            ArianeSettings._configurationDirectoryPath = PlatformSettings.getHomePath() + File.separator + ArianeSettings.CONFIG_STATIC_DIRNAME;
            return configurationFile1;
         }
      }
      // 4. default configuration file
      ArianeSettings._configurationDirectoryPath = PlatformSettings.getHomePath() + File.separator + ArianeSettings.CONFIG_STATIC_DIRNAME;
      return null;
   }

   public static List<XMLElement> readXMLFile( File configFile ) throws IOException {
      try {
         XMLFileInputStream convert = new XMLFileInputStream(configFile);
         convert.parse();
         convert.close();
         return convert.getElements();
      } catch (SAXException | IOException | ParserConfigurationException e) {
         throw new IOException(e);
      }
   }
}
