/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Frame;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.StringTokenizer;

import javax.swing.ToolTipManager;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.model.xml.XMLFileOutputStream;
import fr.greyc.pantheon.ariane.view.Workspace;
import fr.greyc.pantheon.ariane.view.graph.GraphView;
import fr.greyc.pantheon.ariane.view.graph.item.etc.ControlColor;
import fr.greyc.pantheon.ariane.view.graph.item.etc.ControlStructureColor;
import fr.greyc.pantheon.ariane.view.graph.item.etc.DataFileColor;
import fr.greyc.pantheon.ariane.view.graph.item.etc.NoteColor;
import fr.greyc.pantheon.ariane.view.graph.item.etc.OperatorColor;

/**
 * The class <code>UserPreferences</code> stores the current values
 * for all user's preferences.
 * <br>
 * This class sounds like the @link{java.util.prefs.Preferences Preferences class},
 * but the persistence is ensured by our own methods.
 * Each preference is handled using the static get and put methods:
 * <pre>
 * Object value = PreferencesSettings.get("property_name",default_value);
 * PreferencesSettings.set("property_name", value);
 * </pre>
 */
public final class UserPreferences {

   private static final String UNDO_HISTORY_SIZE = "UndoHistorySize";
   private static final String USE_OPERATOR_LIBRARY = "UseOperatorLibrary";
   private static final String PREFERRED_BINARY_DIRECTORY = "BinaryDirectory";
   private static final String LANGUAGE = "Language";
   private static final String FRAME_SIZE = "Frame.Size";
   private static final String TEXT_EDITOR = "TextEditor";
   private static final String USE_TOOLTIP = "UseToolTip";
   private static final String WEB_BROWSER = "WebBrowser";
   private static final String PREFERRED_FILE_DIRECTORY = "FileDirectory";
   private static final String FONT_SIZE = "FontSize";
   private static final String CODE_GENERATOR_PREFERENCE = "CodeGenerator.Preference";
   private static final String TEMPORARY_FILE_EXTENSION = "TemporaryFileExtension";
   private static final String FRAME_MAXIMIZED = "Frame.Maximized";
   private static final String SPLITPANE_H1_RATIO = "SplitPane.H1.Ratio";
   private static final String SPLITPANE_H2_RATIO = "SplitPane.H2.Ratio";
   private static final String SPLITPANE_V1_RATIO = "SplitPane.V1.Ratio";
   private static final String SPLITPANE_V2_RATIO = "SplitPane.V2.Ratio";
   private static final String COMMAND_STATUS = "Command.Status";
   private static final String COMMAND_VISU = "Command.Visu";
   private static final String DATAFILE_COLOR = "Datafile.Color";
   private static final String OPERATOR_COLOR = "Operator.Color";
   private static final String CONTROL_COLOR = "Control.Color";
   private static final String COMPOSITE_COLOR = "Composite.Color";
   private static final String NOTE_COLOR = "Note.Color";
   private static final String DATA_LINK_COLOR = "Image.Link.Color";
   private static final String VALUE_LINK_COLOR = "Value.Link.Color";
   private static final String MASK_LINK_COLOR = "Mask.Link.Color";

   private static String _commandStatus = "";
   private static String _commandVisu = "";

   private static String _temporaryDirectory;
   private static String _currentDirectory;
   private static String _temporaryFileExtension = ".tmp";
   private static final String TEMPORARY_FILE_PREFIX = "data";

   private static Locale _locale;
   private static Font _defaultFont;

   private static final Map<String, Object> PREFERENCES = new HashMap<>();

   private UserPreferences() {
   }

   public static void initializePreferencesWithDefaultValues() {
      putDimension(FRAME_SIZE, new Dimension(1000, 660));

      putInt(SPLITPANE_H1_RATIO, 436);
      putInt(SPLITPANE_V1_RATIO, 232);
      putInt(SPLITPANE_H2_RATIO, 433);
      putInt(SPLITPANE_V2_RATIO, 584);
      putInt(CODE_GENERATOR_PREFERENCE, 0);
      putBoolean(USE_OPERATOR_LIBRARY, true);
      putBoolean(USE_TOOLTIP, true);
      putInt(UNDO_HISTORY_SIZE, 200);
      putInt(FONT_SIZE, 10);

      setDefaultLocale();
      _defaultFont = new Font("Arial", Font.PLAIN, getDefaultFontSize());
      setCurrentDirectory(System.getProperty("user.dir") + File.separator);
      try {
         setTemporaryDirectoryBase(getSystemTemporaryDirectory());
      } catch (IOException e) {
      }
   }

   public static void setTemporaryDirectoryBase( String directory ) throws IOException {
      if (directory.endsWith(File.separator)) {
         createTemporaryFolder(directory);
      } else {
         createTemporaryFolder(directory + File.separator);
      }
   }

   public static String getTemporaryDirectoryBase() {
      File file = new File(_temporaryDirectory);
      return file.getParent() + File.separator;
   }

   public static void setDefaultFontSize( int size ) {
      putInt(FONT_SIZE, size);
   }

   public static int getDefaultFontSize() {
      return getInt(FONT_SIZE);
   }

   public static Font getDefaultFont() {
      return _defaultFont;
   }

   public static String getTemporaryDirectory() {
      return _temporaryDirectory;
   }

   public static String createTemporaryName( int graphIndex, int itemIndex, int slotIndex ) {
      String name = TEMPORARY_FILE_PREFIX
          + "_"
          + graphIndex
          + "_" + itemIndex
          + "_" + slotIndex + UserPreferences._temporaryFileExtension;
      return name;
   }

   public static String getTemporaryFilePrefix() {
      return TEMPORARY_FILE_PREFIX;
   }

   public static String getTemporaryFileExtension() {
      return _temporaryFileExtension;
   }

   public static void setTemporaryFileExtension( String extension ) {
      _temporaryFileExtension = extension;
   }

   public static Locale getCurrentLocale() {
      return _locale;
   }

   public static void setToolTipState( boolean state ) {
      putBoolean(USE_TOOLTIP, state);
      ToolTipManager.sharedInstance().setEnabled(state);
   }

   public static boolean isToolTipEnabled() {
      return getBoolean(USE_TOOLTIP);
   }

   public static void setCommandVisu( String name ) {
      _commandVisu = name.trim();
   }

   public static String getCommandVisu() {
      return _commandVisu;
   }

   public static void setCommandStatus( String name ) {
      _commandStatus = name.trim();
   }

   public static String getCommandStatus() {
      return _commandStatus;
   }

   public static void changeLanguage( String language ) {
      String currentLanguage = get(LANGUAGE);
      if (!currentLanguage.equals(language)) {
         setLocale(language);
         Ariane.getPresenter().changeLanguage();
      }
   }

   public static String getLanguage() {
      return get(UserPreferences.LANGUAGE);
   }

   public static int getHistorySize() {
      return getInt(UserPreferences.UNDO_HISTORY_SIZE);
   }

   public static void setHistorySize( int size ) {
      putInt(UNDO_HISTORY_SIZE, size);
   }

   public static String getWebBrowserPath() {
      return get(WEB_BROWSER);
   }

   public static void setWebBrowserPath( String path ) {
      put(WEB_BROWSER, path);
   }

   public static String getTextEditorPath() {
      return get(TEXT_EDITOR);
   }

   public static void setTextEditorPath( String path ) {
      put(TEXT_EDITOR, path);
   }

   public static String getPreferredBinaryDirectory() {
      return get(PREFERRED_BINARY_DIRECTORY);
   }

   public static void setBinaryDirectory( String path ) {
      if (path.trim().isEmpty()) {
         put(PREFERRED_BINARY_DIRECTORY, null);
      } else {
         put(PREFERRED_BINARY_DIRECTORY, path.trim());
      }
   }

   public static boolean isOperatorLibraryUsed() {
      return getBoolean(USE_OPERATOR_LIBRARY);
   }

   public static void setOperatorLibraryState( boolean state ) {
      putBoolean(UserPreferences.USE_OPERATOR_LIBRARY, state);
   }

   public static String getPreferredFileDirectory() {
      return get(PREFERRED_FILE_DIRECTORY);
   }

   public static void setPreferedFileDirectory( String path ) {
      if (path.trim().isEmpty()) {
         put(PREFERRED_FILE_DIRECTORY, null);
      } else {
         put(PREFERRED_FILE_DIRECTORY, path.trim());
      }
   }

   public static int getCodeGeneratorIndex() {
      return getInt(CODE_GENERATOR_PREFERENCE);
   }

   public static void getCodeGeneratorIndex( int index ) {
      putInt(CODE_GENERATOR_PREFERENCE, index);
   }

   public static Dimension getFrameSize() {
      return getDimension(FRAME_SIZE);
   }

   public static String getCurrentDirectory() {
      return _currentDirectory;
   }

   public static void setCurrentDirectory( String directory ) {
      _currentDirectory = directory;
   }

   public static boolean isFrameMaximized() {
      return UserPreferences.getBoolean(FRAME_MAXIMIZED);
   }

   public static int[] getSplitPaneH1Ratio() {
      return new int[]{
         getInt(SPLITPANE_H1_RATIO),
         getInt(SPLITPANE_V1_RATIO),
         getInt(SPLITPANE_H2_RATIO),
         getInt(SPLITPANE_V2_RATIO)
      };
   }

   public static NamedColor[] getConfigurableColors() {
      return new NamedColor[]{
         new NamedColor(NoteColor.getColor(), ArianeSettings.getString("PreferenceChooser.note")),
         new NamedColor(DataFileColor.getColor(), ArianeSettings.getString("PreferenceChooser.datafile")),
         new NamedColor(OperatorColor.getColor(), ArianeSettings.getString("PreferenceChooser.operator")),
         new NamedColor(ControlColor.getColor(), ArianeSettings.getString("PreferenceChooser.controlItem")),
         new NamedColor(ControlStructureColor.getColor(), ArianeSettings.getString("PreferenceChooser.controlStructureItem")),
         new NamedColor(GraphView.getBackgroundColor(), ArianeSettings.getString("PreferenceChooser.background")),
         new NamedColor(GraphView.getImageLinkColor(), ArianeSettings.getString("PreferenceChooser.imagelinks")),
         new NamedColor(GraphView.getValueLinkColor(), ArianeSettings.getString("PreferenceChooser.numericlinks")),
         new NamedColor(GraphView.getMaskLinkColor(), ArianeSettings.getString("PreferenceChooser.masklinks"))
      };
   }

   public static void setConfigurableColors( NamedColor[] colors ) {
      int index = 0;
      NoteColor.setColor(colors[index++].getColor());
      DataFileColor.setColor(colors[index++].getColor());
      OperatorColor.setColor(colors[index++].getColor());
      ControlColor.setColor(colors[index++].getColor());
      ControlStructureColor.setColor(colors[index++].getColor());

      GraphView.setBackgroundColor(colors[index++].getColor());
      GraphView.setImageLinkColor(colors[index++].getColor());
      GraphView.setValueLinkColor(colors[index++].getColor());
      GraphView.setMaskLinkColor(colors[index++].getColor());
   }

   public static int deleteTemporaryFiles() {
      File directory = new File(getTemporaryDirectory());
      String[] files = directory.list();
      if (files == null) {
         return 0;
      }

      int numberOfErasedFiles = 0;
      for (String filename : files) {
         if (filename.startsWith(getTemporaryFilePrefix()) && filename.endsWith(getTemporaryFileExtension())) {
            File file = new File(getTemporaryDirectory() + filename);
            if (file.delete()) {
               numberOfErasedFiles++;
            } else {
               Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("PreferenceChooser.environment.error1") + " " + filename);
            }
         }
      }
      directory.delete();
      return numberOfErasedFiles;
   }

   public static boolean updateUserPreferenceFile() {
      File file = new File(ArianeSettings.getPreferencesFilename());
      File directory = file.getParentFile();
      directory.mkdirs();
      try {
         XMLFileOutputStream out = new XMLFileOutputStream(file, null);
         out.write(convertPreferencesToXml());
         out.write(convertRecentFilesToXml());
         out.close();
      } catch (IOException ex) {
         Ariane.getPresenter().printSystemMessage(ArianeSettings.getString("PreferenceChooser.error.saved") + " " + ArianeSettings.getPreferencesFilename());
         return false;
      }
      return true;
   }

   public static void load( List<XMLElement> xmlTree1 ) {
      // Discards tag <ARIANE> from the xmlTree.
      List<XMLElement> properties = xmlTree1.get(0).getChildren();
      Iterator<XMLElement> iterator = properties.iterator();

      XMLElement element;
      try {
         while (iterator.hasNext()) {
            element = iterator.next();
            if ("Preferences".equalsIgnoreCase(element.getName())) {
               loadPreferences(element.getChildren());
            } else if ("RecentFiles".equalsIgnoreCase(element.getName()) && element.getChildren() != null) {
               convertXMLToRecentFiles(element.getChildren());
            }
         }
      } catch (IOException e) {
      }
   }

   private static void setFontSize( int size ) {
      _defaultFont = _defaultFont.deriveFont(size);
      setDefaultFontSize(size);
   }

   private static void setDefaultLocale() {
      _locale = Locale.getDefault();
      if (_locale.getDisplayLanguage().equals(Locale.FRENCH.getDisplayLanguage())) {
         setLocale("Français");
      } else {
         setLocale("English");
      }
   }

   private static void putBoolean( final String key, final boolean value ) {
      PREFERENCES.put(key, value);
   }

   private static boolean getBoolean( String key ) {
      if (PREFERENCES.containsKey(key)) {
         return (Boolean) PREFERENCES.get(key);
      } else {
         return false;
      }
   }

   private static void put( String key, String value ) {
      PREFERENCES.put(key, value);
   }

   private static String get( String key ) {
      return (String) PREFERENCES.get(key);
   }

   private static void putDimension( String key, Dimension value ) {
      PREFERENCES.put(key, value);
   }

   private static Dimension getDimension( String key ) {
      if (PREFERENCES.containsKey(key)) {
         return (Dimension) PREFERENCES.get(key);
      } else {
         return null;
      }
   }

   private static void putInt( String key, int value ) {
      PREFERENCES.put(key, value);
   }

   private static int getInt( String key ) {
      return (Integer) PREFERENCES.get(key);
   }

   /**
    * Creates a directory unique in the current directory base.
    * This allows multiple Ariane at the same time.
    *
    * @param directoryName a regular directory name.
    */
   private static void createTemporaryFolder( String directoryName ) throws IOException {
      File directory = new File(directoryName);
      if (directory.isDirectory()) {
         if (Files.isWritable(directory.toPath())) {
            File file;
            Long index = 1L;
            do {
               file = new File(directoryName + "ariane" + index++);
            } while (!file.mkdirs());
            _temporaryDirectory = file.getAbsolutePath() + File.separator;
         } else {
            throw new IOException(ArianeSettings.getLanguage().getString("UserPreferences.error2") + ": " + directoryName);
         }
      } else {
         throw new IOException(ArianeSettings.getLanguage().getString("UserPreferences.error1") + ": " + directoryName);
      }
   }

   private static String getSystemTemporaryDirectory() {
      return System.getProperty("java.io.tmpdir") + File.separator;
   }

   /**
    * Sets the current language to the specified language.
    * Loads related resources from the resource files.
    *
    * @param language the name of the language (eg., français, english).
    */
   private static void setLocale( String language ) {
      put(LANGUAGE, language);
      if ("Français".equals(language)) {
         _locale = new Locale("fr", "FR");
      } else if ("English".equals(language)) {
         _locale = new Locale("en", "US");
      }
      Locale.setDefault(_locale);
      ArianeSettings.setCurrentLocale(_locale);
   }

   private static String codeAsTriplet( Color color ) {
      return color.getRed() + "," + color.getGreen() + "," + color.getBlue();
   }

   private static Color decodeFromTriplet( String color ) {
      StringTokenizer colorValuePart = new StringTokenizer(color, ",");

      int red = Integer.parseInt(colorValuePart.nextToken());
      int green = Integer.parseInt(colorValuePart.nextToken());
      int blue = Integer.parseInt(colorValuePart.nextToken());
      return new Color(red, green, blue);
   }

   private static XMLElement convertPreferencesToXml() throws IOException {
      XMLElement element = new XMLElement("Preferences");
      element.addChildren(new XMLElement("CurrentDirectory", getCurrentDirectory()));
      if (!getSystemTemporaryDirectory().equals(getTemporaryDirectoryBase())) {
         element.addChildren(new XMLElement("TemporaryDirectory", getTemporaryDirectoryBase()));
      }
      element.addChildren(new XMLElement("OperatorLibraryDirectory", ArianeSettings.getLibraryPath()));
      if (!UserPreferences.getCommandVisu().isEmpty()) {
         element.addChildren(new XMLElement(COMMAND_VISU, UserPreferences.getCommandVisu()));
      }
      if (!UserPreferences.getCommandStatus().isEmpty()) {
         element.addChildren(new XMLElement(COMMAND_STATUS, UserPreferences.getCommandStatus()));
      }
      if (!NoteColor.getDefaultColor().equals(NoteColor.getColor())) {
         element.addChildren(new XMLElement(NOTE_COLOR, codeAsTriplet(NoteColor.getColor())));
      }
      if (!OperatorColor.getDefaultColor().equals(OperatorColor.getColor())) {
         element.addChildren(new XMLElement(OPERATOR_COLOR, codeAsTriplet(OperatorColor.getColor())));
      }
      if (!DataFileColor.getDefaultColor().equals(DataFileColor.getColor())) {
         element.addChildren(new XMLElement(DATAFILE_COLOR, codeAsTriplet(DataFileColor.getColor())));
      }
      if (!ControlColor.getDefaultColor().equals(ControlColor.getColor())) {
         element.addChildren(new XMLElement(CONTROL_COLOR, codeAsTriplet(ControlColor.getColor())));
      }
      if (!ControlStructureColor.getDefaultColor().equals(ControlStructureColor.getColor())) {
         element.addChildren(new XMLElement(COMPOSITE_COLOR, codeAsTriplet(ControlStructureColor.getColor())));
      }
      if (!GraphView.getDefaultBackgroundColor().equals(GraphView.getBackgroundColor())) {
         element.addChildren(new XMLElement("Graph.Background.Color", codeAsTriplet(GraphView.getBackgroundColor())));
      }
      if (!GraphView.getDefaultImageLinkColor().equals(GraphView.getImageLinkColor())) {
         element.addChildren(new XMLElement(DATA_LINK_COLOR, codeAsTriplet(GraphView.getImageLinkColor())));
      }
      if (!GraphView.getDefaultValueLinkColor().equals(GraphView.getValueLinkColor())) {
         element.addChildren(new XMLElement(VALUE_LINK_COLOR, codeAsTriplet(GraphView.getValueLinkColor())));
      }
      if (!GraphView.getDefaultMaskLinkColor().equals(GraphView.getMaskLinkColor())) {
         element.addChildren(new XMLElement(MASK_LINK_COLOR, codeAsTriplet(GraphView.getMaskLinkColor())));
      }
      element.addChildren(new XMLElement(LANGUAGE, get(LANGUAGE)));
      element.addChildren(new XMLElement(FONT_SIZE, Integer.toString(getInt(FONT_SIZE))));
      element.addChildren(new XMLElement("ToolTip", Boolean.toString(getBoolean(USE_TOOLTIP))));
      String web = get(WEB_BROWSER);
      if (web != null && !web.trim().isEmpty()) {
         element.addChildren(new XMLElement(WEB_BROWSER, web));
      }
      String editor = get(TEXT_EDITOR);
      if (editor != null && !editor.trim().isEmpty()) {
         element.addChildren(new XMLElement(TEXT_EDITOR, editor));
      }

      Dimension dimension = Ariane.getFrame().getFinalSize();
      element.addChildren(new XMLElement("Frame.Width", Integer.toString(dimension.width)));
      element.addChildren(new XMLElement("Frame.Height", Integer.toString(dimension.height)));
      element.addChildren(new XMLElement(FRAME_MAXIMIZED, Boolean.toString(Ariane.getFrame().getExtendedState() == Frame.MAXIMIZED_BOTH)));
      int[] ratios = Ariane.getPresenter().getSplitPaneRatios();
      element.addChildren(new XMLElement("SplitH1.ratio", Integer.toString(ratios[0])));
      element.addChildren(new XMLElement("SplitV1.ratio", Integer.toString(ratios[1])));
      element.addChildren(new XMLElement("SplitH2.ratio", Integer.toString(ratios[2])));
      element.addChildren(new XMLElement("SplitV2.ratio", Integer.toString(ratios[3])));
      element.addChildren(new XMLElement(USE_OPERATOR_LIBRARY, Boolean.toString(getBoolean(USE_OPERATOR_LIBRARY))));
      element.addChildren(new XMLElement(TEMPORARY_FILE_EXTENSION, _temporaryFileExtension));
      element.addChildren(new XMLElement(UNDO_HISTORY_SIZE, Integer.toString(getInt(UNDO_HISTORY_SIZE))));
      element.addChildren(new XMLElement(CODE_GENERATOR_PREFERENCE, Integer.toString(getInt(CODE_GENERATOR_PREFERENCE))));
      if (get(PREFERRED_FILE_DIRECTORY) != null) {
         element.addChildren(new XMLElement(PREFERRED_FILE_DIRECTORY, get(PREFERRED_FILE_DIRECTORY)));
      }
      if (get(PREFERRED_BINARY_DIRECTORY) != null) {
         element.addChildren(new XMLElement(PREFERRED_BINARY_DIRECTORY, get(PREFERRED_BINARY_DIRECTORY)));
      }
      return element;
   }

   private static void loadPreferences( List<XMLElement> xmlTree ) throws IOException {
      Iterator<XMLElement> iterator = xmlTree.iterator();
      while (iterator.hasNext()) {
         XMLElement element = iterator.next();
         // A l'interieur de <Preferences>
         String name = element.getName();
         if ("CurrentDir".equalsIgnoreCase(name) || "CurrentDirectory".equalsIgnoreCase(name)) {
            setCurrentDirectory(element.getValue());
         } else if ("TempDir".equalsIgnoreCase(name) || "TemporaryDirectory".equalsIgnoreCase(name)) {
            setTemporaryDirectoryBase(element.getValue());
         } else if ("PandoreDir".equalsIgnoreCase(name) || "OperatorLibraryDirectory".equalsIgnoreCase(name)) {
            ArianeSettings.setLibraryPath(element.getValue());
         } else if (COMMAND_VISU.equalsIgnoreCase(name)) {
            setCommandVisu(element.getValue());
         } else if (COMMAND_STATUS.equalsIgnoreCase(name)) {
            setCommandStatus(element.getValue());
         } else if (NOTE_COLOR.equalsIgnoreCase(name)) {
            NoteColor.setColor(decodeFromTriplet(element.getValue()));
         } else if (OPERATOR_COLOR.equalsIgnoreCase(name)) {
            OperatorColor.setColor(decodeFromTriplet(element.getValue()));
         } else if (DATAFILE_COLOR.equalsIgnoreCase(name)) {
            DataFileColor.setColor(decodeFromTriplet(element.getValue()));
         } else if (CONTROL_COLOR.equalsIgnoreCase(name)) {
            ControlColor.setColor(decodeFromTriplet(element.getValue()));
         } else if (COMPOSITE_COLOR.equalsIgnoreCase(name)) {
            ControlStructureColor.setColor(decodeFromTriplet(element.getValue()));
         } else if ("Graph.Background.Color".equalsIgnoreCase(name)) {
            GraphView.setBackgroundColor(decodeFromTriplet(element.getValue()));
         } else if (DATA_LINK_COLOR.equalsIgnoreCase(name)) {
            GraphView.setImageLinkColor(decodeFromTriplet(element.getValue()));
         } else if (VALUE_LINK_COLOR.equalsIgnoreCase(name)) {
            GraphView.setValueLinkColor(decodeFromTriplet(element.getValue()));
         } else if (MASK_LINK_COLOR.equalsIgnoreCase(name)) {
            GraphView.setMaskLinkColor(decodeFromTriplet(element.getValue()));
         } else if (LANGUAGE.equalsIgnoreCase(name)) {
            setLocale(element.getValue());
         } else if (FONT_SIZE.equalsIgnoreCase(name)) {
            int size = Integer.parseInt(element.getValue());
            setFontSize(size);
         } else if ("ToolTip".equalsIgnoreCase(name)) {
            setToolTipState(Boolean.parseBoolean(element.getValue()));
         } else if (WEB_BROWSER.equalsIgnoreCase(name)) {
            put(WEB_BROWSER, element.getValue());
         } else if (TEXT_EDITOR.equalsIgnoreCase(name)) {
            put(TEXT_EDITOR, element.getValue());
         } else if ("Frame.Width".equalsIgnoreCase(name)) {
            Dimension dimension = getDimension(FRAME_SIZE);
            dimension.width = Integer.parseInt(element.getValue());
            putDimension(FRAME_SIZE, dimension);
         } else if ("Frame.Height".equalsIgnoreCase(name)) {
            Dimension dimension = getDimension(FRAME_SIZE);
            dimension.height = Integer.parseInt(element.getValue());
            putDimension(FRAME_SIZE, dimension);
         } else if (FRAME_MAXIMIZED.equalsIgnoreCase(name)) {
            putBoolean("Frame.Maximized", Boolean.parseBoolean(element.getValue()));
         } else if ("SplitH1.ratio".equalsIgnoreCase(name)) {
            putInt(SPLITPANE_H1_RATIO, Integer.parseInt(element.getValue()));
         } else if ("SplitH2.ratio".equalsIgnoreCase(name)) {
            putInt(SPLITPANE_H2_RATIO, Integer.parseInt(element.getValue()));
         } else if ("SplitV1.ratio".equalsIgnoreCase(name)) {
            putInt(SPLITPANE_V1_RATIO, Integer.parseInt(element.getValue()));
         } else if ("SplitV2.ratio".equalsIgnoreCase(name)) {
            putInt(SPLITPANE_V2_RATIO, Integer.parseInt(element.getValue()));
         } else if (USE_OPERATOR_LIBRARY.equalsIgnoreCase(name)) {
            putBoolean(USE_OPERATOR_LIBRARY, Boolean.parseBoolean(element.getValue()));
         } else if (TEMPORARY_FILE_EXTENSION.equalsIgnoreCase(name)) {
            _temporaryFileExtension = element.getValue();
         } else if (UNDO_HISTORY_SIZE.equalsIgnoreCase(name)) {
            putInt(UNDO_HISTORY_SIZE, Integer.parseInt(element.getValue()));
         } else if (CODE_GENERATOR_PREFERENCE.equalsIgnoreCase(name)) {
            putInt(CODE_GENERATOR_PREFERENCE, Integer.parseInt(element.getValue()));
         } else if (PREFERRED_FILE_DIRECTORY.equalsIgnoreCase(name)) {
            element.addChildren(new XMLElement(PREFERRED_FILE_DIRECTORY, get(PREFERRED_FILE_DIRECTORY)));
            put(PREFERRED_FILE_DIRECTORY, element.getValue());
         } else if (PREFERRED_BINARY_DIRECTORY.equalsIgnoreCase(name)) {
            element.addChildren(new XMLElement(PREFERRED_BINARY_DIRECTORY, get(PREFERRED_BINARY_DIRECTORY)));
            put(PREFERRED_BINARY_DIRECTORY, element.getValue());
         }
      }
   }

   private static XMLElement convertRecentFilesToXml() throws IOException {
      XMLElement element = new XMLElement("RecentFiles");
      Workspace.getRecentProjectFilenames().stream().forEach((filename) -> {
         element.addChildren(new XMLElement("File", filename));
      });
      return element;
   }

   private static void convertXMLToRecentFiles( List<XMLElement> xmlTree ) throws IOException {
      Iterator<XMLElement> iterator = xmlTree.iterator();
      List<String> filenames = new ArrayList<>();
      while (iterator.hasNext()) {
         XMLElement element = iterator.next();
         if ("File".equalsIgnoreCase(element.getName())) {
            filenames.add(element.getValue());
         }
      }
      Workspace.setRecentProjectFilenames(filenames);
   }

   public static final class NamedColor {

      private Color _color;
      private final String _name;

      public NamedColor( Color color, String name ) {
         _color = color;
         _name = name;
      }

      public Color getColor() {
         return _color;
      }

      public String getName() {
         return _name;
      }

      public void setColor( Color color ) {
         _color = color;
      }
   }

   public static String getVisuCommandPath() {
      return getAbsoluteBinary(UserPreferences.getCommandVisu());
   }

   public static String getStatusCommandPath() {
      return getAbsoluteBinary(UserPreferences.getCommandStatus());
   }

   /**
    * Returns the absolute path of the executable if possible.
    * Possible cases are:
    * <ul>
    * <li>If the binary already contains the absolute path,
    * <li>if the binary is in the associated library.
    * </ul>
    * In all other cases, the binary is supposed to be accessible by
    * the environment variable PATH, and the function returns the binary itself.
    *
    * @param binary the abstract name of the executable.
    * @return the absolute path to a known executable or the binary itself.
    */
   public static String getAbsoluteBinary( String binary ) {
      // !!!
      // !!! Note: even if I modify the environment variable PATH, the processBuilder cannot find
      // !!! the required Shell command. It seems that it needs the absolute path !
      // !!!
      if (binary == null || binary.isEmpty()) {
         return null;
      }
      // User operator
      File unixFile;
      File windowsFile;
      String binaryDirectory = UserPreferences.getPreferredBinaryDirectory();

      if (binaryDirectory != null) {
         windowsFile = new File(binaryDirectory + File.separator + File.separator + binary + ".exe");
         if (windowsFile.exists() && windowsFile.isFile()) {
            // An operator of the library
            return windowsFile.getAbsolutePath();
         }
         unixFile = new File(binaryDirectory + File.separator + File.separator + binary);
         if (unixFile.exists() && unixFile.isFile()) {
            // An operator of the library
            return unixFile.getAbsolutePath();
         }
      }

      // Already an absolute path
      if (binary.contains(File.separator)) {
         return binary;
      }

      // Unix case
      windowsFile = new File(ArianeSettings.getLibraryPath() + File.separator + "bin" + File.separator + binary + ".exe");
      if (windowsFile.exists()) {
         // An operator of the library
         return windowsFile.getAbsolutePath();
      }
      unixFile = new File(ArianeSettings.getLibraryPath() + File.separator + "bin" + File.separator + binary);
      if (unixFile.exists()) {
         // An operator of the library
         return unixFile.getAbsolutePath();
      }
      // Default case: return the binary itself
      return binary;
   }
}
