/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.generator;

import java.io.File;
import java.io.IOException;
import java.text.DateFormat;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Deque;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeBadGraphException;
import fr.greyc.pantheon.ariane.model.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.model.graph.item.ConstantItem;
import fr.greyc.pantheon.ariane.model.graph.item.ForControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.IfControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.MergeControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.SelectControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.WhileControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.EvalItem;
import fr.greyc.pantheon.ariane.model.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.model.graph.item.FormatItem;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.item.IScriptArgument;
import fr.greyc.pantheon.ariane.model.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.item.MacroOperatorItem;
import fr.greyc.pantheon.ariane.model.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.model.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.model.graph.item.ScriptItem;
import fr.greyc.pantheon.ariane.model.graph.item.SystemItem;
import fr.greyc.pantheon.ariane.model.graph.item.ViewerItem;
import fr.greyc.pantheon.ariane.model.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.Slot;
import fr.greyc.pantheon.ariane.presenter.Presenter;

/**
 * The <code>CodeGenerator</code> is the base class for all the language generators.
 * A <code>CodeGenerator</code> generates a script file that contains code of a specified
 * language for encoding a specified graph.
 *
 * It implements the <b>Template Method</b> Design Pattern.
 */
public abstract class CodeGenerator {

   public static final String[][] AVAILABLE_LANGUAGES = {
      {"fr.greyc.pantheon.ariane.model.generator.BashGenerator", "Bash", "sh"},
      {"fr.greyc.pantheon.ariane.model.generator.PythonGenerator", "Python", "py"},
      {"fr.greyc.pantheon.ariane.model.generator.PerlGenerator", "Perl", "pl"}, //		{"fr.greyc.pantheon.ariane.model.generator.MsdosGenerator", "MsDos", "bat"}
   };
   private File _scriptFile;
   private FileOutputCode _file;

   protected CodeGenerator() {
      // empty
   }

   public static CodeGenerator createCodeGeneratorFromLanguage( File chosenFile, String language ) {
      for (int i = 0; i < AVAILABLE_LANGUAGES.length; i++) {
         if (AVAILABLE_LANGUAGES[i][1].equals(language)) {
            try {
               CodeGenerator generator = (CodeGenerator) Class.forName(AVAILABLE_LANGUAGES[i][0]).newInstance();
               generator._scriptFile = chosenFile;
               return generator;
            } catch (IllegalAccessException | InstantiationException | ClassNotFoundException e) {
               return null;
            }
         }
      }
      return null;
   }

   public static CodeGenerator createCodeGeneratorFromSuffix( File chosenFile, String suffix ) {
      for (int i = 0; i < AVAILABLE_LANGUAGES.length; i++) {
         if (AVAILABLE_LANGUAGES[i][2].equals(suffix)) {
            try {
               CodeGenerator generator = (CodeGenerator) Class.forName(AVAILABLE_LANGUAGES[i][0]).newInstance();
               generator._scriptFile = chosenFile;
               return generator;
            } catch (IllegalAccessException | InstantiationException | ClassNotFoundException e) {
               return null;
            }
         }
      }
      return null;
   }

   public final void generate( Presenter presenter, List<GraphItem> items ) {
      try {
         _file = new FileOutputCode(_scriptFile);
         generateCode(items, _scriptFile.getName());
         presenter.printSystemMessage(ArianeSettings.getString("Generator.success") + " " + _scriptFile);
      } catch (IOException | ArianeBadGraphException e) {
         presenter.printErrorMessage(ArianeSettings.getString("Generator.failure") + e.getMessage());
      } finally {
         if (_file != null) {
            _file.close();
         }
      }
   }

   private void generateCode( List<GraphItem> listOfGraphItems, String filename ) throws ArianeBadGraphException {
      List<GraphItem> orderedListOfGraphItems = buildOrderedItemsSequenceUsingTopologicalSort(listOfGraphItems);
      List<GraphItem> args = getListOfAllScriptArguments(orderedListOfGraphItems);
      List<GraphItem> inputs = getListOfAllLoaderItems(orderedListOfGraphItems);
      List<GraphItem> outputs = getListOfAllSaverItems(orderedListOfGraphItems);

      generateScriptHeader(filename, args, inputs, outputs);
      generateUtilitySubroutinesDefinition();
      generateSubroutinesDefinition(orderedListOfGraphItems);
      generateMainPrologue(orderedListOfGraphItems, args, inputs, outputs);
      generateSubroutineBody(orderedListOfGraphItems, 1, 1);
      generateMainEpilogue();
      generateScriptFooter();
   }

   protected final List<GraphItem> buildOrderedItemsSequenceUsingTopologicalSort( List<GraphItem> litems ) throws ArianeBadGraphException {
      List<GraphItem> graphItems = new ArrayList<>(litems);
      List<GraphItem> orderedItems = new ArrayList<>();
      Map<GraphItem, Integer> incomings = computeNumberOfIncomings(graphItems);
      Deque<GraphItem> stack = getAllNodesWithNoIncomings(incomings);
      // depth-first search: Do not change this order.
      while (!stack.isEmpty()) {
         GraphItem currentItem = stack.removeFirst();
         orderedItems.add(currentItem);
         graphItems.remove(currentItem);
         List<GraphItem> successors = getSuccessorsOfItem(currentItem);
         Collections.reverse(successors);
         for (GraphItem successorItem : successors) {
            int nbOfIncomings = incomings.get(successorItem);
            incomings.put(successorItem, --nbOfIncomings);
            if (nbOfIncomings == 0) {
               stack.addFirst(successorItem);
            }
         }
      }
      return orderedItems;
   }

   private List<GraphItem> getSuccessorsOfItem( GraphItem item ) {
      List<GraphItem> successors = new ArrayList<>();
      for (OutputDataSlot outputSlots : item.getOutputDataSlots()) {
         for (InputSlot inputSlot : outputSlots.getConnectedInputSlots()) {
            GraphItem nextItem = inputSlot.getGraphItem();
            successors.add(nextItem);
         }
      }
      OutputResultSlot outputResult = item.getOutputResultSlot();
      if (outputResult != null) {
         for (InputSlot inputSlot : outputResult.getConnectedInputSlots()) {
            GraphItem nextItem = inputSlot.getGraphItem();
            successors.add(nextItem);
         }
      }
      return successors;
   }

   private Map<GraphItem, Integer> computeNumberOfIncomings( List<GraphItem> graphItems ) {
      Map<GraphItem, Integer> incomings = new HashMap<>();

      for (GraphItem item : graphItems) {
         incomings.put(item, 0);
      }

      for (GraphItem item : graphItems) {
         for (GraphItem nextItem : getSuccessorsOfItem(item)) {
            if (!(nextItem instanceof VirtualItem)) {
               int numberOfIncomings = incomings.get(nextItem);
               incomings.put(nextItem, ++numberOfIncomings);
            }
         }
      }
      return incomings;
   }

   private Deque<GraphItem> getAllNodesWithNoIncomings( Map<GraphItem, Integer> incomings ) {
      Deque<GraphItem> litems = new ArrayDeque<>();

      incomings.entrySet().stream()
          .filter((entry) -> (incomings.get(entry.getKey()) == 0))
          .forEach((entry) -> litems.add(entry.getKey()));
      return litems;
   }

   private List<GraphItem> getListOfAllLoaderItems( List<GraphItem> litems ) {
      List<GraphItem> loaderItems = new ArrayList<>();
      for (GraphItem item : litems) {
         if (item instanceof LoaderItem && !item.getInputParameterSlots().get(0).isConnected()) {
            loaderItems.add(item);
         }
      }
      return loaderItems;
   }

   private List<GraphItem> getListOfAllScriptArguments( List<GraphItem> litems ) {
      List<GraphItem> items = new ArrayList<>();
      for (GraphItem item : litems) {
         if (item instanceof IScriptArgument && ((IScriptArgument) item).isArgumentOfRelatedScript()) {
            items.add(item);
         }
      }
      return items;
   }

   private List<GraphItem> getListOfAllSaverItems( List<GraphItem> litems ) {
      ArrayList<GraphItem> items = new ArrayList<>();
      for (GraphItem item : litems) {
         if (item instanceof SaverItem && !item.getInputParameterSlots().get(0).isConnected()) {
            items.add(item);
         }
      }
      return items;
   }

   protected final void generateSubroutinesDefinition( List<GraphItem> litems ) throws ArianeBadGraphException {
      for (GraphItem graphItem : litems) {
         if (graphItem instanceof WhileControlStructureItem) {
            generateSubGraph((WhileControlStructureItem) graphItem);
         } else if (graphItem instanceof ForControlStructureItem) {
            generateSubGraph((ForControlStructureItem) graphItem);
         } else if (graphItem instanceof MacroOperatorItem) {
            generateSubGraph((MacroOperatorItem) graphItem);
         }
      }
   }

   protected final void generateSubroutineBody( List<GraphItem> orderedListOfGraphItems, int graphNumber, int indentation ) throws ArianeBadGraphException {
      // 'IF' generation is more complex that is seems
      // when there exists items that are shared by operators in
      // both IF and ELSE parts.
      List<GraphItem> ifItems = new ArrayList<>();
      List<List<GraphItem>> elseItems = new ArrayList<>();
      List<GraphItem> generatedItems = new ArrayList<>();
      int indent = indentation;

      for (GraphItem graphItem : orderedListOfGraphItems) {
         generatedItems.add(graphItem);

         updateIfItems(graphNumber, ifItems, elseItems, indent, graphItem);

         if (graphItem instanceof LoaderItem) {
            generateItem((LoaderItem) graphItem, graphNumber, indent);
         } else if (graphItem instanceof SaverItem) {
            generateItem((SaverItem) graphItem, graphNumber, indent);
         } else if (graphItem instanceof ViewerItem) {
            generateItem((ViewerItem) graphItem, graphNumber, indent);
         } else if (graphItem instanceof EvalItem) {
            generateItem((EvalItem) graphItem, graphNumber, indent);
         } else if (graphItem instanceof FileNameItem) {
            generateItem((FileNameItem) graphItem, graphNumber, indent);
         } else if (graphItem instanceof FormatItem) {
            generateItem((FormatItem) graphItem, graphNumber, indent);
         } else if (graphItem instanceof ConstantItem) {
            generateItem((ConstantItem) graphItem, graphNumber, indent);
         } else if (graphItem instanceof ScriptItem) {
            generateItem((ScriptItem) graphItem, graphNumber, indent);
         } else if (graphItem instanceof SystemItem) {
            generateItem((SystemItem) graphItem, graphNumber, indent);
         } else if (graphItem instanceof OperatorItem) {
            generateItem((OperatorItem) graphItem, graphNumber, indent);
         } else if (graphItem instanceof SelectControlStructureItem) {
            generateItem((SelectControlStructureItem) graphItem, graphNumber, indent);
         } else if (graphItem instanceof IfControlStructureItem) {
            ifItems.add(0, graphItem);
            generateItem((IfControlStructureItem) graphItem, graphNumber, indent);
            generateThenPart((IfControlStructureItem) graphItem, graphNumber, indent++);
            elseItems.add(0, searchForFirstItemsOfTheElsePart(graphItem, generatedItems));
         } else if (graphItem instanceof MergeControlStructureItem) {
            if (!ifItems.isEmpty()) {
               generateElsePart((IfControlStructureItem) ifItems.get(0), graphNumber, indent - 1);
               elseItems.remove(0);
               ifItems.remove(0);
            }
            generateFiPart(null, graphNumber, --indent);
            generateItem((MergeControlStructureItem) graphItem, graphNumber, indent);
         } else if (graphItem instanceof MacroOperatorItem) {
            generateItem((MacroOperatorItem) graphItem, graphNumber, indent);
         } else if (graphItem instanceof WhileControlStructureItem) {
            generateItem((CompositeItem) graphItem, graphNumber, indent);
         } else if (graphItem instanceof ForControlStructureItem) {
            generateItem((CompositeItem) graphItem, graphNumber, indent);
         }
      }
      for (int i = 0; i < ifItems.size(); i++) {
         generateFiPart(null, graphNumber, indent);
      }
   }

   private void updateIfItems( int graphNumber, List<GraphItem> ifItems, List<List<GraphItem>> elseItems, int indent, GraphItem graphItem ) throws ArianeBadGraphException {
      if (!ifItems.isEmpty() && !elseItems.isEmpty()) {
         List<GraphItem> items = elseItems.get(0);
         if (items.contains(graphItem)) {
            generateElsePart((IfControlStructureItem) ifItems.get(0), graphNumber, indent - 1);
            ifItems.remove(0);
         }
      }
   }

   private List<GraphItem> searchForFirstItemsOfTheElsePart( GraphItem ifItem, List<GraphItem> generatedItems ) {
      OutputDataSlot elseSlot = ifItem.getOutputDataSlots().get(1);
      List<InputSlot> connectedSlots = elseSlot.getConnectedInputSlots();

      // add items that are directly connected to the else slot of the if item (if not merge).
      ArrayList<GraphItem> graphItems = new ArrayList<>();
      for (InputSlot slot : connectedSlots) {
         if (!(slot.getGraphItem() instanceof MergeControlStructureItem)) {
            graphItems.add(slot.getGraphItem());
         }
      }

      // add items that compute the parameters of connected items
      // and that are not already generated.
      ArrayList<GraphItem> graphItemsBis = new ArrayList<>();
      for (GraphItem item : graphItems) {
         for (InputParameterSlot parameter : item.getInputParameterSlots()) {
            if (parameter.isConnected()) {
               GraphItem connectedItem = parameter.getConnectedSlot().getGraphItem();
               if (!generatedItems.contains(connectedItem)) {
                  graphItemsBis.add(connectedItem);
               }
            }
         }
      }

      // add operators that computes the input images
      // that are not already generated.
      for (GraphItem item : graphItems) {
         for (InputDataSlot slot : item.getInputDataSlots()) {
            GraphItem connectedItem = slot.getConnectedSlot().getGraphItem();
            if (slot.isConnected() && !generatedItems.contains(connectedItem)) {
               graphItemsBis.add(connectedItem);
            }
         }
      }
      graphItems.addAll(graphItemsBis);
      return graphItems;
   }

   protected abstract void generateScriptHeader( String filename, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs );

   protected abstract void generateScriptFooter();

   protected abstract void generateUtilitySubroutinesDefinition();

   protected abstract void generateMainPrologue( List<GraphItem> items, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs );

   protected abstract void generateMainEpilogue();

   protected abstract void generateItem( LoaderItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateItem( SaverItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateItem( ViewerItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateItem( OperatorItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateItem( EvalItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateItem( FileNameItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateItem( FormatItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateItem( ConstantItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateItem( ScriptItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateItem( SystemItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateItem( SelectControlStructureItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateThenPart( IfControlStructureItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateElsePart( IfControlStructureItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateFiPart( IfControlStructureItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateItem( IfControlStructureItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateItem( MergeControlStructureItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateItem( CompositeItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

   protected abstract void generateSubGraph( MacroOperatorItem item ) throws ArianeBadGraphException;

   protected abstract void generateSubGraph( ForControlStructureItem item ) throws ArianeBadGraphException;

   protected abstract void generateSubGraph( WhileControlStructureItem item ) throws ArianeBadGraphException;

   protected final String getCurrentDate() {
      final Date creationDate = new Date();
      final DateFormat dateFormat = DateFormat.getDateInstance(DateFormat.MEDIUM);
      return dateFormat.format(creationDate);
   }

   protected final GraphItem getVirtualOperator( List<GraphItem> litems ) {
      for (GraphItem graphItem : litems) {
         if (graphItem instanceof VirtualItem) {
            return graphItem;
         }
      }
      return null;
   }

   protected final String createInputImageName( InputSlot slot, int graphId ) {
      return UserPreferences.createTemporaryName(graphId, slot.getConnectedSlot().getGraphItem().getId(), slot.getConnectedSlot().getRank());
   }

   protected final boolean hasMergeItemInItsSuccessors( GraphItem item ) {
      for (OutputSlot slot : item.getOutputDataSlots()) {
         if (slot.isConnected()) {
            for (Slot connectedSlot : slot.getConnectedSlots()) {
               if (connectedSlot.getGraphItem() instanceof MergeControlStructureItem) {
                  return true;
               }
            }
         }
      }
      return false;
   }

   protected final void println( String text ) {
      _file.println(text);
   }

   protected final void println( String text, int numberOfWhiteSpaces ) {
      _file.println(text, numberOfWhiteSpaces);
   }

   protected final void print( String text, int numberOfWhiteSpaces ) {
      _file.print(text, numberOfWhiteSpaces);
   }

   protected final void print( String text ) {
      _file.print(text);
   }
}
