/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.generator;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeBadGraphException;
import fr.greyc.pantheon.ariane.model.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.model.graph.item.ConstantItem;
import fr.greyc.pantheon.ariane.model.graph.item.ForControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.IfControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.MergeControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.SelectControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.EvalItem;
import fr.greyc.pantheon.ariane.model.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.model.graph.item.FormatItem;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.item.MacroOperatorItem;
import fr.greyc.pantheon.ariane.model.graph.item.NoteItem;
import fr.greyc.pantheon.ariane.model.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.model.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.model.graph.item.ScriptItem;
import fr.greyc.pantheon.ariane.model.graph.item.SystemItem;
import fr.greyc.pantheon.ariane.model.graph.item.ViewerItem;
import fr.greyc.pantheon.ariane.model.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.model.graph.item.WhileControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.presenter.SubGraph;

/**
 * The <code>PerlGenerator</code> generates Perl code for the specified graph.
 * It operates in three steps:<ol>
 * <li>transforms the graph into a list of operators usable for the topological sort;
 * <li>executes the topological sort;
 * <li>generates the related code.
 * </ol>
 */
public final class PerlGenerator extends CodeGenerator {

   private final List<String> _outputImages = new ArrayList<>();

   protected PerlGenerator() {
      // Only created by the code generator.
   }

   @Override
   protected void generateScriptHeader( String filename, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs ) {
      println("#!/usr/bin/env perl");
      println("#");
      println("# @(#)" + filename + "				" + getCurrentDate());
      println("#");
      println("# This file was automatically generated by Ariane.");
      println("# Ariane is part of the Pantheon Project.\n");

      println("use Getopt::Std;");
      println("use File::Copy;");
      println("use Math::Trig;\n");
      println("#\n# Displays the program usage\n#");
      println("sub displayUsage {");
      print("print \"Usage: $0 [-vh]", 1);
      for (int i = 0; i < arguments.size(); i++) {
         print(" " + arguments.get(i).getNickName());
      }
      for (int i = 0; i < inputs.size(); i++) {
         print(" input-file" + i);
      }
      for (int i = 0; i < outputs.size(); i++) {
         print(" output-file" + i);
      }
      println("\\n\\nOptions:\\n\";");
      println("print \"  -h display this usage\\n\";", 1);
      println("print \"  -v verbose\\n\";", 1);
      print("print \"Example:\\n  $0", 1);
      for (int i = 0; i < arguments.size(); i++) {
         if (arguments.get(i).getFormula() != null && !arguments.get(i).getFormula().isEmpty()) {
            print(" " + arguments.get(i).getFormula());
         } else {
            print(" " + arguments.get(i).getNickName());
         }
      }
      for (int i = 0; i < inputs.size(); i++) {
         print(" " + inputs.get(i).getNickName());
      }
      for (int i = 0; i < outputs.size(); i++) {
         String name = outputs.get(i).getNickName();
         if ("saver".equals(name)) {
            name += outputs.get(i).getId();
         }
         print(" " + name);
      }

      println("\\n\";");
      println("exit;\n}", 1);

      println("\n#\n# Reads the command line\n#");
      println("getopts(\"hv\");");
      println("$verbose = 0;");
      println("if ($opt_h) { displayUsage }");
      // Verbose mode.
      println("if ($opt_v) { $verbose = 1; }");
      println("if ($#ARGV + 1 != " + (arguments.size() + inputs.size() + outputs.size()) + ") {");
      println("print \"Bad argument numbers\\n\\n\";", 1);
      println("displayUsage;", 1);
      println("}\nmain();");
   }

   @Override
   protected void generateScriptFooter() {
      // Do nothing
   }

   @Override
   protected void generateUtilitySubroutinesDefinition() {
      println("\n#\n# Executes an operator and return the status.\n#");
      println("sub execute_with_status {");
      println("my ($operator) = @_;", 1);
      println("print \"$operator\\n\" if $verbose;", 1);
      println("system($operator);", 1);
      println("$result = `" + UserPreferences.getCommandStatus() + "`;", 1);
      println("chomp($result);", 1);
      println("return $result;", 1);
      println("}\n");
      println("\n#\n# Executes an operator.\n#");
      println("sub execute {");
      println("my ($operator) = @_;", 1);
      println("print \"$operator\\n\" if $verbose;", 1);
      println("system($operator);", 1);
      println("}\n");
      println("\n# Some math complements.\n#");
      println("sub max {");
      println("my ($a, $b) = @_;", 1);
      println("print \"$operator\\n\" if $verbose;", 1);
      println("return ($a > $b) ? $a : $b;", 1);
      println("}\n");
      println("sub min {");
      println("my ($a, $b) = @_;", 1);
      println("print \"$operator\\n\" if $verbose;", 1);
      println("return ($a < $b) ? $a : $b;", 1);
      println("}\n");
      println("sub log10 {");
      println("my $n = shift;", 1);
      println("return log($n) / log(10);", 1);
      println("}\n");
      println("sub log2 {");
      println("my $n = shift;", 1);
      println("return log($n) / log(10);", 1);
      println("}\n");
   }

   @Override
   protected void generateMainPrologue( List<GraphItem> litems, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs ) {
      println("\n### The main program");
      for (GraphItem item : litems) {
         if (item instanceof NoteItem) {
            println("# " + convertCommentIntoPerl(item.getFormula()), 0);
         }
      }
      println("sub main {");
      println("unlink(<" + UserPreferences.getTemporaryFilePrefix() + "*" + UserPreferences.getTemporaryFileExtension() + ">);", 1);
      for (GraphItem item : arguments) {
         println(createVariableName(1, item.getId(), 0) + "= shift(@ARGV);", 1);
      }
      for (GraphItem item : inputs) {
         println(createVariableName(1, item.getId(), 0) + " = $ARGV[0]; "
             + "symlink(shift(@ARGV), \"" + createOutputImageName(1, item.getId(), 0) + "\");", 1);
      }
   }

   @Override
   protected void generateMainEpilogue() {
      println("}");
   }

   @Override
   protected void generateItem( LoaderItem item, int graphId, int indent ) throws ArianeBadGraphException {
      if (item.getInputParameterSlots().get(0).isConnected()) {
         String parameters = getActualParameterValue(item, 0, graphId);
         String imagename = createOutputImageName(graphId, item.getId(), 0);
         println("copy(" + parameters + ", \"" + imagename + "\");", indent);
         println(createVariableName(graphId, item.getId(), 0) + " = \"" + imagename + "\";", indent);
      }
   }

   @Override
   protected void generateItem( SaverItem item, int graphId, int indent ) throws ArianeBadGraphException {
      if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
         if (item.getInputParameterSlots().get(0).isConnected()) {
            String parameters = getActualParameterValue(item, 0, graphId);
            println("copy(\"" + createInputImageName(item.getInputDataSlots().get(0), graphId)
                + "\", " + parameters + ");", indent);
         } else {
            println("copy(\"" + createInputImageName(item.getInputDataSlots().get(0), graphId)
                + "\", shift(@ARGV));", indent);
         }
         String imagename = createOutputImageName(graphId, item.getId(), 0);
         println("symlink(\""
             + createInputImageName(item.getInputDataSlots().get(0), graphId)
             + "\", \""
             + imagename + "\");", indent);
         println(createVariableName(graphId, item.getId(), 0) + " = \"" + imagename + "\";", indent);
      }
   }

   @Override
   protected void generateItem( ViewerItem item, int graphId, int indent ) throws ArianeBadGraphException {
      if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
         String commandShell = item.getViewerCommandName();
         println("system(\"" + commandShell + " " + createInputImageName(item.getInputDataSlots().get(0), graphId) + "\");", indent);
         println("symlink (\""
             + createInputImageName(item.getInputDataSlots().get(0), graphId)
             + "\", \""
             + createOutputImageName(graphId, item.getId(), 0)
             + "\");", indent);
      }
   }

   @Override
   protected void generateItem( OperatorItem item, int graphId, int indent ) throws ArianeBadGraphException {
      String codeName = OperatorItem.getCommandPart(item.getBinary());
      if (codeName == null) {
         throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error2") + item.getName());
      }

      // Get input images.
      List<String> inputLine = new LinkedList<>();
      for (InputDataSlot image : item.getInputDataSlots()) {
         if (image.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
         }
         inputLine.add(createInputImageName(image, graphId));
      }

      // Get parameters
      List<String> parameterLine = new LinkedList<>();
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         parameterLine.add(getActualParameterValue(item, i, graphId));
      }

      // Get mask
      List<String> maskLine = new LinkedList<>();
      InputMaskSlot mask = item.getInputMaskSlot();
      if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, graphId))) {
         maskLine.add(createInputImageName(mask, graphId));
      }

      // Get output images.
      List<String> outputLine = new LinkedList<>();
      for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
         outputLine.add(createOutputImageName(graphId, item.getId(), i));
      }

      String code = null;
      List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
      for (String command : commandLine) {
         if (code == null) {
            code = command;
         } else {
            code += " " + command;
         }
      }
      if (item.getOutputResultSlot().isConnected() || hasMergeItemInItsSuccessors(item)) {
         println(createVariableName(graphId, item.getId(), 0) + " = &execute_with_status(\"" + code + "\");", indent);
      } else {
         println("&execute(\"" + code + "\");", indent);
      }
   }

   @Override
   protected void generateItem( SystemItem item, int graphId, int indent ) throws ArianeBadGraphException {
      String codeName = OperatorItem.getCommandPart(item.getBinary());
      if (codeName == null || codeName.trim().isEmpty()) {
         throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error3") + item.getName() + "(" + item.getId() + ")");
      }

      // Get parameters
      List<String> parameterLine = new LinkedList<>();
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         parameterLine.add(getActualParameterValue(item, i, graphId));
      }

      // Get mask
      List<String> maskLine = new LinkedList<>();
      InputMaskSlot mask = item.getInputMaskSlot();
      if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, graphId))) {
         maskLine.add(createInputImageName(mask, graphId));
      }

      // Get input images.
      List<String> inputLine = new LinkedList<>();
      for (InputDataSlot image : item.getInputDataSlots()) {
         if (image.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
         }
         inputLine.add(createInputImageName(image, graphId));
      }

      // Get output images.
      List<String> outputLine = new LinkedList<>();
      for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
         outputLine.add(createOutputImageName(graphId, item.getId(), i));
      }

      List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
      String code = null;
      for (String command : commandLine) {
         if (code == null) {
            code = "\"" + command + " ";
         } else if (command.startsWith("\"$result")) {
            code += "\\\"\" . " + command.substring(1, command.length() - 1) + " . \"\\\" ";
         } else if (command.startsWith("$result")) {
            code += "\" . " + command + " . \" ";
         } else if (command.startsWith("\"")) {
            code += command + " ";
         } else if (command.startsWith("\'")) {
            code += " " + command + " ";
         } else {
            code += command.replaceAll("\"", "\\\\\"") + " ";
         }
      }

      if (item.getOutputResultSlot().isConnected() || hasMergeItemInItsSuccessors(item)) {
         println(createVariableName(graphId, item.getId(), 0) + " = " + "&execute_with_status(" + code + "\");", indent);
      } else {
         println("&execute(" + code + "\");", indent);
      }
   }

   @Override
   protected void generateItem( EvalItem item, int graphId, int indent ) throws ArianeBadGraphException {
      String[] parameters = new String[item.getInputParameterSlots().size()];
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         parameters[i] = getActualParameterValue(item, i, graphId);
      }
      println(createVariableName(graphId, item.getId(), 0)
          + " = "
          + convertArithmeticFormulaIntoPerl(item.getFormula(), parameters)
          + ";", indent);
   }

   @Override
   protected void generateItem( FormatItem item, int graphId, int indent ) throws ArianeBadGraphException {
      // Get parameters
      String[] parameters = new String[item.getInputParameterSlots().size()];
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         parameters[i] = getActualParameterValue(item, i, graphId);
      }

      String code = "";
      StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), " $() +'\"", true);
      String etape;
      while (stringTokenizer.hasMoreTokens()) {
         etape = stringTokenizer.nextToken();
         if ("+".equals(etape)) {
            code += " . ";
         } else if ("quote".equals(etape)) {
            code += "";
         } else if ("\"".equals(etape)) {
            code += "\"";
         } else if ("(".equals(etape)) {
            code += "\"'\" . ";
         } else if (")".equals(etape)) {
            code += " . \"'\"";
         } else if ("$".equals(etape)) {
            int numberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
            code += parameters[numberOfParameters];
         } else {
            code += etape;
         }
      }
      println(createVariableName(graphId, item.getId(), 0) + " = " + code + ";", indent);
   }

   @Override
   protected void generateItem( SelectControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
      println("# Select", indent);

      String value = getActualParameterValue(item, 0, graphId);
      for (int i = 0; i < item.getInputDataSlots().size(); i++) {
         if (!item.getInputDataSlots().get(i).isConnected()) {
            throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
         }
         println("if (" + value + "==" + i + ") {", indent);
         println("symlink(\""
             + createInputImageName(item.getInputDataSlots().get(i), graphId) + "\", \""
             + createOutputImageName(graphId, item.getId(), 0) + "\");", indent + 1);
         println(createVariableName(graphId, item.getId(), 0) + " = " + i + ";", indent + 1);
         println("}", indent);
      }
   }

   @Override
   protected void generateItem( IfControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
      String[] parameters = new String[2];
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         parameters[i] = getActualParameterValue(item, i, graphId);
      }

      // Parse the boolean expression.
      StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), "&|$^!0123456789.E><=() +*-/", true);
      String etape;
      String commandShell = "";
      while (stringTokenizer.hasMoreTokens()) {
         etape = stringTokenizer.nextToken();
         if ("&".equals(etape)) {
            commandShell += " && ";
         } else if ("|".equals(etape)) {
            commandShell += " || ";
         } else if ("$".equals(etape)) {
            if (stringTokenizer.nextToken().charAt(0) == '0') {
               commandShell += parameters[0];
            } else {
               commandShell += parameters[1];
            }
         } else {
            commandShell += etape;
         }
      }

      println("if (" + commandShell + ") {", indent);
   }

   @Override
   protected void generateThenPart( IfControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
      println("symlink(\""
          + createInputImageName(item.getInputDataSlots().get(0), graphId) + "\", \""
          + createOutputImageName(graphId, item.getId(), 0) + "\");", indent + 1);
      println(createVariableName(graphId, item.getId(), 0) + " = 0;", indent + 1);
   }

   @Override
   protected void generateElsePart( IfControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
      println("} else {", indent);
      println("symlink(\""
          + createInputImageName(item.getInputDataSlots().get(0), graphId) + "\", \""
          + createOutputImageName(graphId, item.getId(), 1) + "\");", indent + 1);
      println(createVariableName(graphId, item.getId(), 0) + "= 1;", indent + 1);
   }

   @Override
   protected void generateFiPart( IfControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
      println("}", indent);
   }

   @Override
   protected void generateItem( MergeControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
      InputDataSlot in0 = item.getInputDataSlots().get(0);
      int id0 = in0.getConnectedSlot().getGraphItem().getId();

      print("if (defined(" + createVariableName(graphId, id0, 0) + ") && ", indent);
      print(createVariableName(graphId, id0, 0) + " ne \"FAILURE\" && ");
      println("-e \"" + createInputImageName(item.getInputDataSlots().get(0), graphId) + "\"){");
      println("symlink(\""
          + createInputImageName(item.getInputDataSlots().get(0), graphId)
          + "\", \""
          + createOutputImageName(graphId, item.getId(), 0) + "\");", indent + 1);
      println(createVariableName(graphId, item.getId(), 0) + " = 0;", indent + 1);
      print("} else {");
      println("symlink(\""
          + createInputImageName(item.getInputDataSlots().get(1), graphId) + "\", \""
          + createOutputImageName(graphId, item.getId(), 0) + "\");", indent + 1);
      println(createVariableName(graphId, item.getId(), 0) + " = 1;", indent + 1);
      println("}", indent);
   }

   @Override
   protected void generateItem( FileNameItem item, int graphId, int indent ) throws ArianeBadGraphException {
      if (!item.isArgumentOfRelatedScript()) {
         println(createVariableName(graphId, item.getId(), 0) + " = \"" + item.getFormula() + "\";", indent);
      }
   }

   @Override
   protected void generateItem( ConstantItem item, int graphId, int indent ) {
      if (!item.isArgumentOfRelatedScript()) {
         String value = item.getFormula().replaceAll("\"", "\\\\\"");
         println(createVariableName(graphId, item.getId(), 0) + " = \"" + value + "\";", indent);
      }
   }

   @Override
   protected void generateItem( ScriptItem item, int graphId, int indent ) throws ArianeBadGraphException {
      String value = item.getFormula().replaceAll("\"", "\\\\\\\\\"");
      println(createVariableName(graphId, item.getId(), 0) + " = \"" + value + "\";", indent);
   }

   @Override
   protected void generateItem( CompositeItem item, int graphId, int indent ) throws ArianeBadGraphException {
      String code = "&macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "(";
      boolean comma = false;

      // Get parameters
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         if (comma) {
            code += ", ";
         } else {
            comma = true;
         }
         code += "\"" + getActualParameterValue(item, i, graphId) + "\"";
      }

      // Get mask
      InputMaskSlot mask = item.getInputMaskSlot();
      if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, graphId))) {
         if (comma) {
            code += ", ";
         } else {
            comma = true;
         }
         code += "\"" + createInputImageName(mask, graphId) + "\"";
      }

      // Get input images.
      for (InputDataSlot image : item.getInputDataSlots()) {
         if (image.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
         }
         if (comma) {
            code += ", ";
         } else {
            comma = true;
         }
         code += "\"" + createInputImageName(image, graphId) + "\"";
      }

      // Get output images.
      for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
         if (comma) {
            code += ", ";
         } else {
            comma = true;
         }
         code += "\"" + createOutputImageName(graphId, item.getId(), i) + "\"";
      }
      // Get output result
      if (item.getOutputResultSlot().isConnected()) {
         println(createVariableName(graphId, item.getId(), 0) + " = " + code + ");", indent);
      } else {
         println(code + ");", indent);
      }
   }

   @Override
   protected void generateSubGraph( ForControlStructureItem item ) throws ArianeBadGraphException {
      SubGraph subGraph = item.getSubGraph();

      generateComment(item, subGraph.getListOfGraphItems());

      println("sub macro" + item.getId() + "_" + item.getGraph().getIdentifier() + " {");

      // Get parameters
      int subgraphId = subGraph.getIdentifier();
      println("my $result" + subgraphId + ";", 1);
      println("$begin=shift(@_);", 1);
      println("$end=shift(@_);", 1);
      println("$step=shift(@_);", 1);
      for (int i = 3; i < item.getInputParameterSlots().size(); i++) {
         println(createVariableName(subgraphId, 0, i - 2) + " = shift(@_);\n", 1);
      }
      println(createVariableName(subgraphId, 0, 0) + " = $begin;", 1);

      // Get mask
      InputMaskSlot mask = item.getInputMaskSlot();
      if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, item.getGraph().getIdentifier()))) {
         int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
         println("symlink(shift(@_), \"" + createOutputImageName(subgraphId, 0, maskIndex) + "\");", 1);
      }

      // Get input images
      int rank = 0;
      for (InputDataSlot isi : item.getInputDataSlots()) {
         // Unlinked input!
         if (isi.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
         }
         println("copy(shift(@_), \"" + createOutputImageName(subgraphId, 0, rank++) + "\");", 1);
      }

      // Create the output images in case of no iteration
      VirtualItem virtualItem = subGraph.getVirtualItem();
      for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
         InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
         if (isi.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
         }
         println("$f=open(\"" + createInputImageName(isi, subgraphId) + "\"); close($f);", 1);
      }

      // Initialize the return variable in case of no iteration
      {
         InputParameterSlot isp = (InputParameterSlot) virtualItem.getVirtualOutputResultSlot();
         println(createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + " = \"SUCCESS\";", 1);
      }

      println("");

      println("for ($i=$begin ; $i <= $end ; $i += $step) {", 1);
      println(createVariableName(subgraphId, 0, 0) + " = $i;", 2);

      List<GraphItem> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());
      generateSubroutineBody(litems, subgraphId, 2);

      // Get intermediate images
      for (int i = 0; i < item.getInputDataSlots().size(); i++) {
         InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
         // Unlinked input!
         if (isi.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
         }
         String input = createInputImageName(isi, subgraphId);
         String output = createOutputImageName(subgraphId, 0, i);
         if (!input.equals(output)) {
            println("copy(\"" + input + "\",\"" + output + "\");", 2);
         }
      }

      println("}", 1);
      // Get output images
      println("");

      for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
         InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
         println("symlink(\"" + createInputImageName(isi, subgraphId) + "\",shift(@_));", 1);
      }

      // Output result
      InputParameterSlot isp = (InputParameterSlot) virtualItem.getVirtualOutputResultSlot();
      if (isp.getConnectedSlot() == null) {
         println("return 0;", 1);
      } else {
         println("return " + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + ";", 1);
      }
      println("}");

      generateSubroutinesDefinition(litems);
   }

   @Override
   protected void generateSubGraph( WhileControlStructureItem item ) throws ArianeBadGraphException {
      SubGraph subGraph = item.getSubGraph();

      generateComment(item, subGraph.getListOfGraphItems());

      println("sub macro" + item.getId() + "_" + item.getGraph().getIdentifier() + " {");

      // Get parameters
      String[] parameters = new String[item.getInputParameterSlots().size()];
      int subgraphId = subGraph.getIdentifier();
      println("my $result" + subgraphId + ";", 1);
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         println(createVariableName(subgraphId, 0, i) + " = shift(@_);", 1);
         parameters[i] = createVariableName(subgraphId, 0, i);
      }

      // Get mask
      InputMaskSlot mask = item.getInputMaskSlot();
      if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, item.getGraph().getIdentifier()))) {
         int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
         println("symlink(shift(@_), \"" + createOutputImageName(subgraphId, 0, maskIndex) + "\");", 1);
      }

      // Get input images
      int rank = 0;
      for (InputDataSlot isi : item.getInputDataSlots()) {
         // Unlinked input!
         if (isi.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
         }
         println("copy(shift(@_), \"" + createOutputImageName(subgraphId, 0, rank++) + "\");", 1);
      }

      // Parse the boolean expression.
      StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), "&|$^!0123456789.E><=() +*-/", true);
      String etape;
      String testFormula = "";
      while (stringTokenizer.hasMoreTokens()) {
         etape = stringTokenizer.nextToken();
         if ("&".equals(etape)) {
            testFormula += " && ";
         } else if ("|".equals(etape)) {
            testFormula += " || ";
         } else if ("$".equals(etape)) {
            if (stringTokenizer.nextToken().charAt(0) == '0') {
               testFormula += parameters[0];
            } else {
               testFormula += parameters[1];
            }
         } else {
            testFormula += etape;
         }
      }

      println("");

      println("while (" + testFormula + ") {", 1);

      List<GraphItem> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());
      generateSubroutineBody(litems, subgraphId, 2);

      // Get intermediate images
      VirtualItem virtualItem = subGraph.getVirtualItem();
      for (int i = 0; i < item.getInputDataSlots().size(); i++) {
         InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
         // Unlinked input!
         if (isi.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
         }
         String input = createInputImageName(isi, subgraphId);
         String output = createOutputImageName(subgraphId, 0, i);
         if (!input.equals(output)) {
            println("copy(\"" + input + "\",\"" + output + "\");", 2);
         }
      }

      // Get intermediate parameters
      for (int i = 0; i < 2; i++) {
         InputParameterSlot isp = virtualItem.getVirtualIntermediateResultSlot(i);
         if (isp.getConnectedSlot() != null) {
            println(parameters[i] + "=" + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + ";", 2);
         }
      }
      println("}", 1);

      // Get output images
      for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
         InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
         println("symlink(\"" + createInputImageName(isi, subgraphId) + "\",shift(@_));", 1);
      }

      // Output result
      InputParameterSlot isp = (InputParameterSlot) virtualItem.getVirtualOutputResultSlot();
      if (isp.getConnectedSlot() == null) {
         println("return 0;", 1);
      } else {
         println("return " + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + ";", 1);
      }
      println("}");

      generateSubroutinesDefinition(litems);
   }

   @Override
   protected void generateSubGraph( MacroOperatorItem item ) throws ArianeBadGraphException {
      SubGraph subGraph = item.getSubGraph();
      List<GraphItem> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());

      generateComment(item, subGraph.getListOfGraphItems());

      println("sub macro" + item.getId() + "_" + item.getGraph().getIdentifier() + " {");

      // Get parameters.
      int subgraphId = subGraph.getIdentifier();
      println("my $result" + subgraphId + ";", 1);
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         println(createVariableName(subgraphId, 0, i) + " = shift(@_);", 1);
      }

      // Get mask.
      InputMaskSlot mask = item.getInputMaskSlot();
      if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, item.getGraph().getIdentifier()))) {
         int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
         println("symlink(shift(@_), \"" + createOutputImageName(subgraphId, 0, maskIndex) + "\");", 1);
      }

      // Get input images.
      int rank = 0;
      for (InputDataSlot isi : item.getInputDataSlots()) {
         // Unlinked input!
         if (isi.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
         }
         println("symlink(shift(@_), \"" + createOutputImageName(subgraphId, 0, rank++) + "\");", 1);
      }

      println("");

      generateSubroutineBody(litems, subgraphId, 1);

      println("");

      // Get output images.
      VirtualItem virtualItem = subGraph.getVirtualItem();
      for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
         InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
         println("symlink(\"" + createInputImageName(isi, subgraphId) + "\",shift(@_));", 1);
      }
      // Output result.
      InputParameterSlot isp = (InputParameterSlot) virtualItem.getVirtualOutputResultSlot();
      if (isp.getConnectedSlot() == null) {
         println("return 0;", 1);
      } else {
         println("return " + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + ";", 1);
      }
      println("}");

      generateSubroutinesDefinition(litems);
   }

   private void generateComment( GraphItem item, List<GraphItem> litems ) {
      println("\n#\n# " + item.getNickName());
      for (GraphItem note : litems) {
         if (note instanceof NoteItem) {
            println("# " + convertCommentIntoPerl(note.getFormula()));
         }
      }
      println("#");
   }

   private String createOutputImageName( int graphId, int itemId, int slotRank ) {
      String name = UserPreferences.createTemporaryName(graphId, itemId, slotRank);
      _outputImages.add(name);
      return name;
   }

   private String createVariableName( int graphId, int itemId, int slotRank ) {
      return "$result" + graphId + "[" + itemId + "][" + slotRank + "]";
   }

   private String getActualParameterValue( GraphItem item, int slotRank, int graphId ) {
      InputParameterSlot isp = (InputParameterSlot) item.getSlot("InSlotParameter", slotRank);
      if (isp.getConnectedSlot() == null) {
         return ((InputParameterSlot) item.getSlot("InSlotParameter", slotRank)).getValue();
      } else {
         return createVariableName(graphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank());
      }
   }

   private String convertCommentIntoPerl( String comment ) {
      if (comment != null) {
         return comment.replaceAll("\n", "\n# ");
      }
      return "";
   }

   private String convertArithmeticFormulaIntoPerl( String formula, String[] parameters ) {
      // Parse the formula.
      StringTokenizer stringTokenizer = new StringTokenizer(formula, "$ -+ ()*/^0123456789.E", true);
      String etape;
      String code = "";
      while (stringTokenizer.hasMoreTokens()) {
         etape = stringTokenizer.nextToken();
         if ("^".equals(etape)) {
            code += "**";
         } else if ("e".equals(etape) || "E".equals(etape)) {
            code += "*10^";
         } else if ("ln".equals(etape)) {
            code += "log";
         } else if ("atan".equals(etape)) {
            code += "a";
         } else if ("$".equals(etape)) {
            int nnumberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
            code += parameters[nnumberOfParameters];
         } else {
            code += etape;
         }
      }
      return code;
   }
}
