/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.generator;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeBadGraphException;
import fr.greyc.pantheon.ariane.model.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.model.graph.item.ConstantItem;
import fr.greyc.pantheon.ariane.model.graph.item.ForControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.IfControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.MergeControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.SelectControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.EvalItem;
import fr.greyc.pantheon.ariane.model.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.model.graph.item.FormatItem;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.item.MacroOperatorItem;
import fr.greyc.pantheon.ariane.model.graph.item.NoteItem;
import fr.greyc.pantheon.ariane.model.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.model.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.model.graph.item.ScriptItem;
import fr.greyc.pantheon.ariane.model.graph.item.SystemItem;
import fr.greyc.pantheon.ariane.model.graph.item.ViewerItem;
import fr.greyc.pantheon.ariane.model.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.model.graph.item.WhileControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.presenter.SubGraph;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * The <code>PerlGenerator</code> generates Python code for the specified graph.
 * It operates in three steps:<ol>
 * <li>transforms the graph into a list of operators usable for the topological sort;
 * <li>executes the topological sort;
 * <li>generates the related code.
 * </ol>
 */
public final class PythonGenerator extends CodeGenerator {

   /**
    * The list of generated output images.
    */
   private final List<String> _outputImages = new ArrayList<>();

   protected PythonGenerator() {
      // Only created by the code generator.
   }

   @Override
   protected void generateScriptHeader( String filename, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs ) {
      println("# -*- coding: utf-8 -*-");
      println("#");
      println("# @(#)" + filename + "				" + getCurrentDate());
      println("#");
      println("# This file was automatically generated by Ariane.");
      println("# Ariane is part of the Pantheon Project.\n");

      println("import sys, getopt");
      println("import subprocess");
      println("import os, errno");
      println("import os.path");
      println("import shutil");
      println("import glob");
      println("from math import *\n");

      println("verbose = False\n", 0);
      println("def printUsage( scriptName ):", 0);
      print("print(\"Usage: \" + scriptName + \" [-hv]", 1);
      for (int i = 0; i < arguments.size(); i++) {
         print(" " + arguments.get(i).getNickName());
      }
      for (int i = 0; i < inputs.size(); i++) {
         print(" input-file" + i);
      }
      for (int i = 0; i < outputs.size(); i++) {
         print(" output-file" + i);
      }
      println("\")");

      println("print(\"  -h print this usage\")", 1);
      println("print(\"  -v verbose\")", 1);
      print("print(\"  example: \" + scriptName + \"", 1);
      for (int i = 0; i < arguments.size(); i++) {
         if (arguments.get(i).getFormula() != null && !arguments.get(i).getFormula().isEmpty()) {
            print(" " + arguments.get(i).getFormula());
         } else {
            print(" " + arguments.get(i).getNickName());
         }
      }
      for (int i = 0; i < inputs.size(); i++) {
         print(" " + inputs.get(i).getNickName());
      }
      for (int i = 0; i < outputs.size(); i++) {
         String name = outputs.get(i).getNickName();
         if ("saver".equals(name)) {
            name += outputs.get(i).getId();
         }
         print(" " + name);
      }
      println("\")");
      println("sys.exit(1)\n", 1);
   }

   @Override
   protected void generateScriptFooter() {
      // Do nothing
   }

   @Override
   protected void generateUtilitySubroutinesDefinition() {
      println("def execute( *command ):");
      println("''' Executes an operator. '''", 1);
      println("if (verbose): print(command)", 1);
      println("subprocess.Popen(command, stdout = subprocess.PIPE).communicate()[0]", 1);

      println("\ndef executeWithStatus( *command ):");
      println("''' Executes an operator and return the status. '''", 1);
      println("if (verbose): print(command)", 1);
      println("subprocess.Popen(command, stdout = subprocess.PIPE).communicate()[0]", 1);
      println("return subprocess.Popen(\"" + UserPreferences.getCommandStatus() + "\", stdout = subprocess.PIPE).communicate()[0]", 1);

      println("\ndef link( src, dest ):");
      println("''' Creates a symbolic link between the two files. '''", 1);
      println("try:", 1);
      println("os.symlink(src, dest)", 2);
      println("except OSError as e:", 1);
      println("if e.errno == errno.EEXIST:", 2);
      println("os.remove(dest)", 3);
      println("os.symlink(src, dest)", 3);
      println("", 1);
   }

   @Override
   protected void generateMainPrologue( List<GraphItem> litems, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs ) {
      println("\n### The main program");
      println("def main( args ):");

      // Comments
      println("''' Main ", 1);
      for (GraphItem item : litems) {
         if (item instanceof NoteItem) {
            println(convertCommentIntoPython(item.getFormula()), 1);
         }
      }
      println("'''", 1);

      println("if (len(args) != " + (arguments.size() + inputs.size() + outputs.size()) + " ):", 1);
      println("printUsage(sys.argv[0])", 2);

      println("fileList = glob.glob(\"" + UserPreferences.getTemporaryFilePrefix() + "*" + UserPreferences.getTemporaryFileExtension() + "\")", 1);
      println("for file in fileList: os.remove(file)", 1);
      println("argc = 0", 1);
      for (GraphItem item : arguments) {
         println(createVariableName(1, item.getId(), 0) + " = args[argc]; argc += 1", 1);
      }
      for (GraphItem item : inputs) {
         println("link(args[argc], \"" + createOutputImageName(1, item.getId(), 0) + "\"); "
             + createVariableName(1, item.getId(), 0)
             + " = args[argc]; argc += 1", 1);
      }
   }

   @Override
   protected void generateMainEpilogue() {
      println("\nif __name__ == \"__main__\":", 0);
      println("try:", 1);
      println("opts, args = getopt.getopt(sys.argv[1:], \"hv\")", 2);
      println("except getopt.GetoptError:", 1);
      println("printUsage(sys.argv[0])", 2);
      println("for opt, arg in opts:", 1);
      println("if opt == '-h':", 2);
      println("printUsage(sys.argv[0])", 3);
      println("if opt == '-v':", 2);
      println("verbose = True", 3);
      println("main(args)", 1);
   }

   @Override
   protected void generateItem( LoaderItem item, int graphId, int indent ) throws ArianeBadGraphException {
      if (item.getInputParameterSlots().get(0).isConnected()) {
         String parameter = getActualParameterValue(item, 0, graphId);
         String imagename = createOutputImageName(graphId, item.getId(), 0);
         println("shutil.copy(" + parameter + ", " + "\"" + imagename + "\")", indent);
         println(createVariableName(graphId, item.getId(), 1) + " = \"" + imagename + "\"", indent);
      }
   }

   @Override
   protected void generateItem( SaverItem item, int graphId, int indent ) throws ArianeBadGraphException {
      if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
         if (item.getInputParameterSlots().get(0).isConnected()) {
            String parameters = getActualParameterValue(item, 0, graphId);
            println("shutil.copy(\"" + createInputImageName(item.getInputDataSlots().get(0), graphId)
                + "\", " + parameters + ")", indent);
         } else {
            println("shutil.copy(\"" + createInputImageName(item.getInputDataSlots().get(0), graphId)
                + "\", args[argc]); argc += 1", indent);
         }
         String imagename = createOutputImageName(graphId, item.getId(), 0);
         println("link(\""
             + createInputImageName(item.getInputDataSlots().get(0), graphId)
             + "\", \"" + imagename + "\")", indent);
         println(createVariableName(graphId, item.getId(), 0) + " = \"" + imagename + "\"", indent);
      }
   }

   @Override
   protected void generateItem( ViewerItem item, int graphId, int indent ) {
      if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
         final String commandShell = item.getViewerCommandName();
         println("subprocess.call([\"" + commandShell + "\", \"" + createInputImageName(item.getInputDataSlots().get(0), graphId) + "\"])", indent);
         println("link(\""
             + createInputImageName(item.getInputDataSlots().get(0), graphId)
             + "\", \""
             + createOutputImageName(graphId, item.getId(), 0)
             + "\")", indent);
      }
   }

   @Override
   protected void generateItem( OperatorItem item, int graphId, int indent ) throws ArianeBadGraphException {
      String codeName = OperatorItem.getCommandPart(item.getBinary());
      if (codeName == null) {
         throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error2") + item.getName());
      }

      // Get input images.
      List<String> inputLine = new LinkedList<>();
      for (InputDataSlot image : item.getInputDataSlots()) {
         if (image.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
         }
         inputLine.add("\"" + createInputImageName(image, graphId) + "\"");
      }

      // Get parameters
      List<String> parameterLine = new LinkedList<>();
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         parameterLine.add(getActualParameterValue(item, i, graphId));
      }

      // Get mask
      List<String> maskLine = new LinkedList<>();
      InputMaskSlot mask = item.getInputMaskSlot();
      if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, graphId))) {
         maskLine.add("\"" + createInputImageName(mask, graphId) + "\"");
      }

      // Get output images.
      List<String> outputLine = new LinkedList<>();
      List<OutputDataSlot> outputDataSlots = item.getOutputDataSlots();
      for (int i = 0; i < outputDataSlots.size(); i++) {
         outputLine.add("\"" + createOutputImageName(graphId, item.getId(), i) + "\"");
      }

      String code = null;
      List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
      for (String command : commandLine) {
         if ("-m".equals(command)) {
            code += ", \"" + command + "\"";
         } else if (code == null) {
            code = "\"" + command + "\"";
         } else {
            code += ", " + command;
         }
      }
      if (item.getOutputResultSlot().isConnected() || hasMergeItemInItsSuccessors(item)) {
         println(createVariableName(graphId, item.getId(), 0) + " = " + "executeWithStatus(" + code + ")", indent);
      } else {
         println("execute(" + code + ")", indent);
      }
   }

   @Override
   protected void generateItem( SystemItem item, int graphId, int indent ) throws ArianeBadGraphException {
      String codeName = OperatorItem.getCommandPart(item.getBinary());
      if (codeName == null || codeName.trim().isEmpty()) {
         throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error3") + item.getName() + "(" + item.getId() + ")");
      }

      // Get parameters
      List<String> parameterLine = new LinkedList<>();
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         parameterLine.add(getActualParameterValue(item, i, graphId));
      }

      // Get mask
      List<String> maskLine = new LinkedList<>();
      InputMaskSlot mask = item.getInputMaskSlot();
      if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, graphId))) {
         maskLine.add(createInputImageName(mask, graphId));
      }

      // Get input images.
      List<String> inputLine = new LinkedList<>();
      for (InputDataSlot image : item.getInputDataSlots()) {
         if (image.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
         }
         inputLine.add("\"" + createInputImageName(image, graphId) + "\"");
      }

      // Get output images.
      List<OutputDataSlot> outputDataSlots = item.getOutputDataSlots();
      List<String> outputLine = new LinkedList<>();
      for (int i = 0; i < outputDataSlots.size(); i++) {
         outputLine.add("\"" + createOutputImageName(graphId, item.getId(), i) + "\"");
      }

      List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
      String code = null;
      for (String command : commandLine) {
         if (code == null) {
            code = "\"" + command + " ";
         } else if (command.startsWith("\"var")) {
            code += "\\\"\" + " + command.substring(1, command.length() - 1) + " + \"\\\" ";
         } else if (command.startsWith("var")) {
            code += "\" + " + command + " + \" ";
         } else if (command.startsWith("\'")) {
            code += " " + command + " ";
         } else {
            code += command.replaceAll("\"", "\\\\\"") + " ";
         }
      }

      println(createVariableName(graphId, item.getId(), 0) + " = " + "os.system(" + code + "\")", indent);
   }

   @Override
   protected void generateItem( EvalItem item, int graphId, int indent ) throws ArianeBadGraphException {
      String[] parameters = new String[item.getInputParameterSlots().size()];
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         parameters[i] = getActualParameterValue(item, i, graphId);
      }
      println(createVariableName(graphId, item.getId(), 0)
          + " = str("
          + convertArithmeticFormulaIntoPython(item.getFormula(), parameters)
          + ")", indent);
   }

   @Override
   protected void generateItem( FormatItem item, int graphId, int indent ) throws ArianeBadGraphException {
      // Get parameters
      String[] parameters = new String[item.getInputParameterSlots().size()];
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         parameters[i] = getActualParameterValue(item, i, graphId);
      }

      String code = "";
      StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), " $()+'\"", true);
      String etape;
      while (stringTokenizer.hasMoreTokens()) {
         etape = stringTokenizer.nextToken();
         if ("+".equals(etape)) {
            code += " + ";
         } else if ("quote".equals(etape)) {
            code += "";
         } else if ("\"".equals(etape)) {
            code += "\"";
         } else if ("(".equals(etape)) {
            code += "\"'\"+";
         } else if (")".equals(etape)) {
            code += "+\"'\"";
         } else if ("$".equals(etape)) {
            int numberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
            code += parameters[numberOfParameters];
         } else {
            code += etape;
         }
      }
      println(createVariableName(graphId, item.getId(), 0) + " = " + code, indent);
   }

   @Override
   protected void generateItem( SelectControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
      println("# Select", indent);

      final String value = getActualParameterValue(item, 0, graphId);
      for (int i = 0; i < item.getInputDataSlots().size(); i++) {
         if (!item.getInputDataSlots().get(i).isConnected()) {
            throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
         }
         println("if (int(" + value + ") == " + i + "):", indent);
         println("link(\""
             + createInputImageName(item.getInputDataSlots().get(i), graphId) + "\", \""
             + createOutputImageName(graphId, item.getId(), 0) + "\")", indent + 1);
         println(createVariableName(graphId, item.getId(), 0) + " = \"" + i + "\"", indent + 1);
      }
   }

   @Override
   protected void generateItem( IfControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
      String[] parameters = new String[2];
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         parameters[i] = getActualParameterValue(item, i, graphId);
      }

      // Parse the boolean expression.
      StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), "&|$^!0123456789.E><=() +*-/", true);
      String code = "";
      while (stringTokenizer.hasMoreTokens()) {
         String token = stringTokenizer.nextToken();
         if ("&".equals(token)) {
            code += " and ";
         } else if ("|".equals(token)) {
            code += " or ";
         } else if ("$".equals(token)) {
            String token2 = stringTokenizer.nextToken();
            if (token2.charAt(0) == '0') {
               code += "float(" + parameters[0] + ")";
            } else {
               code += "float(" + parameters[1] + ")";
            }
         } else {
            code += token;
         }
      }
      println("if " + code + ":", indent);
   }

   @Override
   protected void generateThenPart( IfControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
      println("link(\""
          + createInputImageName(item.getInputDataSlots().get(0), graphId)
          + "\", \""
          + createOutputImageName(graphId, item.getId(), 0) + "\")", indent + 1);
      println(createVariableName(graphId, item.getId(), 0) + " = 0", indent + 1);
   }

   @Override
   protected void generateElsePart( IfControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
      println("else:", indent);
      println("link(\""
          + createInputImageName(item.getInputDataSlots().get(0), graphId)
          + "\", \""
          + createOutputImageName(graphId, item.getId(), 1) + "\")", indent + 1);
      println(createVariableName(graphId, item.getId(), 0) + " = 1", indent + 1);
   }

   @Override
   protected void generateFiPart( IfControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
      // Nothing to do
   }

   @Override
   protected void generateItem( MergeControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
      InputDataSlot in0 = item.getInputDataSlots().get(0);

      int id0 = in0.getConnectedSlot().getGraphItem().getId();

      println("# Merge", indent);
      println("try: " + createVariableName(graphId, id0, 0) + "; input = 0", indent);
      println("except UnboundLocalError: input = 1", indent);
      print("if input == 0 and ", indent);
      println("os.path.isfile(\"" + createInputImageName(item.getInputDataSlots().get(0), graphId) + "\"):");
      println("link(\""
          + createInputImageName(item.getInputDataSlots().get(0), graphId)
          + "\", \""
          + createOutputImageName(graphId, item.getId(), 0) + "\")", indent + 1);
      println(createVariableName(graphId, item.getId(), 0) + " = 0", indent + 1);
      println("else:", indent);
      println("link(\""
          + createInputImageName(item.getInputDataSlots().get(1), graphId) + "\", \""
          + createOutputImageName(graphId, item.getId(), 0) + "\")", indent + 1);
      println(createVariableName(graphId, item.getId(), 0) + " = 1", indent + 1);
   }

   @Override
   protected void generateItem( FileNameItem item, int graphId, int indent ) throws ArianeBadGraphException {
      if (!item.isArgumentOfRelatedScript()) {
         println(createVariableName(graphId, item.getId(), 0) + " = '" + item.getFormula() + "'", indent);
      }
   }

   @Override
   protected void generateItem( ConstantItem item, int graphId, int indent ) {
      if (!item.isArgumentOfRelatedScript()) {
         String value = item.getFormula().replaceAll("\"", "\\\\\"");
         println(createVariableName(graphId, item.getId(), 0) + " = \"" + value + "\"", indent);
      }
   }

   @Override
   protected void generateItem( ScriptItem item, int graphId, int indent ) throws ArianeBadGraphException {
      String value = item.getFormula().replaceAll("\"", "\\\\\"");
      println(createVariableName(graphId, item.getId(), 0) + " = \"\"\"" + value + "\"\"\"", indent);
   }

   @Override
   protected void generateItem( CompositeItem item, int graphId, int indent ) throws ArianeBadGraphException {
      String code = "macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "(";
      boolean comma = false;

      // Get parameters
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         if (comma) {
            code += ", ";
         } else {
            comma = true;
         }
         code += getActualParameterValue(item, i, graphId);
      }

      // Get mask
      InputMaskSlot mask = item.getInputMaskSlot();
      if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, graphId))) {
         if (comma) {
            code += ", ";
         } else {
            comma = true;
         }
         code += "\"" + createInputImageName(mask, graphId) + "\"";
      }

      // Get input images.
      for (InputDataSlot image : item.getInputDataSlots()) {
         if (image.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
         }
         if (comma) {
            code += ", ";
         } else {
            comma = true;
         }
         code += "\"" + createInputImageName(image, graphId) + "\"";
      }

      // Get output images.
      Collection<OutputDataSlot> outputDataSlots = item.getOutputDataSlots();
      for (int i = 0; i < outputDataSlots.size(); i++) {
         if (comma) {
            code += ", ";
         } else {
            comma = true;
         }
         code += "\"" + createOutputImageName(graphId, item.getId(), i) + "\"";
      }
      // Get output result.
      if (item.getOutputResultSlot().isConnected()) {
         println(createVariableName(graphId, item.getId(), 0) + " = " + code + ")", indent);
      } else {
         println(code + ")", indent);
      }
   }

   @Override
   protected void generateSubGraph( ForControlStructureItem item ) throws ArianeBadGraphException {
      SubGraph subGraph = item.getSubGraph();

      println("def macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "( *args ):");

      generateComment(item, subGraph.getListOfGraphItems());

      println("argc = 0", 1);

      // Get parameters
      int subgraphId = subGraph.getIdentifier();
      println("begin = int(args[argc]); argc += 1", 1);
      println("end = int(args[argc]); argc += 1", 1);
      println("step = int(args[argc]); argc += 1", 1);
      for (int i = 3; i < item.getInputParameterSlots().size(); i++) {
         println(createVariableName(subgraphId, 0, i - 2) + " = args[argc]; argc += 1", 1);
      }
      println(createVariableName(subgraphId, 0, 0) + " = str(begin)", 1);

      // Get mask.
      InputMaskSlot mask = item.getInputMaskSlot();
      if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, item.getGraph().getIdentifier()))) {
         int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
         println("link(args[argc], \"" + createOutputImageName(subgraphId, 0, maskIndex) + "\"); argc += 1", 1);
      }

      // Get input files
      int rank = 0;
      for (InputDataSlot isi : item.getInputDataSlots()) {
         // Unlinked input!
         if (isi.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
         }
         println("shutil.copy(args[argc], \"" + createOutputImageName(subgraphId, 0, rank++) + "\"); argc += 1", 1);
      }

      // Create the output images in case of no iteration
      VirtualItem virtualItem = subGraph.getVirtualItem();
      for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
         InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
         if (isi.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
         }
         println("if not os.path.exists(\"" + createInputImageName(isi, subgraphId) + "\"): open( \"" + createInputImageName(isi, subgraphId) + "\", 'w').close()", 1);
      }

      // Initialize the return variable in case of no iteration
      {
         InputParameterSlot isp = (InputParameterSlot) virtualItem.getVirtualOutputResultSlot();
         println(createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "=\"SUCCESS\"", 1);
      }

      println("");
      println("for i in range(begin, end+1, step):", 1);
      println(createVariableName(subgraphId, 0, 0) + " = str(i)", 2);

      List<GraphItem> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());
      generateSubroutineBody(litems, subgraphId, 2);

      // Get intermediate images
      for (int i = 0; i < item.getInputDataSlots().size(); i++) {
         InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
         // Unlinked input!
         if (isi.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
         }
         String input = createInputImageName(isi, subgraphId);
         String output = createOutputImageName(subgraphId, 0, i);
         if (!input.equals(output)) {
            println("shutil.copy(\"" + input + "\",\"" + output + "\")", 2);
         }
      }

      // Get output images.
      println("");

      for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
         InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
         println("link(\"" + createInputImageName(isi, subgraphId) + "\",args[argc]); argc += 1", 1);
      }

      // Output result
      InputParameterSlot isp = (InputParameterSlot) virtualItem.getVirtualOutputResultSlot();
      if (isp.getConnectedSlot() == null) {
         println("return 0", 1);
      } else {
         println("return " + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()), 1);
      }
      println("");

      generateSubroutinesDefinition(litems);
   }

   @Override
   protected void generateSubGraph( WhileControlStructureItem item ) throws ArianeBadGraphException {
      SubGraph subGraph = item.getSubGraph();

      println("def macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "( *args ):");

      generateComment(item, subGraph.getListOfGraphItems());

      println("argc = 0", 1);

      // Get parameters
      int subgraphId = subGraph.getIdentifier();
      String[] parameters = new String[item.getInputParameterSlots().size()];
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         println(createVariableName(subgraphId, 0, i) + " = args[argc]; argc += 1", 1);
         parameters[i] = createVariableName(subgraphId, 0, i);
      }

      // Get mask.
      InputMaskSlot mask = item.getInputMaskSlot();
      if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, item.getGraph().getIdentifier()))) {
         int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
         println("link(args[argc], \"" + createOutputImageName(subgraphId, 0, maskIndex) + "\"); argc += 1", 1);
      }

      // Get input files
      int rank = 0;
      for (InputDataSlot isi : item.getInputDataSlots()) {
         // Unlinked input!
         if (isi.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
         }
         println("shutil.copy(args[argc], \"" + createOutputImageName(subgraphId, 0, rank++) + "\"); argc += 1", 1);
      }

      // Parse the boolean expression.
      StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), "&|$^!0123456789.E><=() +*-/", true);
      String etape;
      String testFormula = "";
      while (stringTokenizer.hasMoreTokens()) {
         etape = stringTokenizer.nextToken();
         if ("&".equals(etape)) {
            testFormula += " and ";
         } else if ("|".equals(etape)) {
            testFormula += " or ";
         } else if ("$".equals(etape)) {
            if (stringTokenizer.nextToken().charAt(0) == '0') {
               testFormula += "float(" + parameters[0] + ")";
            } else {
               testFormula += "float(" + parameters[1] + ")";
            }
         } else {
            testFormula += etape;
         }
      }

      println("");

      println("while (" + testFormula + "):", 1);

      List<GraphItem> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());
      generateSubroutineBody(litems, subgraphId, 2);

      // Get intermediate images
      VirtualItem virtualItem = subGraph.getVirtualItem();
      for (int i = 0; i < item.getInputDataSlots().size(); i++) {
         InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
         // Unlinked input!
         if (isi.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
         }
         String input = createInputImageName(isi, subgraphId);
         String output = createOutputImageName(subgraphId, 0, i);
         if (!input.equals(output)) {
            println("shutil.copy(\"" + input + "\",\"" + output + "\")", 2);
         }
      }

      // Get intermediate parameters
      for (int i = 0; i < 2; i++) {
         InputParameterSlot isp = virtualItem.getVirtualIntermediateResultSlot(i);
         if (isp.getConnectedSlot() != null) {
            println(parameters[i] + " = " + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()), 2);
         }
      }

      // Get output images.
      println("");

      for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
         InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
         println("link(\"" + createInputImageName(isi, subgraphId) + "\",args[argc]); argc += 1", 1);
      }

      // Output result
      InputParameterSlot isp = (InputParameterSlot) virtualItem.getVirtualOutputResultSlot();
      if (isp.getConnectedSlot() == null) {
         println("return 0", 1);
      } else {
         println("return " + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()), 1);
      }
      println("");

      generateSubroutinesDefinition(litems);
   }

   @Override
   protected void generateSubGraph( MacroOperatorItem item ) throws ArianeBadGraphException {
      SubGraph subGraph = item.getSubGraph();

      println("def macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "( *args ):");

      generateComment(item, subGraph.getListOfGraphItems());

      // Get parameters.
      println("argc = 0", 1);
      int subgraphId = subGraph.getIdentifier();
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         println(createVariableName(subgraphId, 0, i) + " = args[argc]; argc += 1", 1);
      }

      // Get mask.
      InputMaskSlot mask = item.getInputMaskSlot();
      if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, item.getGraph().getIdentifier()))) {
         int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
         println("link(args[argc], \"" + createOutputImageName(subgraphId, 0, maskIndex) + "\"); argc += 1", 1);
      }

      // Get input images.
      int rank = 0;
      for (InputDataSlot isi : item.getInputDataSlots()) {
         // Unlinked input!
         if (isi.getConnectedSlot() == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
         }
         println("link(args[argc], \"" + createOutputImageName(subgraphId, 0, rank++) + "\"); argc += 1", 1);
      }

      List<GraphItem> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());
      generateSubroutineBody(litems, subgraphId, 1);

      println("");

      // Get output images.
      VirtualItem virtualItem = subGraph.getVirtualItem();
      for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
         InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
         println("link(\"" + createInputImageName(isi, subgraphId) + "\",args[argc]); argc += 1", 1);
      }

      // Output result.
      InputParameterSlot isp = (InputParameterSlot) virtualItem.getVirtualOutputResultSlot();
      if (isp.getConnectedSlot() == null) {
         println("return 0", 1);
      } else {
         println("return " + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()), 1);
      }
      println("");

      generateSubroutinesDefinition(litems);
   }

   private void generateComment( GraphItem item, List<GraphItem> litems ) {
      println("''' " + item.getNickName() + "\n", 1);
      for (GraphItem note : litems) {
         if (note instanceof NoteItem) {
            println(note.getFormula(), 1);
         }
      }
      println("'''", 1);
   }

   private String createOutputImageName( int graphId, int itemId, int slotRank ) {
      String name = UserPreferences.createTemporaryName(graphId, itemId, slotRank);
      _outputImages.add(name);
      return name;
   }

   private String createVariableName( int graphId, int itemId, int slotRank ) {
      return "var" + graphId + "_" + itemId + "_" + slotRank;
   }

   private String getActualParameterValue( GraphItem item, int slotRank, int graphId ) {
      InputParameterSlot isp = (InputParameterSlot) item.getSlot("InSlotParameter", slotRank);
      if (isp.getConnectedSlot() == null) {
         return "\"" + ((InputParameterSlot) item.getSlot("InSlotParameter", slotRank)).getValue() + "\"";
      } else {
         return createVariableName(graphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank());
      }
   }

   private String convertCommentIntoPython( String comment ) {
      if (comment != null) {
         return comment.replaceAll("\n", "\n   ");
      }
      return "";
   }

   private String convertArithmeticFormulaIntoPython( String formula, String[] parameters ) {
      StringTokenizer stringTokenizer = new StringTokenizer(formula, "$ -+()*/^0123456789.E", true);
      String etape;
      String code = "";
      while (stringTokenizer.hasMoreTokens()) {
         etape = stringTokenizer.nextToken();
         if ("^".equals(etape)) {
            code += "**";
         } else if ("ln".equals(etape)) {
            code += "log";
         } else if ("atan".equals(etape)) {
            code += "a";
         } else if ("$".equals(etape)) {
            int nnumberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
            code += "float(" + parameters[nnumberOfParameters] + ")";
         } else {
            code += etape;
         }
      }
      return code;
   }
}
