/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph;

import java.io.File;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

import fr.greyc.pantheon.ariane.model.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.slot.Slot;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.model.xml.XMLFileOutputStream;
import fr.greyc.pantheon.ariane.presenter.Graph;

/**
 * A graph represents an executable program as an oriented graph of items.
 */
public final class GraphModel {

   private final int _identifier;

   private static int _numberOfCreatedGraphs;
   private int _numberOfCreatedItemsInThisGraph;

   private final List<GraphItem> _graphItems = new LinkedList<>();
   private File _storageFile;

   public GraphModel() {
      _identifier = ++_numberOfCreatedGraphs;
   }

   public GraphModel( File absoluteFile ) {
      this();
      _storageFile = absoluteFile;
   }

   public int getIdentifier() {
      return _identifier;
   }

   public int getNextItemIdentifier() {
      return ++_numberOfCreatedItemsInThisGraph;
   }

   public void dispose() {
      for (GraphItem item : _graphItems) {
         item.dispose();
      }
      _graphItems.clear();
   }

   public List<GraphItem> getListOfGraphItems() {
      return _graphItems;
   }

   public File getStorageFile() {
      return _storageFile;
   }

   public void add( GraphItem item ) {
      _graphItems.add(item);
   }

   public void remove( GraphItem item ) {
      _graphItems.remove(item);
   }

   public void resetAll() {
      for (GraphItem item : _graphItems) {
         item.resetAll();
      }
   }

   public void reset() {
      for (GraphItem item : _graphItems) {
         item.reset();
      }
   }

   public GraphItem getGraphItem( int identifier ) {
      for (GraphItem item : _graphItems) {
         if (item.isInSameReadPasteSession() && item.getInternalId() == identifier) {
            return item;
         }
      }
      return null;
   }

   public void save( File file, Graph graph ) throws IOException {
      XMLFileOutputStream out = new XMLFileOutputStream(file, null);
      out.write(convertToXMLTree(file, graph));
      out.close();
      _storageFile = file;
   }

   public XMLElement convertToXMLTree( File file, Graph graph ) throws IOException {
      XMLElement element;
      if (graph.isRoot()) {
         element = new XMLElement("Graph");
      } else {
         element = new XMLElement("SubGraph");
      }
      XMLElement items = new XMLElement("ItemList");
      for (GraphItem item : _graphItems) {
         items.addChildren(item.convertToXML(file));
      }
      XMLElement connections = new XMLElement("ConnectionList");
      for (GraphItem item : _graphItems) {
         connections.addChildren(item.convertSlotsToXML());
      }
      element.addChildren(items);
      element.addChildren(connections);
      return element;
   }

   public void createFromXMLTree( List<XMLElement> xmlTree, Graph graph ) throws IOException {
      List<XMLElement> xmlSubTree = null;
      boolean found = false;
      if (graph.isRoot()) {
         xmlSubTree = xmlTree.get(0).getChildren();
         String tagName = xmlSubTree.get(0).getName();
         if ("Graph".equalsIgnoreCase(tagName) || "Graphe".equalsIgnoreCase(tagName)) {
            xmlSubTree = xmlSubTree.get(0).getChildren();
            found = true;
         }
      } else {
         for (XMLElement element : xmlTree) {
            String tagName = element.getName();
            if ("SubGraph".equalsIgnoreCase(tagName) || "SousGraphe".equalsIgnoreCase(tagName)) {
               xmlSubTree = element.getChildren();
               found = true;
               break;
            }
         }
      }

      if (!found) {
         throw new IOException("Incorrect XML file.");
      }
      for (XMLElement element : xmlSubTree) {
         String tagName = element.getName();
         if ("ItemList".equalsIgnoreCase(tagName) || "ListeElements".equalsIgnoreCase(tagName)) {
            GraphItem.createFromXML(graph, element.getChildren());
         } else if ("ConnectionList".equalsIgnoreCase(tagName) || "ListeConnexions".equalsIgnoreCase(tagName)) {
            Slot.createFromXML(graph, element.getChildren());
         }
      }
   }

   public void synchronizeItemFilesWithSystemFiles() {
      for (GraphItem item : _graphItems) {
         if (item instanceof LoaderItem) {
            ((LoaderItem) item).synchronizeWithFileSystem();
         } else if (item instanceof FileNameItem) {
            ((FileNameItem) item).synchronizeWithFileSystem();
         }
      }
   }
}
