/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph;

import java.io.File;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import javax.swing.SwingUtilities;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.graph.item.DataFile;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.item.SaverItem;

/**
 * A <code>ThreadDirectoryExecution</code> controls the execution
 * of a graph for a whole directory.
 */
public final class ThreadDirectoryExecution extends Thread {

   private GraphItem[] _itemToExecute;
   private final LoaderItem _loader;
   private final File[] _files;
   private final Map<SaverItem, File> _saverDirectories;
   private final File _loaderFile;

   public ThreadDirectoryExecution( List<GraphItem> selectedItems, LoaderItem loader, File loaderFile, File[] files, Map<SaverItem, File> saverDirectories ) {
      // We use a copy of the selectItems since user can modify the list
      // while the program is executing it.
      _itemToExecute = selectedItems.toArray(new GraphItem[selectedItems.size()]);
      _loader = loader;
      _files = Arrays.copyOf(files, files.length);
      _loaderFile = loaderFile;
      _saverDirectories = saverDirectories;
   }

   @Override
   public void run() {
      Ariane.getPresenter().setExecution(true);
      for (File file : _files) {
         final String basename = getBasename(file.getName());
         _loader.setBinary(file.getAbsolutePath());
         updateItemName(file.getName());
         for (SaverItem item : _saverDirectories.keySet()) {
            String suffix = getSuffix(_saverDirectories.get(item).getName());
            item.setBinary(_saverDirectories.get(item).getParent() + File.separator + basename + suffix);
         }
         _loader.resetAll();
         for (GraphItem item : _itemToExecute) {
            item.execute();
         }
      }
      _itemToExecute = null;
      Ariane.getPresenter().setExecution(false);

      // Restore old file name to the loader and the savers (modified the execution).
      _loader.setBinary(_loaderFile.getAbsolutePath());

      updateItemName(DataFile.getFileNameWithoutExtension(_loaderFile.getAbsolutePath()));

      for (SaverItem item : _saverDirectories.keySet()) {
         item.setBinary(_saverDirectories.get(item).getAbsolutePath());
      }
      _loader.resetAll();
   }

   private String getSuffix( String filename ) {
      int suffixPosition1 = filename.lastIndexOf('.');
      if (suffixPosition1 > 0) {
         return filename.substring(suffixPosition1);
      } else {
         return "";
      }
   }

   private String getBasename( String filename ) {
      int suffixPosition = filename.lastIndexOf('.');
      if (suffixPosition > 0) {
         return filename.substring(0, suffixPosition);
      } else {
         return filename;
      }
   }

   private void updateItemName( final String name ) {
      SwingUtilities.invokeLater(new Runnable() {
         @Override
         public void run() {
            _loader.setTemporaryNickName(DataFile.getFileNameWithoutExtension(name));
            _loader.getGraph().revalidate();
         }
      });
   }
}
