/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeArithmeticException;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingFormulaException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.item.etc.ArithmeticExpression;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.item.DefaultControlItemView;

import java.io.IOException;
import java.util.List;

/**
 * A <code>EvalItem</code> implements a calculator item.
 */
public final class EvalItem extends ControlItem {

   public static final String TAGNAME = "Eval";

   public EvalItem( TreeItem treeItem, Graph graph ) throws InterruptedException {
      super(new DefaultControlItemView(), treeItem, graph, treeItem.getName());
      askProperties(null, 0, 2, 0);
   }

   public EvalItem( Graph graph, List<XMLElement> elements ) throws IOException {
      super(new DefaultControlItemView(), graph);
      createFromXML(elements);
      String expression;
      if (getFormula() == null && getBinary() != null) { // Keep this for sake of compatibility
         expression = getBinary();
         setBinary(null);
      } else {
         expression = getFormula();
      }
      try {
         checkExpression(expression);
         setFormula(expression);
         setComment(expression);
      } catch (ArianeArithmeticException e) {
         Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
         setFormula("");
      }
   }

   protected EvalItem( TreeItem treeItem, Graph graph, String name ) { // For test
      super(new DefaultControlItemView(), treeItem, graph, name);
      createSlots(treeItem.getNumberOfInputDataSlots(), treeItem.getNumberOfParameterSlots(), 0, treeItem.getNumberOfOutputDataSlots(), 1);
   }

   @Override
   public void setPrototype( String name, int iIm, int iP, int oIm ) {
      setAbstractName(name);
      setFormula(""); // IMPERATIVE to signify that it uses formula.
      createSlots(0, iP, 0, 0, 0);
   }

   ///// 
   ///// A C C E S S O R S
   /////
   /**
    * @return the tag name.
    */
   @Override
   public String getTagName() {
      return TAGNAME;
   }

   ///// 
   ///// E X E C U T I O N
   /////
   @Override
   protected boolean primitiveExecute() throws ArianeRuntimeException {
      try {
         if (getFormula().isEmpty()) {
            throw new ArianeMissingFormulaException(ArianeSettings.getString("arithmetic.error0"));
         }
         int numberOfParameters = getInputParameterSlots().size();
         double[] val = new double[numberOfParameters];
         for (int i = 0; i < numberOfParameters; i++) {
            InputParameterSlot parameter = getInputParameterSlots().get(i);
            if (parameter.hasNumericalValue(true)) {
               val[i] = parameter.getNumericalValue();
            } else {
               return false;
            }
         }
         ArithmeticExpression expression = getCompiledExpression();
         final String result = Double.toString(expression.evaluate(val));
         getOutputResultSlot().setValue(result);
         Ariane.getPresenter().printSystemMessage("eval(" + getFormula() + ")=" + result);
         changeStatus(Status.EXECUTED);
         return true;
      } catch (ArianeException e) {
         if (Graph.isExecutionRunning()) {
            Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
            getGraph().stopExecution();
            changeStatus(Status.ERROR);
         }
         throw new ArianeRuntimeException(e);
      }
   }

   @Override
   public boolean changeFormula( String arithmeticExpression ) {
      if (getFormula() == null || !getFormula().equals(arithmeticExpression)) {
         try {
            checkExpression(arithmeticExpression);
            setFormula(arithmeticExpression);
            setComment(arithmeticExpression);
         } catch (ArianeArithmeticException e) {
            Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
            return false;
         }
      }
      return true;
   }

   private ArithmeticExpression getCompiledExpression() {
      try {
         return new ArithmeticExpression(getFormula(), getInputParameterSlots().size());
      } catch (ArianeArithmeticException e) {
         return null;
      }
   }

   private void checkExpression( String textFormula ) throws ArianeArithmeticException {
      new ArithmeticExpression(textFormula, getInputParameterSlots().size());
   }
}
