/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.item.FileNameItemView;

import java.io.File;
import java.io.IOException;
import java.util.List;

/**
 * A <code>FileNameItem</code> implements a filename item other than an image file.
 * Such file is used as a parameter of an other item.
 */
public final class FileNameItem extends ControlItem implements IScriptArgument {

   public static final String TAGNAME = "Filename";

   private boolean _isArgument;
   private long _lastModified;

   public FileNameItem( TreeItem treeItem, Graph graph ) {
      super(new FileNameItemView(), treeItem, graph, treeItem.getName());
      super.setFormula(""); // IMPERATIVE to signify that it uses formula.
      setComment(getAbstractFileName());
   }

   public FileNameItem( Graph graph, List<XMLElement> elements ) throws IOException {
      super(new FileNameItemView(), graph);
      super.setFormula(""); // IMPERATIVE to signify that it uses formula.
      createFromXML(elements);
      setComment(getAbstractFileName());
   }

   @Override
   public boolean isArgumentOfRelatedScript() {
      return _isArgument;
   }

   @Override
   public void setArgumentOfRelatedScript( boolean value ) {
      _isArgument = value;
   }

   @Override
   public void setPrototype( String name, int iIm, int iP, int oIm ) {
      // Empty
   }

   @Override
   public void setFormula( String formula ) {
      if (!formula.trim().isEmpty()) {
         String currentDirectory = getGraph().getDirectoryName();
         super.setFormula(DataFile.setPathToFilename(formula, currentDirectory));
      }
   }

   @Override
   public boolean changeFormula( String textFormula ) {
      if (getFormula().equals(textFormula)) {
         return true;
      }
      setFormula(textFormula);
      setComment(getAbstractFileName());

      return true;
   }

   @Override
   public String getFormula( String directory ) {
      return DataFile.getRelativePath(super.getFormula(), directory, UserPreferences.getPreferredFileDirectory());
   }

   @Override
   public String getFormulaLabel() {
      return ArianeSettings.getString("PropertyViewer.filename.formula");
   }

   @Override
   public String getFormulaText() {
      if (!getFormula().isEmpty()) {
         File file = new File(getFormula());
         return file.getName();
      } else {
         return ArianeSettings.getString("PropertyViewer.browser");
      }
   }

   public void synchronizeWithFileSystem() {
      if (!getFormula().isEmpty()) {
         File file = new File(getFormula());
         if (_lastModified < file.lastModified()) {
            resetAll();
            _lastModified = file.lastModified();
         }
      }
   }

   ///// 
   ///// A C C E S S O R S
   /////
   @Override
   public String getTagName() {
      return TAGNAME;
   }

   private String getAbstractFileName() {
      if (getFormula().trim().length() == 0) {
         return "";
      } else {
         File filename = new File(getFormula());
         return filename.getName();
      }
   }

   ///// 
   ///// E X E C U T I O N
   /////
   @Override
   protected boolean primitiveExecute() throws ArianeRuntimeException {
      try {
         final String filename = getFormula();
         if (filename.isEmpty()) {
            throw new ArithmeticException(ArianeSettings.getString("arithmetic.error18"));
         }
         File file = new File(filename);
         _lastModified = file.lastModified();
         getOutputResultSlot().setValue(filename);
         changeStatus(Status.EXECUTED);
         return true;
      } catch (ArithmeticException e) {
         if (Graph.isExecutionRunning()) {
            Ariane.getPresenter().printErrorMessage(ArianeSettings.ErrorCode.ARITHMETICS, getAbsoluteName(), getId(), e.getLocalizedMessage());
            getGraph().stopExecution();
            changeStatus(Status.ERROR);
         }
         throw new ArianeRuntimeException(e);
      }
   }
}
