/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.item;

import java.io.IOException;
import java.util.List;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.presenter.SubGraph;

/**
 * A <code>ForControlStructureItem</code> implements the iterative control
 * structure by a graph item. The sequence of Items to has be iterated is
 * inclosed inside a subgraph. The loop is made from the value of the
 * parameter $0 until the value of the parameter $1 by step of the value
 * of the parameter $2.
 */
public final class ForControlStructureItem extends CompositeItem {

   public static final String TAGNAME = "For";
   private static final String ABSTRACT_NAME = "for";

   // From item tree
   public ForControlStructureItem( TreeItem treeItem, Graph graph ) throws InterruptedException {
      super(treeItem, graph, ABSTRACT_NAME);
      askProperties(ABSTRACT_NAME + getId(), 1, 1, 1);
      initComment();
      setSignature(getRelatedTreeItem().getSignature());
   }

   public ForControlStructureItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
      super(graph);
      createFromXML(xmlTree);
      // Keep the following lines for the sake of compatibility
      if (getRelatedTreeItem() == null) {
         setRelatedTreeItem(restoreDefaultTreeItem());
      }
      initComment();
      setSignature(getRelatedTreeItem().getSignature());
   }

   public ForControlStructureItem( Graph graph, List<XMLElement> xmlTree, TreeItem treeItem ) throws IOException {
      super(graph, treeItem);
      createFromXML(xmlTree);
      initComment();
      setSignature(getRelatedTreeItem().getSignature());
      setAbstractName(treeItem.getName());
      setNickName(treeItem.getName());
   }

   public ForControlStructureItem( TreeItem treeItem, Graph graph, int inputDataSlotNumber, int inputParameterSlotNumber, int outputDataSlotNumber ) { //For test
      super(treeItem, graph, ABSTRACT_NAME);
      createSlots(inputDataSlotNumber, inputParameterSlotNumber + 3, 0, outputDataSlotNumber, 1);
      initComment();
      setSignature(getRelatedTreeItem().getSignature());
   }

   @Override
   public void setPrototype( String name, int iIm, int iP, int oIm ) {
      setPrototype(name, iIm, 3 + iP, oIm, iIm, 1);
   }

   @Override
   protected SubGraph createSubGraph( String name, int iIm, int iP, int oIm ) {
      return createSubGraph(getNickName(), iIm, iP - 2, oIm, iIm, 0);
   }

   @Override
   public String getTagName() {
      return TAGNAME;
   }

   @Override
   public TreeItem restoreDefaultTreeItem() {
      setAbstractName(ABSTRACT_NAME);
      return Ariane.getPresenter().getTree().getTreeItem(ABSTRACT_NAME, null);
   }

   ///// 
   ///// E X E C U T I O N
   /////
   @Override
   public void resetHookHook() {
      initComment();
   }

   @Override
   protected boolean linkItemWithVirtualItem( VirtualItem virtualItem ) throws ArianeException {
      // 1. Get and check the 3 parameter values used in the conditional expression
      int[] loop = new int[3];
      if (!getLoopParameters(loop)) {
         return false;
      }

      // 2. Set the first loop value to the virtual operator
      OutputResultSlot result1 = virtualItem.getVirtualInputParameterSlots().get(0);
      result1.setValue(Integer.toString(loop[0]));

      // 3. Get and check the rest of the parameter values
      for (int i = loop.length; i < getInputParameterSlots().size(); i++) {
         InputParameterSlot parameter = getInputParameterSlots().get(i);
         if (parameter.hasValue(true)) {
            OutputResultSlot result = virtualItem.getVirtualInputParameterSlots().get(1 + i - loop.length);
            result.setValue(parameter.getValue());
         } else {
            return false;
         }
      }

      getMask(virtualItem);

      return getInputDataFiles(virtualItem);
   }

   @Override
   protected boolean primitiveExecute() throws ArianeRuntimeException {
      try {
         VirtualItem virtualItem = getVirtualItem();
         if (!linkItemWithVirtualItem(virtualItem)) {
            return false;
         }

         int[] loop = new int[3];
         if (!getLoopParameters(loop)) {
            return false;
         }

         Ariane.getPresenter().printSystemMessage("> For loop from " + loop[0] + " to " + loop[1] + " step " + loop[2] + " do");

         if (loop[0] > loop[1] && loop[2] > 0 || loop[0] < loop[1] && loop[2] < 0) {
            Ariane.getPresenter().printSystemMessage("> end loop for");
            return false;
         }

         changeStatus(Status.RUNNING);
         // Iteration
         for (int loopv = loop[0]; loopv <= loop[1] && Graph.isExecutionRunning(); loopv += loop[2]) {
            reset();
            virtualItem.changeStatus(Status.EXECUTED);

            Ariane.getPresenter().printSystemMessage("> loop indice " + loopv);
            setComment(Integer.toString(loopv));
            changeStatus(Status.RUNNING);

            // Set the value of the inner parameter with the current loop value.
            virtualItem.getVirtualInputParameterSlots().get(0).setValue(Integer.toString(loopv));

            // The body execution is split in 2 steps
            // to avoid problems when two operators use the same entry.
            executeBody(virtualItem);
            setIntermediateInputs(virtualItem);
         }

         Ariane.getPresenter().printSystemMessage("> end loop for");

         setOutputData(virtualItem);

         if (!setOutputResult(virtualItem)) {
            return false;
         }
         changeStatus(Status.EXECUTED);
         return true;
      } catch (ArianeException e) {
         throw displayError(e, e.getErrorCode());
      } catch (IOException e) {
         throw displayError(e, ArianeSettings.ErrorCode.INPUTOUTPUT);
      }
   }

   private boolean getLoopParameters( int[] loop ) throws ArianeMissingParameterException {
      for (int i = 0; i < 3; i++) {
         InputParameterSlot parameter = getInputParameterSlots().get(i);
         if (parameter.hasNumericalValue(true)) {
            loop[i] = (int) parameter.getNumericalValue();
         } else {
            return false;
         }
      }
      return true;
   }

   ///// 
   ///// P R O P E R T I E S
   /////
   @Override
   protected void askProperties() throws InterruptedException {
      askProperties(null, getInputDataSlots().size(), 0, getOutputDataSlots().size());
   }
}
