/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.item.etc.ItemSignatureParser;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.Slot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.presenter.Presenter;
import fr.greyc.pantheon.ariane.presenter.SubGraph;
import fr.greyc.pantheon.ariane.presenter.command.ChangeBinaryEdit;
import fr.greyc.pantheon.ariane.presenter.command.ChangeFormulaEdit;
import fr.greyc.pantheon.ariane.presenter.command.RenameItemEdit;
import fr.greyc.pantheon.ariane.view.graph.item.GraphItemView;

import java.awt.Point;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.MalformedParameterizedTypeException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JTextField;

/**
 * A <code>GraphItem</code> defines an item that can be displayed in a graph.
 * It is graphically represented by a box with input and output slots.
 */
public abstract class GraphItem {

   private final int _globalId;
   private int _internalId = -1;
   private int _currentReadOrPasteSessionId;

   private String _name;
   private String _nickName;
   private String _binary;
   private String _category;
   private String _formula;

   private final GraphItemView _graphItemView;
   private TreeItem _treeItem;
   private Graph _graph;

   private boolean _selected;

   private final List<InputDataSlot> _inputDataSlots = new ArrayList<>();
   private final List<OutputDataSlot> _outputDataSlots = new ArrayList<>();
   private final List<InputParameterSlot> _inputParameterSlots = new ArrayList<>();
   private final List<OutputResultSlot> _outputResultSlots = new ArrayList<>();
   private InputMaskSlot _inputMaskSlot;

   private Status _executionStatus = Status.UNEXECUTED;

   public enum Status {
      UNEXECUTED,
      RUNNING,
      EXECUTED,
      ERROR
   }

   public GraphItem( GraphItemView graphItemView, Graph graph ) {
      _graph = graph;
      // Set a unique identifier number for the whole application!
      _globalId = _graph.getNextGraphItemIdentifier();
      _currentReadOrPasteSessionId = Presenter.getCurrentReadOrPasteSessionId();
      _graphItemView = graphItemView;
      _graphItemView.setGraphItem(this);
   }

   public GraphItem( GraphItemView graphItemView, TreeItem treeItem, Graph graph, String name, String category ) {
      this(graphItemView, graph);
      _treeItem = treeItem;
      _name = name;
      _category = category;
   }

   public int getId() {
      return _globalId;
   }

   protected final void setInternalId( int identifier ) {
      _internalId = identifier;
   }

   public final int getInternalId() {
      return _internalId;
   }

   public final boolean isInSameReadPasteSession() {
      return _currentReadOrPasteSessionId == Presenter.getCurrentReadOrPasteSessionId();
   }

   /////
   ///// P R O P E R T Y
   /////
   public final Point getLocation() {
      return _graphItemView.getLocation();
   }

   public final void setLocation( Point point ) {
      _graphItemView.setLocation(point);
   }

   public final GraphItemView getView() {
      return _graphItemView;
   }

   public final void setRelatedTreeItem( TreeItem relatedTreeItem ) {
      _treeItem = relatedTreeItem;
   }

   protected final TreeItem getRelatedTreeItem() {
      if (_treeItem == null) {
         return Ariane.getPresenter().getTree().getTreeItem(getName(), getCategory());
      } else {
         return _treeItem;
      }
   }

   /**
    * Removes the item from the graph, but it is still in the history. Disconnects it from all other items.
    */
   public final void remove() {
      disconnectAllSlots();
      _graphItemView.dispose();
      removeHook();
   }

   protected void removeHook() {
      // empty
   }

   public final void dispose() {
      disposeHook();
      disconnectAllSlots();
      _graphItemView.dispose();
      for (Slot inputDataSlot : _inputDataSlots) {
         inputDataSlot.dispose();
      }
      _inputDataSlots.clear();

      for (Slot inputParameterSlot : _inputParameterSlots) {
         inputParameterSlot.dispose();
      }
      _inputParameterSlots.clear();

      if (_inputMaskSlot != null) {
         _inputMaskSlot.dispose();
      }
      _inputMaskSlot = null;

      for (Slot outputDataSlot : _outputDataSlots) {
         outputDataSlot.dispose();
      }
      _outputDataSlots.clear();

      for (Slot outputResultSlot : _outputResultSlots) {
         outputResultSlot.dispose();
      }
      _outputResultSlots.clear();
      _graph = null;
   }

   protected void disposeHook() {
      // Empty
   }

   public final void delete() {
      _graph.removeItem(this);
   }

   /////
   ///// P R O P E R T Y
   /////
   public final void setSignature( String signature ) {
      if (signature != null && !signature.trim().isEmpty()) {
         // Set the default value for the parameters
         try {
            String[][] values = ItemSignatureParser.parse(signature, _inputParameterSlots.size());
            for (int i = 0; i < _inputParameterSlots.size(); i++) {
               _inputParameterSlots.get(i).setName(values[i][0]);
               _inputParameterSlots.get(i).setValue(values[i][1]);
            }
         } catch (MalformedParameterizedTypeException | ArrayIndexOutOfBoundsException e) {
            Ariane.getPresenter().printErrorMessage(getName() + ": " + ArianeSettings.getString("Operator.error6"));
         }
      }
   }

   /////
   ///// E X E C U T I O N
   /////
   public final boolean execute() {
      if (!Graph.isExecutionRunning()) {
         return false;
      }
      if (getExecutionStatus() == Status.EXECUTED) {
         return true;
      }
      try {
         if (primitiveExecute()) {
            _graphItemView.updatePropertyViewer();
            return true;
         }
      } catch (ArianeRuntimeException e) {
      }
      return false;
   }

   protected abstract boolean primitiveExecute() throws ArianeRuntimeException;

   public final Status getExecutionStatus() {
      return _executionStatus;
   }

   protected final void changeStatus( Status value ) {
      if (!_executionStatus.equals(value)) {
         _executionStatus = value;
         if (_executionStatus == Status.ERROR && _graph.getSuperGraphItem() != null) {
            _graph.getSuperGraphItem().changeStatus(value);
         }
         _graphItemView.displayStatus();
      }
   }

   public void resetAll() {
      if (_executionStatus != Status.UNEXECUTED) {
         reset();
         if (_outputDataSlots != null) {
            for (OutputDataSlot slot : _outputDataSlots) {
               slot.reset();
               // Reset all connected items.
               if (slot.isConnected()) {
                  for (InputSlot inputSlot : slot.getConnectedInputSlots()) {
                     inputSlot.getGraphItem().resetAll();
                  }
               }
            }
         }

         if (!_outputResultSlots.isEmpty()) {
            getOutputResultSlot().reset();
            if (getOutputResultSlot().isConnected()) {
               for (InputSlot inputSlot : getOutputResultSlot().getConnectedInputSlots()) {
                  inputSlot.getGraphItem().resetAll();
               }
            }
         }
         // Its parent
         if (_graph.getSuperGraphItem() != null) {
            _graph.getSuperGraphItem().resetAll();
         }
      }
   }

   public final void reset() {
      if (_executionStatus != Status.UNEXECUTED) {
         changeStatus(Status.UNEXECUTED);
         _graphItemView.updatePropertyViewer();
      }
      resetHook();
   }

   protected void resetHook() {
      // Empty
   }

   /////
   ///// C O N N E C T I O N
   /////
   public final void disconnectAllSlots() {
      for (Slot slot : _outputDataSlots) {
         slot.disconnectAll();
      }
      if (_inputParameterSlots != null) {
         for (InputSlot slot : _inputParameterSlots) {
            slot.disconnect();
         }
      }
      for (InputSlot slot : _inputDataSlots) {
         slot.disconnect();
      }
      if (_inputMaskSlot != null) {
         _inputMaskSlot.disconnect();
      }
      if (!_outputResultSlots.isEmpty()) {
         getOutputResultSlot().disconnectAll();
      }
   }

   public final void fireConnection( InputSlot slot, boolean connection ) {
      fireConnectionHook(slot, connection);
      _graphItemView.fireConnection();
   }

   protected void fireConnectionHook( InputSlot slot, boolean connection ) {
      // Empty
   }

   protected final void createSlots( int inputDataSlotNumber, int inputParameterSlotNumber, int inputMaskSlotNumber, int outputDataSlotNumber, int outputResultSlotNumber ) {
      for (int i = 0; i < inputDataSlotNumber; i++) {
         _inputDataSlots.add(new InputDataSlot(this, i));
      }
      if (inputMaskSlotNumber > 0) {
         _inputMaskSlot = new InputMaskSlot(this);
      }
      for (int i = 0; i < inputParameterSlotNumber; i++) {
         _inputParameterSlots.add(new InputParameterSlot(this, i));
      }
      for (int i = 0; i < outputDataSlotNumber; i++) {
         _outputDataSlots.add(new OutputDataSlot(this, i));
      }
      for (int i = 0; i < outputResultSlotNumber; i++) {
         _outputResultSlots.add(new OutputResultSlot(this, i));
      }
   }

   public final List<InputDataSlot> getInputDataSlots() {
      return _inputDataSlots;
   }

   public final List<OutputDataSlot> getOutputDataSlots() {
      return _outputDataSlots;
   }

   public final InputMaskSlot getInputMaskSlot() {
      return _inputMaskSlot;
   }

   public final List<InputParameterSlot> getInputParameterSlots() {
      return _inputParameterSlots;
   }

   public final List<OutputResultSlot> getOutputResultSlots() {
      return _outputResultSlots;
   }

   public final OutputResultSlot getOutputResultSlot() {
      if (_outputResultSlots.isEmpty()) {
         return null;
      } else {
         return _outputResultSlots.get(0);
      }
   }

   public final Slot getSlot( String typeSlot, int rank ) {
      if ("InSlotImage".equalsIgnoreCase(typeSlot)) {
         return getInputDataSlots().get(rank);
      } else if ("InSlotMask".equalsIgnoreCase(typeSlot)) {
         return getInputMaskSlot();
      } else if ("InSlotParameter".equalsIgnoreCase(typeSlot)) {
         return getInputParameterSlots().get(rank);
      } else if ("OutSlotImage".equalsIgnoreCase(typeSlot)) {
         if (rank >= getOutputDataSlots().size()) {
            return null;
         }
         return getOutputDataSlots().get(rank);
      } else if ("OutSlotResult".equalsIgnoreCase(typeSlot)) {
         // For sake of ascendent compatibility
         if (rank < getOutputResultSlots().size()) {
            return getOutputResultSlots().get(rank);
         } else {
            return null;
         }
      }
      return null;
   }

   ///// 
   ///// A C C E S S O R S
   /////
   public SubGraph getSubGraph() {
      return null;
   }

   protected final void setComment( String comment ) {
      _graphItemView.setComment(comment);
   }

   public final String getCategory() {
      if (_treeItem == null) {
         return _category;
      }
      return _treeItem.getCategory();
   }

   public final String getDocumentationManualURL() {
      if (_treeItem == null) {
         return "";
      }
      return _treeItem.getDocumentationURL();
   }

   public String getBinary() {
      return _binary;
   }

   public String getBinary( String filePath ) {
      return _binary;
   }

   public void setBinary( String binary ) {
      _binary = binary;
   }

   public String getFormula( String filepath ) {
      return _formula;
   }

   public boolean isDisplayableFormula() {
      return true;
   }

   public final String getFormula() {
      return _formula;
   }

   public boolean changeFormula( String textFormula ) {
      return true;
   }

   public void setFormula( String formula ) {
      _formula = formula.trim();
      resetAll();
      getGraph().setModified(true);
      _graphItemView.firePropertyName(getNickName());
      getGraph().getView().repaint();
   }

   public String getFormulaText() {
      return getFormula();
   }

   public String getFormulaLabel() {
      return ArianeSettings.getString("PropertyViewer.formula");
   }

   public final String getName() {
      return _name;
   }

   public final String getAbsoluteName() {
      return getGraph().getName() + "/" + getNickName();
   }

   public final void setAbstractName( String name ) {
      if (name.length() > 0) {
         _name = name;
      }
   }

   public final void setTemporaryNickName( String name ) {
      _nickName = name;
      updateName();
   }

   public void setNickName( String name ) {
      if (name != null && name.length() > 0) {
         _nickName = name;
         _graphItemView.firePropertyTextName(name);
      } else {
         _nickName = null;
         _graphItemView.firePropertyTextName("");
      }
      _graphItemView.updatePropertyViewer();
      getGraph().setModified(true);
   }

   public final void setNickName( String newName, boolean addInHistory ) {
      String oldName = _nickName;
      if (!newName.equals(oldName)) {
         setNickName(newName);
         if (addInHistory) {
            getGraph().validateOneAction(new RenameItemEdit(this, newName, oldName));
         }
         updateName();
         getGraph().revalidate();
      }
   }

   public final String getNickName() {
      return (_nickName == null) ? getName() : _nickName;
   }

   public final Graph getGraph() {
      return _graph;
   }

   public abstract String getTagName();

   /**
    * Changes the formula expression and store it in undo -redo manager.
    *
    * @param component the component that displays the formula.
    * @param newValue the new formula expression.
    * @param historical true if the action must be stored in the action history (for use with undo/redo).
    */
   public final void changeFormula( final JComponent component, String newValue, boolean historical ) {
      final String oldFormula = getFormula();
      changeFormula(newValue);
      if (historical) {
         getGraph().validateOneAction(new ChangeFormulaEdit(this, component, getFormula(), oldFormula));
      }
      if (component instanceof JTextField) {
         ((JTextField) component).setText(getFormulaText());
      } else if (component instanceof JButton) {
         ((JButton) component).setText(getFormulaText());
      }
      _graphItemView.updatePropertyViewer();
   }

   /**
    * Changes the binary expression and store it in undo -redo manager.
    *
    * @param component the component that displays the formula.
    * @param newValue the new formula expression.
    * @param historical true if the action must be stored in the action history (for use with undo/redo).
    */
   public final void changeBinary( final JComponent component, String newValue, boolean historical ) {
      setBinary(newValue);
      // Very ugly but KISS principle...
      if (component instanceof JTextField) {
         ((JTextField) component).setText(getBinary());
      } else if (component instanceof JButton) {
         ((JButton) component).setText(getBinary());
      }
      _graphItemView.updatePropertyViewer();
      if (historical) {
         String oldBinary = getBinary();
         getGraph().validateOneAction(new ChangeBinaryEdit(this, component, getBinary(), oldBinary));
      }
   }

   /**
    * Updates the item contents after modification through the properties dialog box.
    */
   public final void updateName() {
      _graphItemView.firePropertyName(getNickName());
      _graphItemView.setToolTipText(toString());
   }

   ///// 
   ///// S E L E C T I O N
   ///// 
   public final boolean isSelected() {
      return _selected;
   }

   public final void setSelected( boolean selected ) {
      if (_selected != selected) {
         for (Slot slot : _inputDataSlots) {
            slot.getView().setSelected(selected);
         }
         for (Slot slot : _inputParameterSlots) {
            slot.getView().setSelected(selected);
         }
         for (Slot slot : _outputDataSlots) {
            slot.getView().setSelected(selected);
         }
         for (Slot slot : _outputResultSlots) {
            slot.getView().setSelected(selected);
         }
         Slot slot = _inputMaskSlot;
         if (slot != null) {
            slot.getView().setSelected(selected);
         }
         _selected = selected;
         _graphItemView.setSelected(selected);
      }
   }
   /////
   /////  I N P U T S   &   O U T P U T S
   /////

   public XMLElement convertToXML( File file ) throws IOException {
      XMLElement element = new XMLElement(getTagName());
      element.addChildren(convertPropertiesToXML(file));
      return element;
   }

   protected final List<XMLElement> convertPropertiesToXML( File file ) throws IOException {
      LinkedList<XMLElement> xmlTree = new LinkedList<>();
      xmlTree.add(new XMLElement("NickName", _nickName));
      xmlTree.add(new XMLElement("Name", getName()));
      xmlTree.add(new XMLElement("Identifier", Integer.toString(getId())));
      xmlTree.add(new XMLElement("X-coordinate", Integer.toString(getLocation().x)));
      xmlTree.add(new XMLElement("Y-coordinate", Integer.toString(getLocation().y)));

      String currentPath;
      if (file != null) {
         currentPath = file.getParentFile().getAbsolutePath();
      } else {
         currentPath = null;
      }
      String tmp = getBinary(currentPath);
      if (tmp != null) {
         xmlTree.add(new XMLElement("Binary", tmp));
      }
      if (getCategory() != null) {
         xmlTree.add(new XMLElement("Category", getCategory()));
      }
      if (getFormula() != null) {
         xmlTree.add(new XMLElement("Formula", getFormula(currentPath)));
      }
      if (!_inputDataSlots.isEmpty()) {
         xmlTree.add(new XMLElement("InputFiles", Integer.toString(_inputDataSlots.size())));
      }
      if (!_inputParameterSlots.isEmpty()) {
         xmlTree.add(new XMLElement("Parameters", Integer.toString(_inputParameterSlots.size())));
      }
      if (!_outputDataSlots.isEmpty()) {
         xmlTree.add(new XMLElement("OutputFiles", Integer.toString(_outputDataSlots.size())));
      }
      if (this instanceof IScriptArgument) {
         IScriptArgument item = (IScriptArgument) this;
         xmlTree.add(new XMLElement("Is-Argument", Boolean.toString(item.isArgumentOfRelatedScript())));
      }
      return xmlTree;
   }

   public final List<XMLElement> convertSlotsToXML() throws IOException {
      List<XMLElement> domTree = new ArrayList<>();
      for (InputSlot slot : _inputDataSlots) {
         XMLElement element = slot.convertToXML();
         if (element != null) {
            domTree.add(element);
         }
      }
      for (InputSlot slot : _inputParameterSlots) {
         XMLElement element = slot.convertToXML();
         if (element != null) {
            domTree.add(element);
         }
      }
      if (_inputMaskSlot != null) {
         XMLElement element = _inputMaskSlot.convertToXML();
         if (element != null) {
            domTree.add(element);
         }
      }
      return domTree;
   }

   public final void createFromXML( List<XMLElement> xmlTree ) throws IOException {
      createPropertyFromXML(xmlTree);
      createAdditionalPropertyFromXML(xmlTree);
   }

   protected final void createPropertyFromXML( List<XMLElement> xmlTree ) throws IOException {
      int inputParameterNumber = 0;
      int inputDataNumber = 0;
      int outputDataNumber = 0;
      Point coordinates = new Point(0, 0);
      for (XMLElement element : xmlTree) {
         final String tagName = element.getName();
         final String value = element.getValue();
         // French tags are kept for sake of compatibility.
         if ("NickName".equalsIgnoreCase(tagName)) {
            setNickName(value);
         } else if ("Name".equalsIgnoreCase(tagName) || "Nom".equalsIgnoreCase(tagName)) {
            setAbstractName(value);
         } else if ("Identifier".equalsIgnoreCase(tagName) || "Identifieur".equalsIgnoreCase(tagName)) {
            setInternalId(Integer.parseInt(value));
         } else if ("X-coordinate".equalsIgnoreCase(tagName) || "Abscisse".equalsIgnoreCase(tagName)) {
            coordinates.x = Integer.parseInt(value);
         } else if ("Y-coordinate".equalsIgnoreCase(tagName) || "Ordonnee".equalsIgnoreCase(tagName)) {
            coordinates.y = Integer.parseInt(value);
         } else if ("Binary".equalsIgnoreCase(tagName) || "Binaire".equalsIgnoreCase(tagName)) {
            setBinary(value);
         } else if ("Category".equalsIgnoreCase(tagName) || "Categorie".equalsIgnoreCase(tagName)) {
            _category = value;
         } else if ("Formula".equalsIgnoreCase(tagName)) {
            setFormula(value);
         } else if ("Parameters".equalsIgnoreCase(tagName) || "ParametresEntree".equalsIgnoreCase(tagName)) {
            inputParameterNumber = Integer.parseInt(value);
         } else if ("InputFiles".equalsIgnoreCase(tagName) || "InputImages".equalsIgnoreCase(tagName) || "ImagesEntree".equalsIgnoreCase(tagName)) {
            inputDataNumber = Integer.parseInt(value);
         } else if ("OutputFiles".equalsIgnoreCase(tagName) || "OutputImages".equalsIgnoreCase(tagName) || "ImagesSortie".equalsIgnoreCase(tagName)) {
            outputDataNumber = Integer.parseInt(value);
         } else if ("Is-Argument".equalsIgnoreCase(tagName) && this instanceof IScriptArgument) {
            IScriptArgument item = (IScriptArgument) this;
            item.setArgumentOfRelatedScript(Boolean.parseBoolean(value));
         }
      }
      for (int i = 0; i < inputDataNumber; i++) {
         _inputDataSlots.add(new InputDataSlot(this, i));
      }
      for (int i = 0; i < inputParameterNumber; i++) {
         _inputParameterSlots.add(new InputParameterSlot(this, i));
      }
      for (int i = 0; i < outputDataNumber; i++) {
         _outputDataSlots.add(new OutputDataSlot(this, i));
      }
      if (_treeItem == null && Ariane.getPresenter() != null) {
         if ("File".equals(_category)) { // keep it for the sake of compatibility.
            if (this instanceof LoaderItem) {
               _name = "loader";
               createSlots(0, 1, 0, 0, 0);
               _category = null;
            } else if (this instanceof SaverItem) {
               _name = "saver";
               createSlots(0, 1, 0, 0, 0);
               _category = null;
            }
         } else if ("Viewer".equals(_category) && this instanceof ViewerItem) {
            _name = "viewer";
            _category = null;
         } else if ("Comment".equals(_category) && this instanceof NoteItem) {
            _category = null;
         } else if ("Function".equals(_category) && this instanceof EvalItem) {
            _category = null;
         }
         _treeItem = getRelatedTreeItem();
      }
      setLocation(coordinates);
   }

   protected void createAdditionalPropertyFromXML( List<XMLElement> xmlTree ) throws IOException {
      // Empty
   }

   public static final List<GraphItem> createFromXML( Graph graph, List<XMLElement> xmlTree ) throws IOException {
      List<GraphItem> items = new ArrayList<>();
      if (xmlTree == null) {
         return items;
      }

      for (XMLElement element : xmlTree) {
         GraphItem item = null;
         String tagName = element.getName();
         if (OperatorItem.TAGNAME.equalsIgnoreCase(tagName) || "Operateur".equalsIgnoreCase(tagName)) {
            item = new OperatorItem(graph, element.getChildren());
         } else if (NoteItem.TAGNAME.equalsIgnoreCase(tagName)) {
            item = new NoteItem(graph, element.getChildren());
         } else if ("Fichier".equalsIgnoreCase(tagName)) {
            // Kept for the sake of compatibility
            item = DataFile.getInstance(graph, element.getChildren());
         } else if (LoaderItem.TAGNAME.equalsIgnoreCase(tagName)) {
            item = new LoaderItem(graph, element.getChildren());
         } else if (SaverItem.TAGNAME.equalsIgnoreCase(tagName)) {
            item = new SaverItem(graph, element.getChildren());
         } else if (ViewerItem.TAGNAME.equalsIgnoreCase(tagName)) {
            item = new ViewerItem(graph, element.getChildren());
         } else if (FileNameItem.TAGNAME.equalsIgnoreCase(tagName) || "Datafile".equalsIgnoreCase(tagName)) {
            item = new FileNameItem(graph, element.getChildren());
         } else if (ConstantItem.TAGNAME.equalsIgnoreCase(tagName)) {
            item = new ConstantItem(graph, element.getChildren());
         } else if (SystemItem.TAGNAME.equalsIgnoreCase(tagName)) {
            item = new SystemItem(graph, element.getChildren());
         } else if (ScriptItem.TAGNAME.equalsIgnoreCase(tagName)) {
            item = new ScriptItem(graph, element.getChildren());
         } else if (MacroOperatorItem.TAGNAME.equalsIgnoreCase(tagName) || "MacroOperateur".equalsIgnoreCase(tagName)) {
            item = new MacroOperatorItem(graph, element.getChildren());
         } else if (MergeControlStructureItem.TAGNAME.equalsIgnoreCase(tagName)) {
            item = new MergeControlStructureItem(graph, element.getChildren());
         } else if (EvalItem.TAGNAME.equalsIgnoreCase(tagName)) {
            item = new EvalItem(graph, element.getChildren());
         } else if (FormatItem.TAGNAME.equalsIgnoreCase(tagName)) {
            item = new FormatItem(graph, element.getChildren());
         } else if (IfControlStructureItem.TAGNAME.equalsIgnoreCase(tagName)) {
            item = new IfControlStructureItem(graph, element.getChildren());
         } else if (SelectControlStructureItem.TAGNAME.equalsIgnoreCase(tagName)) {
            item = new SelectControlStructureItem(graph, element.getChildren());
         } else if (WhileControlStructureItem.TAGNAME.equalsIgnoreCase(tagName)) {
            item = new WhileControlStructureItem(graph, element.getChildren());
         } else if ("Until".equalsIgnoreCase(tagName)) { // Keep for compatibility
            item = new WhileControlStructureItem(graph, element.getChildren());
         } else if (ForControlStructureItem.TAGNAME.equalsIgnoreCase(tagName)) {
            item = new ForControlStructureItem(graph, element.getChildren());
         }
         if (item != null) {
            graph.addItemAtAbsolutePosition(item, item.getLocation(), true);
            items.add(item);
         }
      }
      return items;
   }
}
