/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeArithmeticException;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.item.etc.LogicalExpression;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.item.ControlStructureItemView;

import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

/**
 * A <code>ControlStructureIfItem</code> implements the conditional
 * <code>ControlStructure</code>, as a tradition if (test) then ... else sequence.
 */
public final class IfControlStructureItem extends ControlItem {

   public static final String TAGNAME = "If";

   public IfControlStructureItem( TreeItem treeItem, Graph graph ) {
      super(new ControlStructureItemView(), treeItem, graph, treeItem.getName());
      createSlots(treeItem.getNumberOfInputDataSlots(), treeItem.getNumberOfParameterSlots(), 0, treeItem.getNumberOfOutputDataSlots(), 0);
      setAbstractName(treeItem.getName());
      setFormula(""); // IMPERATIVE to signify that it uses formula.
   }

   public IfControlStructureItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
      super(new ControlStructureItemView(), graph);
      createFromXML(xmlTree);

      String expression;
      if (getFormula() == null && getBinary() != null) { // Keep this for sake of compatibility
         expression = getBinary();
         setBinary(null);
      } else {
         expression = getFormula();
      }
      try {
         checkExpression(expression);
         setFormula(expression);
         setComment(expression);
      } catch (ArianeArithmeticException e) {
         Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
         setFormula("");
      }
   }

   @Override
   public void setPrototype( String name, int iIm, int iP, int oIm ) {
      // Empty
   }

   @Override
   public String getTagName() {
      return TAGNAME;
   }

   ///// 
   /////  E X E C U T I O N 
   /////
   @Override
   protected boolean primitiveExecute() throws ArianeRuntimeException {
      try {
         if (getFormula().isEmpty()) {
            throw new ArianeArithmeticException(ArianeSettings.getString("arithmetic.error0"));
         }

         // gets input parameters.
         double[] val = new double[2];
         if (!getInputparameters(val)) {
            return false;
         }

         // Get the input data
         List<String> inputLine = new LinkedList<>();
         for (InputDataSlot slot : getInputDataSlots()) {
            if (slot.hasValue(true)) {
               inputLine.add(slot.getValue());
            } else {
               return false;
            }
         }

         // gets input images.
         InputDataSlot input = getInputDataSlots().get(0);
         input.hasValue(true);

         // use the conditional expression result to select the right output.
         LogicalExpression conditionExpression = getCompiledExpression();
         int outputRank = conditionExpression.evaluate(val[0], val[1]) ? 0 : 1;

         // validate the selected output.
         getOutputDataSlots().get(outputRank).setValue(input.getValue());
         // unvalidate the other one.
         getOutputDataSlots().get(1 - outputRank).reset();

         getOutputResultSlot().setValue(Integer.toString(outputRank));
         changeStatus(Status.EXECUTED);
         return true;
      } catch (ArianeException e) {
         if (Graph.isExecutionRunning()) {
            Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
            getGraph().stopExecution();
            changeStatus(Status.ERROR);
         }
         throw new ArianeRuntimeException(e);
      }
   }

   protected boolean getInputparameters( double[] val ) throws ArianeMissingParameterException {
      for (int i = 0; i < 2; i++) {
         InputParameterSlot parameter = getInputParameterSlots().get(i);
         if (parameter.hasNumericalValue(true)) {
            val[i] = parameter.getNumericalValue();
         } else {
            return false;
         }
      }
      return true;
   }

   ///// 
   /////  P R O P E R T I E S 
   /////
   @Override
   public boolean changeFormula( String conditionExpression ) {
      if (getFormula() == null || !getFormula().equals(conditionExpression)) {
         try {
            checkExpression(conditionExpression);
            setFormula(conditionExpression);
            setComment(conditionExpression);
         } catch (ArianeArithmeticException e) {
            Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
            return false;
         }
      }
      return true;
   }

   protected LogicalExpression getCompiledExpression() {
      try {
         return new LogicalExpression(getFormula());
      } catch (ArianeArithmeticException e) {
         return null;
      }
   }

   private void checkExpression( String expression ) throws ArianeArithmeticException {
      new LogicalExpression(expression);
   }
}
