/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.item;

import java.io.IOException;
import java.util.Iterator;
import java.util.List;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.presenter.SubGraph;

/**
 * A <code>MacroOperatorItem</code> is a graph item that is composed by a subgraph of items.
 */
public final class MacroOperatorItem extends CompositeItem {

   public static final String TAGNAME = "MacroOperator";
   private static final String ABSTRACT_NAME = "macro-operator";

   // From item tree
   public MacroOperatorItem( TreeItem treeItem, Graph graph ) throws InterruptedException {
      super(treeItem, graph, ABSTRACT_NAME);
      askProperties(ABSTRACT_NAME + getId(), 1, 1, 1);
   }

   public MacroOperatorItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
      super(graph);
      createFromXML(xmlTree);

      // Keep the following lines for the sake of compatibility
      if (getRelatedTreeItem() == null) {
         setRelatedTreeItem(restoreDefaultTreeItem());
      }
      if (getInputMaskSlot() != null) {
         createSlots(0, 0, 1, 0, 0);
      }

      if (getRelatedTreeItem() != null) {
         setSignature(getRelatedTreeItem().getSignature());
      }
   }

   public MacroOperatorItem( Graph graph, List<XMLElement> xmlTree, TreeItem treeItem ) throws IOException {
      super(graph, treeItem);
      createFromXML(xmlTree);
      // Keep the following lines for the sake of compatibility
      if (getInputMaskSlot() != null) {
         createSlots(0, 0, 1, 0, 0);
      }
      setSignature(getRelatedTreeItem().getSignature());
      setAbstractName(treeItem.getName());
      setNickName(treeItem.getName());
   }

   public MacroOperatorItem( TreeItem treeItem, Graph graph, int inputDataSlotNumber, int inputParameterSlotNumber, int outputDataSlotNumber ) { //For test
      super(treeItem, graph, ABSTRACT_NAME);
      createSlots(inputDataSlotNumber, inputParameterSlotNumber, 0, outputDataSlotNumber, 1);
   }

   @Override
   public void setPrototype( String name, int inputDataCount, int parameterCount, int outputDataCount ) {
      setPrototype(name, inputDataCount, parameterCount, outputDataCount, 0, 0);
   }

   @Override
   protected SubGraph createSubGraph( String name, int inputDataCount, int parameterCount, int outputDataCount ) {
      return createSubGraph(getNickName(), inputDataCount, parameterCount, outputDataCount, 0, 0);
   }

   @Override
   public String getTagName() {
      return TAGNAME;
   }

   @Override
   public TreeItem restoreDefaultTreeItem() {
      setAbstractName(ABSTRACT_NAME);
      return Ariane.getPresenter().getTree().getTreeItem(ABSTRACT_NAME, null);
   }

   ///// 
   ///// E X E C U T I O N
   /////
   @Override
   protected boolean linkItemWithVirtualItem( VirtualItem virtualItem ) throws ArianeException {
      Iterator<InputParameterSlot> its1 = getInputParameterSlots().iterator();
      Iterator<OutputResultSlot> itd1 = virtualItem.getVirtualInputParameterSlots().iterator();
      while (its1.hasNext()) {
         InputParameterSlot parameter = its1.next();
         if (parameter.hasValue(true)) {
            OutputResultSlot result = itd1.next();
            result.setValue(parameter.getValue());
         } else {
            return false;
         }
      }

      getMask(virtualItem);

      return getInputDataFiles(virtualItem);
   }

   @Override
   protected boolean primitiveExecute() throws ArianeRuntimeException {
      try {
         VirtualItem virtualItem = getVirtualItem();
         if (!linkItemWithVirtualItem(virtualItem)) {
            return false;
         }

         changeStatus(Status.RUNNING);

         setOutputData(virtualItem);
         if (!setOutputResult(virtualItem)) {
            return false;
         }
         changeStatus(Status.EXECUTED);
         return true;
      } catch (ArianeException e) {
         throw displayError(e, e.getErrorCode());
      }
   }

   ///// 
   ///// P R O P E R T I E S
   /////
   @Override
   protected void askProperties() throws InterruptedException {
      askProperties(getName(), getInputDataSlots().size(), getInputParameterSlots().size(), getOutputDataSlots().size());
   }
}
