/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.item;

import java.io.IOException;
import java.util.List;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.item.ControlStructureItemView;

/**
 * A <code>ControlStructureSelectItem</code> implements the a selector of multiple inputs.
 * The selected input is given by the parameter value.
 */
public final class SelectControlStructureItem extends ControlItem {

   public static final String TAGNAME = "Select";

   public SelectControlStructureItem( TreeItem treeItem, Graph graph ) throws InterruptedException {
      super(new ControlStructureItemView(), treeItem, graph, treeItem.getName());
      askProperties(null, 1, 0, 0);
      setSignature(treeItem.getSignature());
   }

   public SelectControlStructureItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
      super(new ControlStructureItemView(), graph);
      createFromXML(xmlTree);
   }

   public SelectControlStructureItem( TreeItem treeItem, Graph graph, int inputDataSlotNumber, int inputParameterSlotNumber, int outputDataSlotNumber ) { // For tests
      super(new ControlStructureItemView(), treeItem, graph, treeItem.getName());
      createSlots(inputDataSlotNumber, inputParameterSlotNumber, 0, outputDataSlotNumber, 1);
      setSignature(treeItem.getSignature());
   }

   @Override
   public void setPrototype( String name, int iIm, int iP, int oIm ) {
      setAbstractName(name);
      createSlots(iIm, 1, 0, 1, 0);
   }

   @Override
   public String getTagName() {
      return TAGNAME;
   }

   ///// 
   /////  E X E C U T I O N 
   /////
   @Override
   protected boolean primitiveExecute() throws ArianeRuntimeException {
      try {
         // Gets and checks the parameter (-> executes forward operators if needed).
         InputParameterSlot parameter = getInputParameterSlots().get(0);
         int inputRank;
         if (parameter.hasNumericalValue(true)) {
            inputRank = (int) parameter.getNumericalValue();
         } else {
            return false;
         }

         // Checks consistency of the parameter.
         if (inputRank < 0 || inputRank >= getInputDataSlots().size()) {
            throw new ArithmeticException(ArianeSettings.getString("controlstructure.error1"));
         }

         // Gets and checks the input image (-> executes forward operators if needed).
         InputDataSlot isi = getInputDataSlots().get(inputRank);
         if (isi != null && isi.hasValue(true)) {
            // Validate the output.
            (getOutputDataSlots().get(0)).setValue(isi.getValue());
            getOutputResultSlot().setValue(Float.toString(inputRank));
         } else {
            return false;
         }
         changeStatus(Status.EXECUTED);
         return true;
      } catch (ArianeException e) {
         if (Graph.isExecutionRunning()) {
            Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
            getGraph().stopExecution();
            changeStatus(Status.ERROR);
         }
         throw new ArianeRuntimeException(e);
      }
   }

   ///// 
   /////  P R O P E R T I E S
   /////
   protected void askProperties() throws InterruptedException {
      askProperties(null, 1, 0, 0);
   }
}
