/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.item;

import java.io.File;
import java.io.IOException;
import java.util.List;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.Slot;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.SubGraph;
import fr.greyc.pantheon.ariane.view.graph.item.VirtualItemView;

/**
 * A <code>VirtualItem</code> is used to reference inputs and outputs
 * of a <code>SubGraph</code>. It is invisible as it has no graphical
 * representation.
 */
public final class VirtualItem extends GraphItem {

   private final int _outputDataSlotNumber;

   public VirtualItem( SubGraph graph, String name, int inputDataNumber, int inputParameterNumber, int outputDataNumber, int intermediateInputDataNumber, int intermediateInputParameterNumber ) {
      super(new VirtualItemView(), null, graph, name, null);
      setInternalId(0);       // IMPERATIVE = 0.
      _outputDataSlotNumber = outputDataNumber;
      createSlots(_outputDataSlotNumber + intermediateInputDataNumber,
          1 + intermediateInputParameterNumber,
          0,
          inputDataNumber + 1,
          inputParameterNumber);

      buildVirtualInputData(graph, inputDataNumber);
      buildVirtualInputMask(graph, inputDataNumber);

      buildVirtualOutputData(graph);
      buildIntermediateInputData(graph, intermediateInputDataNumber);

      buildOutputResult(graph);
      buildIntermediateParameters(graph, inputParameterNumber, intermediateInputParameterNumber);

      buildInputParameters(graph, inputParameterNumber);

      getView().setFocusable(false);
   }

   private void buildVirtualInputData( SubGraph graph, int inputDataNumber ) {
      List<OutputDataSlot> inputDataSlots = getOutputDataSlots();
      for (int i = 0; i < inputDataNumber; i++) {
         OutputDataSlot slot = inputDataSlots.get(i);
         slot.getView().setToolTipText(graph.getSuperGraphItem().getNickName() + ArianeSettings.getString("VirtualOperator.input") + i);
         inputDataSlots.add(slot);
      }
   }

   private void buildVirtualInputMask( SubGraph graph, int inputDataNumber ) {
      List<OutputDataSlot> inputDataSlots = getOutputDataSlots();
      OutputDataSlot slot = inputDataSlots.get(inputDataNumber);
      slot.getView().setToolTipText(graph.getSuperGraphItem().getNickName() + ArianeSettings.getString("VirtualOperator.mask"));
      inputDataSlots.add(slot);
   }

   private void buildVirtualOutputData( SubGraph graph ) {
      List<InputDataSlot> outputDataSlots = getInputDataSlots();
      for (int i = 0; i < _outputDataSlotNumber; i++) {
         InputDataSlot slot = outputDataSlots.get(i);
         slot.getView().setToolTipText(graph.getSuperGraphItem().getNickName() + ArianeSettings.getString("VirtualOperator.output") + i);
         outputDataSlots.add(slot);
      }
   }

   private void buildIntermediateInputData( SubGraph graph, int intermediateInputDataNumber ) {
      List<InputDataSlot> outputDataSlots = getInputDataSlots();
      for (int i = _outputDataSlotNumber; i < _outputDataSlotNumber + intermediateInputDataNumber; i++) {
         InputDataSlot slot = outputDataSlots.get(i);
         int number = i - _outputDataSlotNumber;
         slot.getView().setToolTipText(graph.getSuperGraphItem().getNickName() + ArianeSettings.getString("VirtualOperator.reinput") + number);
         outputDataSlots.add(slot);
      }
   }

   private void buildOutputResult( SubGraph graph ) {
      List<InputParameterSlot> outResultSlots = getInputParameterSlots();
      InputParameterSlot slot = outResultSlots.get(0);
      slot.getView().setToolTipText(graph.getSuperGraphItem().getNickName() + ArianeSettings.getString("VirtualOperator.result"));
      outResultSlots.add(slot);
   }

   private void buildIntermediateParameters( SubGraph graph, int inputParameterNumber, int intermediateInputParameterNumber ) {
      List<InputParameterSlot> outResultSlots = getInputParameterSlots();
      for (int i = 1; i < 1 + intermediateInputParameterNumber; i++) {
         InputParameterSlot slot = outResultSlots.get(i);
         int number = i + inputParameterNumber - intermediateInputParameterNumber - 1;
         slot.getView().setToolTipText(graph.getSuperGraphItem().getNickName() + ArianeSettings.getString("VirtualOperator.reparameter") + number);
         outResultSlots.add(slot);
      }
   }

   private void buildInputParameters( SubGraph graph, int inputParameterNumber ) {
      List<OutputResultSlot> inputParameterSlots = getOutputResultSlots();
      for (int i = 0; i < inputParameterNumber; i++) {
         OutputResultSlot slot = inputParameterSlots.get(i);
         slot.getView().setToolTipText(graph.getSuperGraphItem().getNickName() + ArianeSettings.getString("VirtualOperator.parameter") + i);
         inputParameterSlots.add(slot);
      }
   }

   @Override
   public int getId() {
      return 0;
   }

   public OutputDataSlot getVirtualInputDataSlot( int index ) {
      return getOutputDataSlots().get(index);
   }

   public List<OutputDataSlot> getVirtualInputDataSlots() {
      return getOutputDataSlots();
   }

   public List<InputDataSlot> getVirtualOutputDataSlots() {
      return getInputDataSlots();
   }

   public List<OutputResultSlot> getVirtualInputParameterSlots() {
      return getOutputResultSlots();
   }

   public InputDataSlot getVirtualOutputDataSlot( int index ) {
      return getInputDataSlots().get(index);
   }

   public OutputResultSlot getVirtualInputParameterSlot( int index ) {
      return getOutputResultSlots().get(index);
   }

   public InputParameterSlot getVirtualIntermediateResultSlot( int index ) {
      return getInputParameterSlots().get(1 + index);
   }

   public Slot getVirtualInputMaskSlot() {
      return getOutputDataSlots().get(getOutputDataSlots().size() - 1);
   }

   public Slot getVirtualOutputResultSlot() {
      return getInputParameterSlots().get(0);
   }

   public InputDataSlot getVirtualIntermediateDataSlots( int index ) {
      return getInputDataSlots().get(index + _outputDataSlotNumber);
   }

   @Override
   public void resetAll() {
      if (getExecutionStatus() != Status.UNEXECUTED) {
         // Overrides the graphitem reset in order to
         // avoid resetting the inter images and inter parameters
         // during a loop since a loop need to reset the
         // contents of the subgraph for each iterations.
         reset();
         if (getOutputDataSlots() != null) {
            for (OutputDataSlot slot : getOutputDataSlots()) {
               slot.reset();
            }
         }
         if (!getOutputResultSlots().isEmpty()) {
            getOutputResultSlot().reset();
         }
         // Its parent
         if (getGraph().getSuperGraphItem() != null) {
            getGraph().getSuperGraphItem().resetAll();
         }
      }
   }

   @Override
   protected boolean primitiveExecute() {
      CompositeItem item = (CompositeItem) getGraph().getSuperGraphItem();
      try {
         if (!item.linkItemWithVirtualItem(this)) {
            return false;
         }
         changeStatus(Status.EXECUTED);
         return true;
      } catch (ArianeException e) {
         changeStatus(Status.UNEXECUTED);
         GraphItem composite = getGraph().getSuperGraphItem();
         Ariane.getPresenter().printErrorMessage(e.getErrorCode(), composite.getAbsoluteName(), composite.getId(), e.getLocalizedMessage());
         return false;
      }
   }

   @Override
   public XMLElement convertToXML( File file ) throws IOException {
      return null;
   }

   @Override
   public String getTagName() {
      return null;
   }
}
