/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeArithmeticException;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.item.etc.LogicalExpression;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.presenter.SubGraph;

import java.io.IOException;
import java.util.List;

/**
 * A <code>WhileControlStructureItem</code> implements the repetitive structure
 * control by a graph item. The sequence of Items to has be repeated is inclosed
 * inside a subgraph.
 */
public final class WhileControlStructureItem extends CompositeItem {

   public static final String TAGNAME = "While";
   private static final String ABSTRACT_NAME = "while";

   // From item tree
   public WhileControlStructureItem( TreeItem treeItem, Graph graph ) throws InterruptedException {
      super(treeItem, graph, ABSTRACT_NAME);
      askProperties(ABSTRACT_NAME + getId(), 1, 1, 1);
      initComment();
      setSignature(getRelatedTreeItem().getSignature());
   }

   // From file
   public WhileControlStructureItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
      super(graph);
      createFromXML(xmlTree);
      String expression;
      if (getFormula() == null && getBinary() != null) { // Keep this for sake of compatibility
         expression = getBinary();
         setBinary(null);
      } else {
         expression = getFormula();
      }

      try {
         checkExpression(expression);
         setFormula(expression);
      } catch (ArianeArithmeticException e) {
         Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
         setFormula("");
      }
      if (getRelatedTreeItem() == null) {
         setRelatedTreeItem(restoreDefaultTreeItem());
      }
      initComment();
      setSignature(getRelatedTreeItem().getSignature());
   }

   // From subroutine
   public WhileControlStructureItem( Graph graph, List<XMLElement> xmlTree, TreeItem treeItem ) throws IOException {
      super(graph, treeItem);
      createFromXML(xmlTree);
      String expression;
      if (getFormula() == null && getBinary() != null) { // Keep this for sake of compatibility
         expression = getBinary();
         setBinary(null);
      } else {
         expression = getFormula();
      }
      try {
         checkExpression(expression);
         setFormula(expression);
      } catch (ArianeArithmeticException e) {
         Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
         setFormula("");
      }
      setAbstractName(treeItem.getName());
      setNickName(treeItem.getName());
      initComment();
      setSignature(getRelatedTreeItem().getSignature());
   }

   // For tests
   protected WhileControlStructureItem( TreeItem treeItem, Graph graph, int inputDataSlotNumber, int inputParameterSlotNumber, int outputDataSlotNumber ) { //For test
      super(treeItem, graph, ABSTRACT_NAME);
      createSlots(inputDataSlotNumber, inputParameterSlotNumber + 3, 0, outputDataSlotNumber, 1);
      initComment();
      setSignature(getRelatedTreeItem().getSignature());
   }

   @Override
   public void setPrototype( String name, int iIm, int iP, int oIm ) {
      super.setPrototype(name, iIm, 2 + iP, oIm, iIm, 2);
      setFormula(""); // IMPERATIVE to signify that it uses formula.
   }

   @Override
   protected SubGraph createSubGraph( String name, int iIm, int iP, int oIm ) {
      return createSubGraph(getNickName(), iIm, iP, oIm, iIm, 2);
   }

   @Override
   public String getTagName() {
      return TAGNAME;
   }

   @Override
   public TreeItem restoreDefaultTreeItem() {
      setAbstractName(ABSTRACT_NAME);
      return Ariane.getPresenter().getTree().getTreeItem(ABSTRACT_NAME, null);
   }

   /////
   ///// E X E C U T I O N
   /////
   @Override
   public void resetHookHook() {
      initComment();
   }

   @Override
   protected boolean linkItemWithVirtualItem( VirtualItem virtualItem ) throws ArianeException {
      // 1. Get the parameter values
      for (int i = 0; i < getInputParameterSlots().size(); i++) {
         InputParameterSlot parameter = getInputParameterSlots().get(i);
         if (parameter.hasValue(true)) {
            OutputResultSlot result = virtualItem.getVirtualInputParameterSlots().get(i);
            result.setValue(parameter.getValue());
         } else {
            return false;
         }
      }

      getMask(virtualItem);

      return getInputDataFiles(virtualItem);
   }

   @Override
   protected boolean primitiveExecute() throws ArianeRuntimeException {
      try {
         if (getFormula().isEmpty()) {
            throw new ArianeArithmeticException(ArianeSettings.getString("arithmetic.error0"));
         }

         VirtualItem virtualItem = getVirtualItem();
         if (!linkItemWithVirtualItem(virtualItem)) {
            return false;
         }

         double[] whilevariables = new double[2];
         if (!getLoopParameters(whilevariables)) {
            return false;
         }

         changeStatus(Status.RUNNING);

         // Repetition
         int iteration = 0;
         LogicalExpression condition = getCompiledExpression();

         Ariane.getPresenter().printSystemMessage("> While loop (" + condition.toHtml() + ") do");
         while (condition.evaluate(whilevariables[0], whilevariables[1]) && Graph.isExecutionRunning()) {
            iteration++;

            reset();
            virtualItem.changeStatus(Status.EXECUTED); // Prevent the rexecution of the virtual operator.

            Ariane.getPresenter().printSystemMessage("> cycle " + iteration);
            setComment(Integer.toString(iteration));
            changeStatus(Status.RUNNING);

            // Set the value of the 2 inner parameters.
            for (int i = 0; i < 2; i++) {
               virtualItem.getVirtualInputParameterSlot(i).setValue(Float.toString((float) whilevariables[i]));
            }

            // The body execution is split in 2 steps
            // to avoid problems when two operators use the same entry.
            executeBody(virtualItem);
            setIntermediateInputs(virtualItem);

            setLoopParameterValues(virtualItem, whilevariables);
         }
         Ariane.getPresenter().printSystemMessage("end while loop");

         setOutputData(virtualItem);

         if (!setOutputResult(virtualItem)) {
            return false;
         }
         changeStatus(Status.EXECUTED);
         return true;
      } catch (ArianeException e) {
         throw displayError(e, e.getErrorCode());
      } catch (IOException e) {
         throw displayError(e, ArianeSettings.ErrorCode.INPUTOUTPUT);
      }
   }

   private boolean getLoopParameters( double[] whilevariables )
       throws ArianeMissingParameterException {
      for (int i = 0; i < 2; i++) {
         InputParameterSlot parameter = getInputParameterSlots().get(i);
         if (parameter.hasNumericalValue(true)) {
            whilevariables[i] = (int) parameter.getNumericalValue();
         } else {
            return false;
         }
      }
      return true;
   }

   private void setLoopParameterValues( VirtualItem virtualItem, double[] whilevariables ) throws ArianeMissingParameterException {
      for (int i = 0; i < 2; i++) {
         InputParameterSlot parameter = virtualItem.getVirtualIntermediateResultSlot(i);
         if (parameter.isConnected() && parameter.hasNumericalValue(true)) {
            whilevariables[i] = (int) parameter.getNumericalValue();
         }
      }
   }

   /////
   ///// P R O P E R T I E S
   /////
   @Override
   public boolean changeFormula( String condition ) {
      if (getFormula() != null && getFormula().equals(condition)) {
         return true;
      }
      try {
         checkExpression(condition);
         setFormula(condition);
         return true;
      } catch (ArianeArithmeticException e) {
         Ariane.getPresenter().printErrorMessage(e.getLocalizedMessage());
      }
      return false;
   }

   private LogicalExpression getCompiledExpression() {
      try {
         return new LogicalExpression(getFormula());
      } catch (ArianeArithmeticException e) {
         return null;
      }
   }

   private void checkExpression( String textFormula ) throws ArianeArithmeticException {
      new LogicalExpression(textFormula);
   }

   @Override
   protected void askProperties() throws InterruptedException {
      askProperties(null, 1, 0, 0);
   }
}
