/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.item.etc;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeArithmeticException;

/**
 * A <code>StringExpression</code> allows string manipulation.
 */
public final class StringExpression {

   private static final byte PLUS = -1;
   private static final byte MINUS = -2;
   private static final byte POWER = -3;
   private static final byte MIN = -4;
   private static final byte QUOTE = -5;

   private final int _nbParameters;

   private final String[] _functionNamesUnary = {"quote"};
   private final String[] _functionNamesBinary = {"concatenate", "substring"};

   private final String _initialExpression;
   private byte[] _codedExpression;
   private final byte[] _variables;
   private String[] _stack;
   private String[] _constantStack;

   private int _pos;
   private int _constantStackIndex;
   private int _codeSize;

   public StringExpression( String formula, int nbParameters ) throws ArianeArithmeticException {
      _nbParameters = nbParameters;
      _variables = new byte[_nbParameters];
      _initialExpression = formula;
      if (formula == null || formula.trim().isEmpty()) {
         return;
      }
      for (int i = 0; i < _nbParameters; i++) {
         _variables[i] = (byte) (-34 - i);
      }
      _codedExpression = new byte[formula.length()];
      _constantStack = new String[formula.length()];

      // Convert formula to computable code
      _pos = 0;
      _constantStackIndex = 0;
      _codeSize = 0;
      parseFormula();
      skipWhiteSpaces();
      if (nextCharacter() != 0) {
         error(ArianeSettings.getString("arithmetic.error1"));
      }
      int stackSize = computeStackUsage();
      _stack = new String[stackSize];
      byte[] code = new byte[_codeSize];
      System.arraycopy(_codedExpression, 0, code, 0, _codeSize);
      _codedExpression = code;
      String[] tmp = new String[_constantStackIndex];
      System.arraycopy(_constantStack, 0, tmp, 0, _constantStackIndex);
      _constantStack = tmp;
   }

   public String evaluate( String[] values ) {
      int top = 0;
      boolean flag;
      int var;
      for (int i = 0; i < _codeSize; i++) {
         var = 0;
         flag = false;
         if (_codedExpression[i] >= 0) { // constants
            _stack[top++] = _constantStack[_codedExpression[i]];
         } else if (_codedExpression[i] >= POWER) { // binary operators
            String op2 = _stack[--top];
            String op1 = _stack[--top];
            String ans = null;
            if (_codedExpression[i] == PLUS) {
               ans = op1 + op2;
            }
            if (ans == null) {
               return ans;
            }
            _stack[top++] = ans;
         } else {
            // parameters
            while (!flag && var < _nbParameters) {
               if (_codedExpression[i] == _variables[var]) {
                  flag = true;
                  _stack[top++] = values[var];
               }
               var++;
            }
            // otherwise
            if (!flag) {
               // string op2;
               String op1 = _stack[--top];
               String ans;
               if (_codedExpression[i] == QUOTE) {
                  ans = "'" + op1 + "'";
               } else {
                  return null;
               }
               _stack[top++] = ans;
            }
         }
      }
      return _stack[0];
   }

   private void error( String message ) throws ArianeArithmeticException {
      throw new ArianeArithmeticException(ArianeSettings.getString("arithmetic.error0.1") + " " + message + " " + ArianeSettings.getString("arithmetic.error0.2") + " " + _pos + ")");
   }

   private int computeStackUsage() {
      int size = 0;
      int max = 0;
      for (int i = 0; i < _codeSize; i++) {
         if (_codedExpression[i] >= 0 || isVariable(i)) {
            size++;
            if (size > max) {
               max = size;
            }
         } else if (_codedExpression[i] >= POWER) {
            size--;
         }
      }
      return max;
   }

   /**
    * Tells if the specified component from the coded expression is a parameter or not.
    *
    * @param rank the rank of the item we want to check.
    * @return true if it is a parameter, false if not.
    */
   private boolean isVariable( int rank ) {
      return _codedExpression[rank] <= -34 && _codedExpression[rank] > -34 - _nbParameters;
   }

   private char nextCharacter() {
      return _pos >= _initialExpression.length() ? 0 : _initialExpression.charAt(_pos);
   }

   private void skipWhiteSpaces() {
      while (Character.isWhitespace(nextCharacter())) {
         _pos++;
      }
   }

   /**
    * Parses a mathematical expression such as: [+|-]Term[+|-]Term.
    */
   private void parseFormula() throws ArianeArithmeticException {
      skipWhiteSpaces();
      parseTerm();
      skipWhiteSpaces();
      while (nextCharacter() == '+' || nextCharacter() == '-') {
         char op = nextCharacter();
         _pos++;
         parseTerm();
         _codedExpression[_codeSize++] = (op == '+') ? PLUS : MINUS;
         skipWhiteSpaces();
      }
   }

   /**
    * Parses an expression such as: [VARIABLE1 | VARIABLE2| FUNCTION | NUMBER | (FORMULA) ].
    */
   private void parseTerm() throws ArianeArithmeticException {
      skipWhiteSpaces();
      char ch = nextCharacter();
      if (ch == '$') {
         _pos++;
         if (Integer.valueOf(nextCharacter()) < (48 + _nbParameters) && Integer.valueOf(nextCharacter()) >= 48) {
            _codedExpression[_codeSize++] = _variables[-48 + Integer.valueOf(nextCharacter())];
         } else {
            error(ArianeSettings.getString("arithmetic.error2") + " " + ch + nextCharacter());
         }
         _pos++;
      } else if (Character.isLetter(ch)) {
         parseFunctionName();
      } else if (ch == '\"') {
         parseString();
      } else if (ch == ')') {
         error(ArianeSettings.getString("arithmetic.error4"));
      } else {
         error(ArianeSettings.getString("arithmetic.error7.1") + ch + ArianeSettings.getString("arithmetic.error7.2"));
      }
   }

   private void parseFunctionName() throws ArianeArithmeticException {
      String functionName = "";
      while (Character.isLetterOrDigit(nextCharacter())) {
         functionName += nextCharacter();
         _pos++;
      }
      functionName = functionName.toLowerCase();
      for (int i = 0; i < _functionNamesUnary.length; i++) {
         if (functionName.equals(_functionNamesUnary[i])) {
            skipWhiteSpaces();
            if (nextCharacter() != '(') {
               error(ArianeSettings.getString("arithmetic.error8.1`") + functionName + ArianeSettings.getString("arithmetic.error8.2"));
            }
            _pos++;
            parseFormula();
            skipWhiteSpaces();
            if (nextCharacter() != ')') {
               error(ArianeSettings.getString("arithmetic.error9") + functionName + "'.");
            }
            _pos++;
            _codedExpression[_codeSize++] = (byte) (QUOTE - i);
            return;
         }
      }
      for (int i = 0; i < _functionNamesBinary.length; i++) {
         if (functionName.equals(_functionNamesBinary[i])) {
            skipWhiteSpaces();
            if (nextCharacter() != '(') {
               error(ArianeSettings.getString("arithmetic.error10.1") + functionName + ArianeSettings.getString("arithmetic.error10.2"));
            }
            _pos++;
            parseFormula();
            skipWhiteSpaces();
            if (nextCharacter() != ',') {
               error(ArianeSettings.getString("arithmetic.error11") + functionName + "'.");
            }
            _pos++;
            parseFormula();
            skipWhiteSpaces();
            if (nextCharacter() != ')') {
               error(ArianeSettings.getString("arithmetic.error12") + functionName + "'.");
            }
            _pos++;
            _codedExpression[_codeSize++] = (byte) (MIN - i);
            return;
         }
      }
      error(ArianeSettings.getString("arithmetic.error13") + functionName);
   }

   private void parseString() {
      String value = "";
      _pos++;
      while (nextCharacter() != '\"') {
         value += nextCharacter();
         _pos++;
      }
      _pos++;
      _codedExpression[_codeSize++] = (byte) _constantStackIndex;
      _constantStack[_constantStackIndex++] = value;
   }

   @Override
   public String toString() {
      return _initialExpression;
   }
}
