/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.slot;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.Line2D;
import java.awt.geom.Path2D;
import java.util.ArrayList;
import java.util.List;

/**
 * A <code>Link</code> is a connection between two slots.
 * It is represented as a polyline, namely a series of straight lines.
 */
public abstract class Link {

   private static final BasicStroke REGULAR_LINE_STROKE = new BasicStroke(2.0F, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
   private static final BasicStroke SELECTED_LINE_STRKE = new BasicStroke(4.0F, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
   private static final int MIN_DISTANCE_BETWEEN_LINK_AND_SLOT = 10;
   protected static final int GAP = 10;

   private final InputSlot _targetSlot;
   private Line2D.Float[] _polyline;
   private int _selectedLinkPartNumber = -1;
   private boolean _userDefined;
   private boolean _selected;

   protected Link( InputSlot slot ) {
      _targetSlot = slot;
   }

   protected final void dispose() {
      _targetSlot.getView().setSelected(false);
   }

   protected final Line2D.Float[] getPolyline() {
      return _polyline.clone();
   }

   protected final void setPath( List<Point> path, boolean isUserDefined ) {
      if (path.isEmpty()) {
         _userDefined = false;
      } else {
         _userDefined = isUserDefined;
         buildPolyline(path);
      }
   }

   /**
    * Returns the current link as a series of intermediate points if this line is reshaped by the user. If the link is the default link, namely not reshaped by the user, the method returns an empty array of size 0.
    *
    * @return the current link as a series of intermediate points.
    */
   protected final List<Point> getPath() {
      ArrayList<Point> path = new ArrayList<>();
      if (_userDefined) {
         for (Line2D.Float line : _polyline) {
            path.add(new Point((int) line.x1, (int) line.y1));
         }
         path.add(new Point((int) _polyline[_polyline.length - 1].x2, (int) _polyline[_polyline.length - 1].y2));
      }
      return path;
   }

   protected final boolean isUserDefined() {
      return _userDefined;
   }

   protected final void setUserDefined( boolean value ) {
      _userDefined = value;
   }

   private boolean hasMoved() {
      int dx1 = (int) (_targetSlot.getConnectedSlot().getView().getAnchor().getX() - _targetSlot.getView().getAnchor().getX());
      int dx2 = (int) (_polyline[0].x1 - _polyline[_polyline.length - 1].x2);
      int dy1 = (int) (_targetSlot.getConnectedSlot().getView().getAnchor().getY() - _targetSlot.getView().getAnchor().getY());
      int dy2 = (int) (_polyline[0].y1 - _polyline[_polyline.length - 1].y2);
      return dx1 == dx2 && dy1 == dy2;
   }

   private boolean isModified() {
      return _polyline[0].x1 != _targetSlot.getConnectedSlot().getView().getAnchor().getX()
          || _polyline[0].y1 != _targetSlot.getConnectedSlot().getView().getAnchor().getY()
          || _polyline[_polyline.length - 1].x2 != _targetSlot.getView().getAnchor().getX()
          || _polyline[_polyline.length - 1].y2 != _targetSlot.getView().getAnchor().getY();
   }

   private void movePolyline() {
      final double dx = _targetSlot.getConnectedSlot().getView().getAnchor().getX() - _polyline[0].x1;
      final double dy = _targetSlot.getConnectedSlot().getView().getAnchor().getY() - _polyline[0].y1;
      for (Line2D.Float line : _polyline) {
         line.x1 += dx;
         line.x2 += dx;
         line.y1 += dy;
         line.y2 += dy;
      }
   }

   protected abstract boolean modifyPolyline();

   private void buildPolyline() {
      if (_polyline == null) {
         buildPolyline(createPolyline());
      } else if (isModified()) {
         if (hasMoved()) {
            movePolyline();
         } else if (!_userDefined || !modifyPolyline()) {
            buildPolyline(createPolyline());
            _userDefined = false;
         }
      }
   }

   private void buildPolyline( List<Point> path ) {
      _polyline = new Line2D.Float[path.size() - 1];
      // create the lines from the intermediate points
      for (int i = 0; i < path.size() - 1; i++) {
         final Point point1 = path.get(i);
         final Point point2 = path.get(i + 1);
         _polyline[i] = new Line2D.Float((float) point1.getX(), (float) point1.getY(), (float) point2.getX(), (float) point2.getY());
      }
   }

   protected abstract List<Point> createPolyline();

   private int getLineRankAtSpecifiedCoordinates( Point coordinates ) {
      int rank = 0;
      double distance = Double.MAX_VALUE;
      for (int i = 0; i < _polyline.length; i++) {
         if (_polyline[i].ptSegDist(coordinates.x, coordinates.y) < distance) {
            distance = _polyline[i].ptSegDist(coordinates.x, coordinates.y);
            rank = i;
         }
      }
      return rank;
   }

   public final boolean closestDistanceToLink( double x, double y ) {
      if (_polyline != null) {
         Shape shape = buildLinkShape();
         if (shape.intersects(x - MIN_DISTANCE_BETWEEN_LINK_AND_SLOT, y - MIN_DISTANCE_BETWEEN_LINK_AND_SLOT, 2 * MIN_DISTANCE_BETWEEN_LINK_AND_SLOT, 2 * MIN_DISTANCE_BETWEEN_LINK_AND_SLOT) && !shape.contains(x, y)) {
            return true;
         }
      }
      return false;
   }

   private Path2D buildLinkShape() {
      if (_polyline.length == 1) {
         Path2D line = new Path2D.Double();
         line.moveTo(_polyline[0].getP1().getX(), _polyline[0].getP1().getY());
         line.lineTo(_polyline[_polyline.length - 1].getP2().getX(), _polyline[_polyline.length - 1].getP2().getY());
         return line;
      } else if (_polyline.length == 3) {
         Path2D path = new Path2D.Double();
         path.moveTo(_polyline[0].getP1().getX(), _polyline[0].getP1().getY());
         path.curveTo(_polyline[1].getP1().getX(), _polyline[1].getP1().getY(),
             _polyline[2].getP1().getX(), _polyline[2].getP1().getY(),
             _polyline[2].getP2().getX(), _polyline[2].getP2().getY());
         return path;
      } else if (_polyline.length == 5) {
         Path2D path = new Path2D.Double();
         path.moveTo(_polyline[0].getP1().getX(), _polyline[0].getP1().getY());
         path.curveTo(_polyline[1].getP1().getX(), _polyline[1].getP1().getY(),
             _polyline[1].getP2().getX(), _polyline[1].getP2().getY(),
             (_polyline[2].getP1().getX() + _polyline[2].getP2().getX()) / 2, (_polyline[2].getP1().getY() + _polyline[2].getP2().getY()) / 2);
         path.curveTo(_polyline[3].getP1().getX(), _polyline[3].getP1().getY(),
             _polyline[3].getP2().getX(), _polyline[3].getP2().getY(),
             _polyline[4].getP2().getX(), _polyline[4].getP2().getY());
         return path;
      } else {
         Path2D path = new Path2D.Double();
         path.moveTo(_polyline[0].getP1().getX(), _polyline[0].getP1().getY());
         path.curveTo((_polyline[1].getP1().getX() + _polyline[1].getP2().getX()) / 2, (_polyline[1].getP1().getY() + _polyline[1].getP2().getY()) / 2,
             (_polyline[2].getP1().getX() + _polyline[2].getP2().getX()) / 2, (_polyline[2].getP1().getY() + _polyline[2].getP2().getY()) / 2,
             _polyline[3].getP2().getX(), _polyline[3].getP2().getY());
         return path;
      }
   }

   public final void draw( Graphics gc, Color color, boolean isSelected ) {
      Graphics2D gc2D = (Graphics2D) gc;

      // store old stroke.
      final Stroke oldStroke = gc2D.getStroke();
      gc2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
      gc2D.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
      gc2D.setStroke(isSelected || _selected ? SELECTED_LINE_STRKE : REGULAR_LINE_STROKE);
      gc2D.setPaint(color);

      buildPolyline();
      Shape shape = buildLinkShape();
      gc2D.draw(shape);
      gc2D.setStroke(oldStroke);
   }

   /**
    * Returns the cursor that corresponds to that the only allowed direction of link reshape (vertically or horizontally) depending of the current selected line in the polyline.
    *
    * @return the convenient cursor.
    */
   public final Cursor getCursor() {
      if (_selectedLinkPartNumber > 0) {
         if (_polyline[_selectedLinkPartNumber].x1 == _polyline[_selectedLinkPartNumber].x2) {
            return Cursor.getPredefinedCursor(Cursor.W_RESIZE_CURSOR);
         } else {
            return Cursor.getPredefinedCursor(Cursor.N_RESIZE_CURSOR);
         }
      }
      return null;
   }

   public final void setSelected( boolean selected, Point xycoord ) {
      if (selected) {
         _selectedLinkPartNumber = getLineRankAtSpecifiedCoordinates(xycoord);
         if (_selectedLinkPartNumber == 0) {
            _selectedLinkPartNumber = 1;
         } else if (_selectedLinkPartNumber == _polyline.length - 1) {
            _selectedLinkPartNumber = _polyline.length - 2;
         }
      } else {
         _selectedLinkPartNumber = -1;
      }
      _selected = selected;
      _targetSlot.getView().setSelected(selected);
   }

   public abstract void move( Point xycoord );

   public final int getSelectedLine() {
      return _selectedLinkPartNumber;
   }

   public final void move( int lineRank, Point xycoord ) {
      _selectedLinkPartNumber = lineRank;
      move(xycoord);
      _selectedLinkPartNumber = -1;
   }

   protected final Point getSourceCoordinate() {
      return _targetSlot.getConnectedSlot().getView().getAnchor();
   }

   protected final Point getTargetCoordinate() {
      return _targetSlot.getView().getAnchor();
   }
}
