/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.slot;

import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.slot.SlotView;

import java.awt.Point;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * The class <code>Slot</code> defines connection slot between 2 graph items.
 * The two derived classes <code>InputSlot</code> and <code>OutputSlot</code>
 * distinguish the input from output slots.
 * A slot is graphically represented an icon (using JLabel).
 * A slot is both a draggable and a droppable component.
 */
public abstract class Slot {

   private final transient SlotView _slotView;
   private transient GraphItem _graphItem;
   private final int _rank;

   public Slot( SlotView slotView, GraphItem item, int rank ) {
      _slotView = slotView;
      _rank = rank;
      _graphItem = item;
      slotView.setModel(this);
   }

   public final SlotView getView() {
      return _slotView;
   }

   /**
    * Disposes the slot (definitively). Tries to avoid memory leak.
    * <b>IT IS SUPPOSED TO BE DISCONNECTED!</b>
    */
   public final void dispose() {
      _graphItem = null;
   }

   public final int getRank() {
      return _rank;
   }

   public final GraphItem getGraphItem() {
      return _graphItem;
   }

   public final Graph getGraph() {
      return _graphItem.getGraph();
   }

   public abstract boolean isConnected();

   public abstract void disconnectAll();

   public abstract List<? extends Slot> getConnectedSlots();

   public abstract void connectWith( Slot slot );

   public abstract boolean isAcceptableDropSite( Slot slot );

   private static void connectCompatibleSlots( GraphItem gFrom, GraphItem gTo, SlotAttributes attributes ) {
      if (gFrom != null && gTo != null) {
         OutputSlot sFrom = (OutputSlot) gFrom.getSlot(attributes.typeFrom, attributes.rankFrom);
         InputSlot sTo = (InputSlot) gTo.getSlot(attributes.typeTo, attributes.rankTo);
         if (sFrom != null && sTo != null) {
            sFrom.connectTo(sTo, null, true);
            Link link = sTo.createLink();
            link.setPath(attributes.path, attributes.userDefined);
            sTo.setLink(link);
         }
      }
   }

   public static final void createFromXML( Graph graph, List<XMLElement> items ) throws IOException {
      if (items == null) {
         return;
      }
      for (XMLElement element : items) {
         if ("connection".equalsIgnoreCase(element.getName()) || "connexion".equalsIgnoreCase(element.getName())) {
            SlotAttributes attributes = new SlotAttributes();

            for (XMLElement elc : element.getChildren()) {
               String tag = elc.getName().toLowerCase();
               if ("enditemid".equals(tag) || "idelementarrivee".equals(tag)) {
                  attributes.idTo = Integer.parseInt(elc.getValue());
               } else if ("startitemid".equals(tag) || "idelementdepart".equals(tag)) {
                  attributes.idFrom = Integer.parseInt(elc.getValue());
               } else if ("endslottype".equals(tag) || "typeslotarrivee".equals(tag)) {
                  attributes.typeTo = elc.getValue();
               } else if ("startslottype".equals(tag) || "typeslotdepart".equals(tag)) {
                  attributes.typeFrom = elc.getValue();
               } else if ("startslotindex".equals(tag) || "numeroslotdepart".equals(tag)) {
                  attributes.rankFrom = Integer.parseInt(elc.getValue());
               } else if ("endslotindex".equals(tag) || "numeroslotarrivee".equals(tag)) {
                  attributes.rankTo = Integer.parseInt(elc.getValue());
               } else if ("parametervalue".equals(tag)) {
                  attributes.parameterValue = elc.getValue();
               } else if ("userdefined".equals(tag)) {
                  attributes.userDefined = "true".equalsIgnoreCase(elc.getValue());
               } else if ("linkpath".equals(tag) && elc.getChildren() != null) {
                  createLinkPathFromXML(elc, attributes);
               }
            }
            if (attributes.typeTo == null) {
               throw new IOException("Unknown connection type");
            }
            GraphItem gTo = graph.getGraphItem(attributes.idTo);
            if (attributes.idFrom > -1) { // Connection parameter -> result
               GraphItem gFrom = graph.getGraphItem(attributes.idFrom);
               connectCompatibleSlots(gFrom, gTo, attributes);
            }
            if ("inslotparameter".equalsIgnoreCase(attributes.typeTo) && gTo != null) { // At least one Parameter value
               InputParameterSlot par = gTo.getInputParameterSlots().get(attributes.rankTo);
               if (par != null) {
                  par.setValue(attributes.parameterValue);
               }
            }
         }
      }
   }

   private static void createLinkPathFromXML( XMLElement element, SlotAttributes attributes ) {
      for (XMLElement tag : element.getChildren()) {
         if ("linkpoint".equalsIgnoreCase(tag.getName())) {
            final Point point = new Point();
            int pointId = 0;
            for (XMLElement pointElement : tag.getChildren()) {
               String tagName = pointElement.getName();
               if ("idpoint".equalsIgnoreCase(tagName)) {
                  pointId = Integer.parseInt(pointElement.getValue());
               } else if ("x-coordinate".equalsIgnoreCase(tagName)) {
                  point.x = Integer.parseInt(pointElement.getValue());
               } else if ("y-coordinate".equalsIgnoreCase(tagName)) {
                  point.y = Integer.parseInt(pointElement.getValue());
               }
            }
            if (pointId > 0) {
               attributes.path.add(pointId, point);
            } else {
               attributes.path.add(point);
            }
         }
      }
   }

   private static class SlotAttributes { // A messenger
      //CHECKSTYLE:OFF

      public int idTo = -1;
      public int idFrom = -1;
      public int rankFrom = -1;
      public int rankTo = -1;
      public String typeTo;
      public String typeFrom;
      public String parameterValue;
      public final List<Point> path = new ArrayList<>();
      public boolean userDefined;
      //CHECKSTYLE:ON
   }
}
