/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 * 
 * https://clouard.users.greyc.fr
 * 
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.tree;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;

import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

/**
 * The class <code>Tree</code> contains all the available graph items organized in a tree. This tree is displayed in the explorer from where the user can drag and drop items into the graph editor.
 */
public final class ItemTreeModel extends DefaultTreeModel {

    private final MutableTreeItem _builtinSubtree = new MutableTreeItem(null, true);
    private final MutableTreeItem _commonOperatorSubtree = new MutableTreeItem(null, true);
    private MutableTreeItem _userOperatorSubtree = new MutableTreeItem(null, true);
    private MutableTreeItem _subRoutineSubtree = new MutableTreeItem(null, true);
    private String _pathToHTMLLibraryDocumentationFile;

    public ItemTreeModel() {
        super(new MutableTreeItem("Library", true));
        initContents();
    }

    private void initContents() {
        createBuiltinItems();
        MutableTreeItem rootNode = (MutableTreeItem) getRoot();
        rootNode.add(_builtinSubtree);
        rootNode.add(_commonOperatorSubtree);
        rootNode.add(_subRoutineSubtree);
        rootNode.add(_userOperatorSubtree);
    }

    private void createBuiltinItems() {
        for (TreeItem.Item itemType : TreeItem.Item.values()) {
            if (itemType != TreeItem.Item.OPERATOR) {
                TreeItem treeItem = getItem(itemType);
                MutableTreeItem fileItem = new MutableTreeItem(treeItem, false);
                _builtinSubtree.add(fileItem);
            }
        }
    }

    public static TreeItem getItem( TreeItem.Item itemType ) {
        switch (itemType) {
            case LOADER:
                return new OperatorTreeItem(TreeItem.Item.LOADER, "loader", "loader.html", 0, 0, 1, "");
            case SAVER:
                return new OperatorTreeItem(TreeItem.Item.SAVER, "saver", "saver.html", 1, 0, 0, "");
            case VIEWER:
                return new OperatorTreeItem(TreeItem.Item.VIEWER, "viewer", "viewer.html", 1, 0, 0, "");
            case FILENAME:
                return new OperatorTreeItem(TreeItem.Item.FILENAME, "filename", "filename.html", 0, 0, 0, "");
            case CONSTANT:
                return new OperatorTreeItem(TreeItem.Item.CONSTANT, "constant", "constant.html", 0, 0, 0, "");
            case SCRIPT:
                return new OperatorTreeItem(TreeItem.Item.SCRIPT, "script", "script.html", 0, 0, 0, "");
            case EVAL:
                return new OperatorTreeItem(TreeItem.Item.EVAL, "eval", "eval.html", 0, 2, 0, "");
            case FORMAT:
                return new OperatorTreeItem(TreeItem.Item.FORMAT, "format", "format.html", 0, 2, 0, "");
            case SYSTEM:
                return new OperatorTreeItem(TreeItem.Item.SYSTEM, "system", "system.html", 0, 0, 0, "");
            case IF_CONTROL:
                return new OperatorTreeItem(TreeItem.Item.IF_CONTROL, "if", "if.html", 1, 2, 2, "");
            case SELECT_CONTROL:
                return new OperatorTreeItem(TreeItem.Item.SELECT_CONTROL, "select", "select.html", 0, 0, 1, "$0:Index");
            case MERGE_CONTROL:
                return new OperatorTreeItem(TreeItem.Item.MERGE_CONTROL, "merge", "merge.html", 2, 0, 1, "");
            case FOR_CONTROL:
                return new OperatorTreeItem(TreeItem.Item.FOR_CONTROL, "for", "for.html", 3, 0, 0, "$0:Initial value;$1:Final value;$2:Increment");
            case WHILE_CONTROL:
                return new OperatorTreeItem(TreeItem.Item.WHILE_CONTROL, "while", "while.html", 2, 0, 0, "");
            case MACRO_OPERATOR:
                return new OperatorTreeItem(TreeItem.Item.MACRO_OPERATOR, "macro-operator", "macro.html", 0, 0, 0, "");
            case NOTE:
                return new OperatorTreeItem(TreeItem.Item.NOTE, "note", "note.html", 0, 0, 0, "");
            default:
                return null;
        }
    }

    public TreeItem getTreeItem( String name, String category ) {
        return getNode((MutableTreeItem) getRoot(), name, category);
    }

    private TreeItem getNode( MutableTreeItem subtree, String name, String category ) {
        for (final Enumeration<?> e = subtree.children(); e.hasMoreElements();) {
            MutableTreeItem node = (MutableTreeItem) e.nextElement();
            if (node.isLeaf()) {
                TreeItem treeItem = node.getTreeItemObject();
                if (name.equals(treeItem.getName()) && (category == null || category.equals(treeItem.getCategory()))) {
                    return treeItem;
                }
            } else {
                TreeItem item = getNode(node, name, category);
                if (item != null) {
                    return item;
                }
            }
        }
        return null;
    }

    public void setInternationalText() {
        _subRoutineSubtree.setUserObject(ArianeSettings.getString("Tree.subroutines"));
        _userOperatorSubtree.setUserObject(ArianeSettings.getString("Tree.personalOperators"));
        _commonOperatorSubtree.setUserObject(ArianeSettings.getString("Tree.operators"));
        _builtinSubtree.setUserObject(ArianeSettings.getString("Tree.builtinItems"));
        reload();
    }

    public String getPathToLibraryDocumentationFile() {
        return _pathToHTMLLibraryDocumentationFile;
    }

    private List<TreeItem> subtreeToList( MutableTreeItem subtree ) {
        ArrayList<TreeItem> treeList = new ArrayList<>();
        Enumeration<?> iterator = subtree.depthFirstEnumeration();
        while (iterator.hasMoreElements()) {
            MutableTreeItem node = (MutableTreeItem) iterator.nextElement();
            if (node.getUserObject() instanceof TreeItem) {
                treeList.add(node.getTreeItemObject());
            }
        }
        return treeList;
    }

    public List<TreeItem> getUserOperatorTreeAsList() {
        return subtreeToList(_userOperatorSubtree);
    }

    public List<TreeItem> getSubroutineTreeAsList() {
        return subtreeToList(_subRoutineSubtree);
    }

    public MutableTreeItem getUserOperatorTree() {
        return _userOperatorSubtree;
    }

    public MutableTreeItem getSubroutineTree() {
        return _subRoutineSubtree;
    }

    private MutableTreeItem getCommonOperatorTree() {
        return _commonOperatorSubtree;
    }

    private MutableTreeItem changeSubtree( MutableTreeItem oldSubtree, MutableTreeItem newSubtree ) {
        MutableTreeItem rootNode = (MutableTreeItem) getRoot();
        int index = getIndexOfChild(rootNode, oldSubtree);
        removeNodeFromParent(oldSubtree);
        insertNodeInto(newSubtree, rootNode, index);
        return newSubtree;
    }

    public void setUserOperatorTree( MutableTreeItem newTree ) {
        _userOperatorSubtree = changeSubtree(_userOperatorSubtree, newTree);
    }

    public void setSubroutineTree( MutableTreeItem newTree ) {
        _subRoutineSubtree = changeSubtree(_subRoutineSubtree, newTree);
    }

    public void setCommonOperators( List<XMLElement> readItems ) {
        addInCommonOperators(loadItemsFromXMLTree(readItems));
    }

    public void addInCommonOperators( List<TreeItem> operatorList ) {
        _commonOperatorSubtree.removeAllChildren();
        addItemsInSubtree(operatorList, _commonOperatorSubtree);
    }

    public void addInUserOperators( List<XMLElement> operatorList ) {
        addItemsInSubtree(loadItemsFromXMLTree(operatorList), _userOperatorSubtree);
    }

    public MutableTreeItem addInSubroutines( List<XMLElement> routinesList ) {
        addItemsInSubtree(loadItemsFromXMLTree(routinesList), _subRoutineSubtree);
        return _subRoutineSubtree;
    }

    public void addSubroutine( TreeItem item ) {
        ArrayList<TreeItem> treeItems = new ArrayList<>();
        treeItems.add(item);
        addItemsInSubtree(treeItems, _subRoutineSubtree);
    }

    public String[] getSubroutineCategories() {
        String[] categories = new String[_subRoutineSubtree.getChildCount()];
        for (int i = 0; i < _subRoutineSubtree.getChildCount(); i++) {
            categories[i] = _subRoutineSubtree.getChildAt(i).toString();
        }
        return categories;
    }

    private void addItemsInSubtree( List<TreeItem> itemList, MutableTreeItem subtree ) {
        for (TreeItem treeItem : itemList) {
            String categoryName = treeItem.getCategory();
            MutableTreeItem category = subtree.getNodeWithSpecifiedName(categoryName);

            if (category == null) { // create a new category.
                category = new MutableTreeItem(categoryName, true);
                subtree.add(category);
            } else if (category.getNodeWithSpecifiedName(treeItem.getName()) != null) {
                category.remove(category.getNodeWithSpecifiedName(treeItem.getName()));
            }
            MutableTreeItem item = new MutableTreeItem(treeItem, false);
            category.add(item); // replace existing subroutine
        }
        reload(subtree);
    }

    public List<TreePath> getItemsWithTheSpecifiedNameInTree( final String name ) {
        return getItemsWithTheSpecifiedNameInSpecifiedSubtree((MutableTreeItem) getRoot(), name);
    }

    public List<TreePath> getItemsWithTheSpecifiedNameInSpecifiedSubtree( MutableTreeItem subtree, String name ) {
        ArrayList<TreePath> paths = new ArrayList<>();
        for (final Enumeration<?> e = subtree.children(); e.hasMoreElements();) {
            MutableTreeItem node = (MutableTreeItem) e.nextElement();
            if (node.isLeaf()) {
                String operatorName = node.getTreeItemName();
                if (operatorName.contains(name)) {
                    paths.add(new TreePath(node.getPath()));
                }
            } else {
                paths.addAll(getItemsWithTheSpecifiedNameInSpecifiedSubtree(node, name));
            }
        }
        return paths;
    }

    public TreeItem getItemEqualsTo( String name ) {
        for (MutableTreeItem subtree : getSubTrees()) {
            List<TreePath> paths = getItemsWithTheSpecifiedNameInSpecifiedSubtree(subtree, name);
            for (TreePath treePath : paths) {
                MutableTreeItem leaf = (MutableTreeItem) treePath.getLastPathComponent();
                TreeItem item = leaf.getTreeItemObject();
                // Ensure that the operator BEGINS with the pattern.
                if (item.getName().equals(name)) {
                    return leaf.getTreeItemObject();
                }
            }
        }
        return null;
    }

    /**
     * The list of all operators that can be added to the graph.
     */
    private MutableTreeItem[] getSubTrees() {
        return new MutableTreeItem[]{getCommonOperatorTree(),
            getUserOperatorTree(),
            getSubroutineTree()};
    }

    public List<String> getItemsWithNameStartingWith( String pattern ) {
        List<String> result = new ArrayList<>();
        for (MutableTreeItem subtree : getSubTrees()) {
            for (TreePath path : getItemsWithTheSpecifiedNameInSpecifiedSubtree(subtree, pattern)) {
                MutableTreeItem leaf = (MutableTreeItem) path.getLastPathComponent();
                TreeItem item = leaf.getTreeItemObject();
                // Ensure that the operator BEGINS with the pattern.
                if (item.getName().startsWith(pattern)) {
                    result.add(item.getName());
                }
            }
        }
        return result;
    }

    ///// 
    ///// I N P U T   /   O U T P U T
    /////
    protected List<TreeItem> loadItemsFromXMLTree( List<XMLElement> xmlTree ) {
        List<TreeItem> treeItems = new ArrayList<>();
        if (xmlTree.isEmpty()) {
            return treeItems;
        }
        // Discard tag <ARIANE>.
        List<XMLElement> xmlContents = xmlTree.get(0).getChildren();
        if (xmlContents == null) {
            return treeItems;
        }
        for (XMLElement element : xmlContents) {
            if ("OperatorLibrary".equalsIgnoreCase(element.getName())
                    || "Library".equalsIgnoreCase(element.getName())
                    || "Bibliotheque".equalsIgnoreCase(element.getName())) {
                for (XMLElement item : element.getChildren()) {
                    String tag = item.getName();
                    if ("DocPath".equalsIgnoreCase(tag)) {
                        ArianeSettings.setLibraryDocumentationPath(item.getValue());
                    } else if ("CommandStatus".equalsIgnoreCase(tag)) {
                        UserPreferences.setCommandStatus(item.getValue());
                    } else if ("CommandVisu".equalsIgnoreCase(tag)) {
                        UserPreferences.setCommandVisu(item.getValue());
                    } else if ("Version".equalsIgnoreCase(tag)) {
                        // kept for sake of compatibility
                        // Gets the version number of library.
                        ((MutableTreeItem) getRoot()).setUserObject(item.getValue());
                        reload();
                        if (item.getValue().startsWith("Pandore")) {
                            UserPreferences.setCommandStatus("pstatus");
                            UserPreferences.setCommandVisu("pvisu");
                        }
                    } else if ("Index".equalsIgnoreCase(tag)) {
                        // kept for sake of compatibility
                        _pathToHTMLLibraryDocumentationFile = item.getValue();
                    } else if ("Operator".equalsIgnoreCase(tag) || "Operateur".equalsIgnoreCase(tag)) {
                        treeItems.add(new OperatorTreeItem(item.getChildren()));
                    }
                }
                return treeItems;
            } else if ("SubroutineLibrary".equalsIgnoreCase(element.getName())) {
                for (XMLElement item : element.getChildren()) {
                    if ("Subroutine".equalsIgnoreCase(item.getName())) {
                        treeItems.add(new SubroutineTreeItem(item.getChildren()));
                    }
                }
                return treeItems;
            }
        }
        return treeItems;
    }
}
