/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 * 
 * https://clouard.users.greyc.fr
 * 
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.tree;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;
import java.util.List;

import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.graph.item.DataFile;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;

/**
 * A <code>TreeItem</code> is an item that can be inserted in the tree of the explorer.
 */
public abstract class TreeItem implements Transferable {

    private Item _typeItem;

    private String _name;

    private String _documentationURL;

    private String _binaryPath;

    private String _categoryName;

    private String _signature;

    private int _numberOfInputDataSlots;

    private int _numberOfInputParameterSlots;

    private int _numberOfOutputDataSlots;

    /**
     * The item magic numbers.
     */
    public enum Item {
        /**
         * Operator.
         */
        OPERATOR,
        /**
         * A file loader.
         */
        LOADER,
        /**
         * A file saver.
         */
        SAVER,
        /**
         * A file viewer.
         */
        VIEWER,
        /**
         * A file name.
         */
        FILENAME,
        /**
         * A constant.
         */
        CONSTANT,
        /**
         * A macro operator.
         */
        MACRO_OPERATOR,
        /**
         * An if control structure.
         */
        IF_CONTROL,
        /**
         * An input selector.
         */
        SELECT_CONTROL,
        /**
         * An iterative control structure.
         */
        FOR_CONTROL,
        /**
         * A repetitive control structure.
         */
        WHILE_CONTROL,
        /**
         * Merge two inputs into one output.
         */
        MERGE_CONTROL,
        /**
         * An calculus unit.
         */
        EVAL,
        /**
         * A string manipulator unit.
         */
        FORMAT,
        /**
         * A comment.
         */
        NOTE,
        /**
         * Execute an operating system command.
         */
        SYSTEM,
        /**
         * A script.
         */
        SCRIPT,
    }

    public TreeItem( Item typeItem, String name, String docURL, int inputDataSlots, int inputParameterSlots, int outputDataSlots, String signature ) {
        _typeItem = typeItem;
        _name = name;
        _documentationURL = docURL;
        _numberOfInputParameterSlots = inputParameterSlots;
        _numberOfInputDataSlots = inputDataSlots;
        _numberOfOutputDataSlots = outputDataSlots;
        _signature = signature;
    }

    public TreeItem( List<XMLElement> xmlTree ) {
        loadItemFromXMLTree(xmlTree);
    }

    public final String getName() {
        return _name;
    }

    public final void setName( String text ) {
        _name = text;
    }

    protected final void setType( Item type ) {
        _typeItem = type;
    }

    protected final Item getType() {
        return _typeItem;
    }

    public final String getDocumentationURL() {
        return _documentationURL;
    }

    public final void setDocumentationURL( String url ) {
        _documentationURL = url;
    }

    public final String getCategory() {
        return _categoryName;
    }

    public final void setCategory( String categoryNname ) {
        _categoryName = categoryNname;
    }

    public final String getBinaryPath() {
        return _binaryPath;
    }

    public final void setBinaryPath( String path ) {
        String preferredDirectory = UserPreferences.getPreferredBinaryDirectory();
        if (preferredDirectory != null && preferredDirectory.trim().isEmpty()) {
            String[] commandLine = path.split(OperatorItem.SEPARATOR);
            String binary = DataFile.getRelativePath(commandLine[0], preferredDirectory, UserPreferences.getPreferredFileDirectory());
            if (commandLine.length > 1) {
                _binaryPath = binary + OperatorItem.SEPARATOR + commandLine[1];
            } else {
                _binaryPath = binary + OperatorItem.SEPARATOR;
            }
        } else {
            _binaryPath = path;
        }
    }

    public final int getNumberOfParameterSlots() {
        return _numberOfInputParameterSlots;
    }

    public final void setNumberOfParameterSlots( int count ) {
        _numberOfInputParameterSlots = count;
    }

    public final String getSignature() {
        return _signature;
    }

    /**
     * Sets the default values for the parameters.
     *
     * @param signature (eg. $p0:toto=0; $p1:titi=1;).
     */
    public final void setSignature( String signature ) {
        _signature = signature;
    }

    public final int getNumberOfInputDataSlots() {
        return _numberOfInputDataSlots;
    }

    public final void setNumberOfInputDataSlots( int count ) {
        _numberOfInputDataSlots = count;
    }

    public final int getNumberOfOutputDataSlots() {
        return _numberOfOutputDataSlots;
    }

    public final void setNumberOfOutputDataSlots( int count ) {
        _numberOfOutputDataSlots = count;
    }

    /**
     * Returns the name of the item. Warning! this method is used to display the name in the tree.
     *
     * @return the description of the node.
     */
    @Override
    public final String toString() {
        return _name;
    }

    @Override
    public abstract DataFlavor[] getTransferDataFlavors();

    @Override
    public abstract Object getTransferData( DataFlavor flavor ) throws UnsupportedFlavorException, IOException;

    @Override
    public abstract boolean isDataFlavorSupported( DataFlavor flavor );

    public abstract GraphItem createGraphItem( Graph graph );

    ///// 
    ///// I N P U T  /  O U T P U T
    /////
    public abstract XMLElement convertToXMLTree() throws IOException;

    protected abstract void loadItemFromXMLTree( List<XMLElement> xmlelements );
}
