/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.presenter;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JViewport;
import javax.swing.undo.UndoableEdit;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.graph.GraphModel;
import fr.greyc.pantheon.ariane.model.graph.ThreadDirectoryExecution;
import fr.greyc.pantheon.ariane.model.graph.ThreadGraphExecution;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.model.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.Link;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.Slot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.command.AddItemEdit;
import fr.greyc.pantheon.ariane.presenter.command.MoveItemEdit;
import fr.greyc.pantheon.ariane.presenter.command.MoveLinkEdit;
import fr.greyc.pantheon.ariane.presenter.command.RemoveItemEdit;
import fr.greyc.pantheon.ariane.view.graph.GraphView;
import fr.greyc.pantheon.ariane.view.graph.XYConstraint;

public abstract class Graph {

   private final GraphView _graphView;
   private final GraphModel _graphModel;

   private boolean _isAlive = true;
   private String _name;

   private static boolean _isRunning;
   private Process _currentProcess;

   private final List<GraphItem> _selectedItems = new LinkedList<>();

   private DraggingSelection _draggingSelection;

   private final UndoRedoManager _undoRedoManager = new UndoRedoManager();

   public Graph() {
      this(new GraphModel());
   }

   public Graph( File file ) {
      this(new GraphModel(file.getAbsoluteFile()));
   }

   protected Graph( GraphModel graphModel ) {
      _graphModel = graphModel;
      _graphView = new GraphView(this, _graphModel);
   }

   public final void dispose() {
      _isAlive = false;
      _graphModel.dispose();
      _undoRedoManager.discardAllEdits();
      unSelectAll();
   }

   /////
   ///// F R A M E S
   /////
   public final void closeFrame() {
      for (GraphItem item : _graphModel.getListOfGraphItems()) {
         if (item.getSubGraph() != null) {
            item.getSubGraph().closeFrame();
         }
      }
      Ariane.getPresenter().removeGraph(this);
   }

   protected final void renameFrame() {
      for (GraphItem item : _graphModel.getListOfGraphItems()) {
         if (item.getSubGraph() != null) {
            item.getSubGraph().renameFrame();
         }
      }
      Ariane.getPresenter().renameGraph(this, getName());
   }

   /////
   ///// A C C E S S O R S
   /////
   public final GraphView getView() {
      return _graphView;
   }

   public final int getIdentifier() {
      return _graphModel.getIdentifier();
   }

   public final void setName( String name ) {
      _name = name;
   }

   public final String getName() {
      return _name;
   }

   public abstract boolean close();

   protected final boolean isCloseConfirmed() {
      return _graphView.isCloseConfirmed();
   }

   protected final boolean isAlive() {
      return _graphView.getParent() != null && _isAlive;
   }

   public abstract GraphItem getSuperGraphItem();

   public final String getDirectoryName() {
      File file = getRoot()._graphModel.getStorageFile();
      if (file != null) {
         return file.getParent();
      }
      return null;
   }

   public final boolean isAssociatedToFile( File file ) {
      File currentFile = getRoot()._graphModel.getStorageFile();
      if (currentFile != null) {
         try {
            return file.getCanonicalPath().equals(currentFile.getCanonicalPath());
         } catch (IOException e) {
            return false;
         }
      }
      return false;
   }

   public final int getNextGraphItemIdentifier() {
      return _graphModel.getNextItemIdentifier();
   }

   public abstract boolean isModified();

   public abstract void setModified( boolean value );

   public abstract boolean isRoot();

   public abstract Graph getRoot();

   public final boolean isSelectedItemsEmpty() {
      return _selectedItems.isEmpty();
   }

   public final List<GraphItem> getListOfGraphItems() {
      return _graphModel.getListOfGraphItems();
   }

   public final GraphItem getGraphItem( int identifier ) {
      return _graphModel.getGraphItem(identifier);
   }

   /////
   ///// A C T I O N S
   /////
   public final void disconnectSlot( Slot incoming ) {
      startActionEdition();
      incoming.disconnectAll();
      endActionEdition();
   }

   public final void addItemAtRelativePosition( TreeItem incoming, Point coordinates ) {
      Point viewPosition = ((JViewport) _graphView.getParent()).getViewPosition();
      Point absolutePosition = new Point(viewPosition.x + coordinates.x, viewPosition.y + coordinates.y);
      Graph.this.addItemAtAbsolutePosition(incoming, absolutePosition);
   }

   public final void addItemAtAbsolutePosition( TreeItem incoming, Point coordinates ) {
      GraphItem newItem = incoming.createGraphItem(this);
      if (newItem != null) {
         startActionEdition();
         addItemAtAbsolutePosition(newItem, coordinates, true);
         endActionEdition();
         unSelectAll();
         addInSelection(newItem, false);
         newItem.getView().requestFocusInWindow();
      }
   }

   public final void addItemAtAbsolutePosition( GraphItem item, Point xycoord, boolean addInHistory ) {
      _graphModel.add(item);
      item.getView().draw();
      _graphView.add(item.getView(), new XYConstraint(xycoord));
      _graphView.setComponentZOrder(item.getView(), 0); // Put it in front
      _graphView.scrollRectToVisible(item.getView().getBounds());
      if (addInHistory) {
         validateAction(new AddItemEdit(this, item));
      }
   }

   public final void removeItem( GraphItem item ) {
      startActionEdition();
      item.remove();
      removeItem(item, true);
      revalidate();
      endActionEdition();
   }

   public final void removeItem( GraphItem item, boolean addInHistory ) {
      _graphModel.remove(item);
      _graphView.remove(item.getView());
      item.remove();
      if (addInHistory) {
         validateAction(new RemoveItemEdit(this, item));
      }
   }

   public final void moveItem( GraphItem item, Point location, boolean addInHistory ) {
      if (addInHistory) {
         validateAction(new MoveItemEdit(this, item, item.getLocation(), location));
      }
      item.setLocation(location);
      _graphView.revalidate();
   }

   private void moveItems( List<GraphItem> litems, Point translation ) {
      for (GraphItem item : litems) {
         Point xy = item.getLocation();
         xy.translate(-translation.x, -translation.y);
         moveItem(item, xy, true);
      }
      _graphView.repaint();
   }

   private void moveItems( List<GraphItem> litems, Point from, Point to ) {
      final Point translation = new Point(from.x - to.x, from.y - to.y);
      for (GraphItem item : litems) {
         Point origin = new Point(item.getLocation().x + translation.x, item.getLocation().y + translation.y);
         validateAction(new MoveItemEdit(this, item, origin, item.getLocation()));
      }
      _graphView.repaint();
   }

   public final void moveLink( Link link, Point from, Point to ) {
      validateOneAction(new MoveLinkEdit(this, link, from, to));
   }

   /////
   ///// E X E C U T I O N
   /////
   public final void revalidate() {
      _graphView.revalidate();
      _graphView.repaint();
   }

   public final void resetAll() {
      _graphModel.resetAll();
   }

   public final void reset() {
      _graphModel.reset();
   }

   public final void resetSelectedItems() {
      for (GraphItem gi : _selectedItems) {
         gi.resetAll();
      }
   }

   public final void setCurrentProcess( Process process ) {
      _currentProcess = process;
   }

   public final synchronized void stopExecution() {
      _isRunning = false;
      Ariane.getPresenter().setExecution(false);
      if (_currentProcess != null) {
         _currentProcess.destroy();
      }
   }

   public static final boolean isExecutionRunning() {
      return _isRunning;
   }

   public final void startTestExecution() {
      _isRunning = true;
   }

   public final void startExecution() {
      if (_selectedItems.isEmpty()) {
         Ariane.getPresenter().printWarning(ArianeSettings.getString("Graph.execution.warning1"));
      } else {
         _isRunning = true;
         // Creates a thread and executes the graph with it.
         ThreadGraphExecution thread = new ThreadGraphExecution(_selectedItems);
         thread.start();
      }
   }

   public final void executeFolder( LoaderItem loader ) {
      if (loader.getBinary() == null) {
         Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Graph.execution.error1") + loader.getName() + "(" + loader.getId() + ")");
         return;
      }
      File loaderFile = new File(loader.getBinary());
      if (!loaderFile.exists()) {
         Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Graph.execution.error1") + loader.getName() + "(" + loader.getId() + ")");
         return;
      }

      List<GraphItem> terminalItems = getTerminalItems(loader);

      // Get the savers to store data in the appropriate files.
      HashMap<SaverItem, File> saverDirectories = new HashMap<>();
      for (GraphItem item : terminalItems) {
         if (item instanceof SaverItem) {
            File saverFile = new File(item.getBinary());
            if (saverFile.getParentFile() == null || !saverFile.getParentFile().exists()) {
               Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Graph.execution.error1") + item.getName() + "(" + item.getId() + ")");
               return;
            }
            saverDirectories.put((SaverItem) item, saverFile);
         }
      }

      int suffixPosition = loaderFile.getName().lastIndexOf('.');
      final String loaderSuffix = suffixPosition > 0 ? loaderFile.getName().substring(suffixPosition) : "";
      final File[] files = loaderFile.getParentFile().listFiles(new FilenameFilter() {
         @Override
         public boolean accept( File dir, String name ) {
            File file = new File(dir, name);
            if (file.isDirectory() || file.isHidden()) {
               return false;
            }
            String suffix = "";
            int suffixPosition = name.lastIndexOf('.');
            if (suffixPosition > 0) {
               suffix = name.substring(suffixPosition);
            }
            return suffix.equals(loaderSuffix);
         }
      });

      if (files == null) {
         Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Graph.execution.error2") + loaderFile.getParent());
         return;
      }
      Arrays.sort(files);

      _isRunning = true;
      ThreadDirectoryExecution thread = new ThreadDirectoryExecution(terminalItems, loader, loaderFile, files, saverDirectories);
      thread.start();
   }

   private List<GraphItem> getTerminalItems( LoaderItem loader ) {
      // Get the list of items to be executed starting from the loader.
      LinkedList<GraphItem> terminalItems = new LinkedList<>();
      LinkedList<GraphItem> visitedItems = new LinkedList<>();
      LinkedList<GraphItem> stack = new LinkedList<>();
      stack.add(loader);

      while (!stack.isEmpty()) {
         GraphItem item = stack.removeLast();
         if (visitedItems.contains(item)) {
            continue;
         }
         visitedItems.add(item);

         for (InputParameterSlot slot : item.getInputParameterSlots()) {
            if (slot.isConnected()) {
               GraphItem neighbour = slot.getConnectedSlot().getGraphItem();
               stack.add(neighbour);
            }
         }

         for (InputDataSlot slot : item.getInputDataSlots()) {
            if (slot.isConnected()) {
               GraphItem neighbour = slot.getConnectedSlot().getGraphItem();
               stack.add(neighbour);
            }
         }

         InputMaskSlot maskSlot = item.getInputMaskSlot();
         if (maskSlot != null && maskSlot.isConnected()) {
            GraphItem neighbour = maskSlot.getConnectedSlot().getGraphItem();
            stack.add(neighbour);
         }

         boolean lastItem = true;
         for (OutputDataSlot slot : item.getOutputDataSlots()) {
            if (slot.isConnected()) {
               lastItem = false;
               for (InputSlot neighbourSlot : slot.getConnectedInputSlots()) {
                  if (neighbourSlot.isConnected()) {
                     stack.add(neighbourSlot.getGraphItem());
                  }
               }
            }
         }

         if (item.getOutputResultSlot() != null) {
            OutputResultSlot slot = item.getOutputResultSlot();
            if (slot.isConnected()) {
               lastItem = false;
               for (InputSlot neighbourSlot : slot.getConnectedInputSlots()) {
                  if (neighbourSlot.isConnected()) {
                     stack.add(neighbourSlot.getGraphItem());
                  }
               }
            }
         }

         if (lastItem) {
            terminalItems.add(item);
         }
      }
      return terminalItems;
   }

   public final void displayOutputs( int output ) {
      for (GraphItem x : _selectedItems) {
         if (x.getOutputDataSlots().size() > output) {
            x.getOutputDataSlots().get(output).showDataFileContents();
         }
      }
   }

   public final void selectAll() {
      unSelectAll();
      for (GraphItem item : _graphModel.getListOfGraphItems()) {
         if (!(item instanceof VirtualItem)) {
            addInSelection(item, true);
         }
      }
   }

   public final void unSelectAll() {
      for (GraphItem gi : _selectedItems) {
         gi.setSelected(false);
      }
      _selectedItems.clear();
      _graphView.resetCurrentSelection();
   }

   private void selectItem( GraphItem item ) {
      unSelectAll();
      _selectedItems.add(item);
      item.setSelected(true);
   }

   /**
    * Adds the given item in the current selection according to the specified added value:
    * <ul>
    * <li>Shift pressed: adds the item to the current selection.
    * <li>Shift pressed and item already in the selection: removes it from the selection.
    * <li>Shift not pressed : builds a new selection with this item.
    * </ul>
    *
    * @param item the item to be added.
    * @param shiftPressed key shift pressed?
    */
   public final void addInSelection( GraphItem item, boolean shiftPressed ) {
      if (shiftPressed) {
         if (item.isSelected()) {
            _selectedItems.remove(item);
            item.setSelected(false);
         } else {
            item.setSelected(true);
            _selectedItems.add(item);
         }
      } else {
         selectItem(item);
      }
   }

   private Rectangle getRectangleAroundSelectedItems() {
      int xs = 100000;
      int ys = 100000;
      int xe = 0;
      int ye = 0;
      for (GraphItem item : _selectedItems) {
         Point itemcoord = item.getLocation();
         if (itemcoord.x < xs) {
            xs = itemcoord.x;
         }
         if (itemcoord.y < ys) {
            ys = itemcoord.y;
         }
         if (item.getView().getWidth() + itemcoord.x > xe) {
            xe = item.getView().getWidth() + itemcoord.x;
         }
         if (item.getView().getHeight() + itemcoord.y > ye) {
            ye = item.getView().getHeight() + itemcoord.y;
         }
      }
      return new Rectangle(xs, ys, xe - xs, ye - ys);
   }

   public final void dragSelection( MouseEvent event ) {
      Point xycoord = event.getPoint();
      if (_draggingSelection == null) {	// start dragging.
         _draggingSelection = new DraggingSelection(getRectangleAroundSelectedItems(), xycoord);
      } else {							// dragging the current selection...
         Point shift = _draggingSelection.getOffset();
         Point newOrigin = new Point(xycoord.x + shift.x, xycoord.y + shift.y);
         if (newOrigin.x < 0) {
            newOrigin.x = 0;
         }
         if (newOrigin.y < 0) {
            newOrigin.y = 0;
         }
         Point translation = _draggingSelection.getTranslation(newOrigin);
         _graphView.scrollRectToVisible(new Rectangle(xycoord.x, xycoord.y, 10, 10));
         dragSelection(translation);
      }
   }

   public final boolean dropSelection() {
      if (_draggingSelection != null) {
         moveSelection(_draggingSelection.getOriginLocation(), _draggingSelection.getLocation());
         _draggingSelection = null;
         return true;
      }
      return false;
   }

   private void moveSelection( Point from, Point to ) {
      startActionEdition();
      moveItems(_selectedItems, from, to);
      endActionEdition();
   }

   public final void moveSelection( Point translation ) {
      for (GraphItem item : _selectedItems) {
         if (item.getLocation().x - translation.x < 0 || item.getLocation().y - translation.y < 0) {
            return;
         }
      }
      startActionEdition();
      moveItems(_selectedItems, translation);
      endActionEdition();
   }

   private void dragSelection( Point translation ) {
      for (GraphItem item : _selectedItems) {
         Point xy = item.getLocation();
         xy.translate(-translation.x, -translation.y);
         moveItem(item, xy, false);
      }
      _graphView.repaint();
   }

   public final void deleteSelectedItems() {
      startActionEdition();
      while (!_selectedItems.isEmpty()) {
         GraphItem item = _selectedItems.remove(0);
         removeItem(item, true);
      }
      endActionEdition();
      unSelectAll();
   }

   public final XMLElement copySelectedItems() {
      try {
         XMLElement element = new XMLElement("ClipBoard");
         XMLElement items = new XMLElement("ItemList");
         for (GraphItem item : _selectedItems) {
            items.addChildren(item.convertToXML(null));
         }
         XMLElement connections = new XMLElement("ConnectionList");
         for (GraphItem item : _selectedItems) {
            connections.addChildren(item.convertSlotsToXML());
         }
         element.addChildren(items);
         element.addChildren(connections);
         return element;
      } catch (IOException ex) {
         return null;
      }
   }

   protected final void pasteSelectedItems( XMLElement clipBoard ) {
      startActionEdition();
      try {
         Point viewPosition = ((JViewport) _graphView.getParent()).getViewPosition();
         List<GraphItem> items = GraphItem.createFromXML(this, clipBoard.getChildren().get(0).getChildren());
         Slot.createFromXML(this, clipBoard.getChildren().get(1).getChildren());

         if (!items.isEmpty()) {
            Point min = items.get(0).getLocation();
            for (GraphItem item : items) {
               Point point = item.getLocation();
               if (point.y < min.y) {
                  min.y = point.y;
               }
               if (point.x < min.x) {
                  min.x = point.x;
               }
            }
            final Point translation = new Point(viewPosition.x + 5 - min.x, viewPosition.y + 5 - min.y);
            // translate all item to be visible into the current viewport.
            for (GraphItem item : items) {
               Point coordinates = item.getLocation();
               coordinates.translate(translation.x, translation.y);
               _graphView.add(item.getView(), new XYConstraint(coordinates));
               _graphView.setComponentZOrder(item.getView(), 0); // Put it in front
            }
         }

         // start the new selection
         unSelectAll();
         for (GraphItem item : _graphModel.getListOfGraphItems()) {
            // Discard virtual operators.
            if (!(item instanceof VirtualItem) && item.isInSameReadPasteSession()) {
               addInSelection(item, true);
            }
         }

         ((JViewport) _graphView.getParent()).setViewPosition(viewPosition);
         revalidate();
      } catch (IOException ex) {
      }
      endActionEdition();
   }

   public final Link detectLink( double x, double y ) {
      for (GraphItem item : _graphModel.getListOfGraphItems()) {
         LinkedList<InputSlot> allSlots = new LinkedList<>();
         allSlots.addAll(item.getInputDataSlots());
         allSlots.addAll(item.getInputParameterSlots());
         if (item.getInputMaskSlot() != null) {
            allSlots.add(item.getInputMaskSlot());
         }
         for (InputSlot slot : allSlots) {
            if (slot.getConnectedSlot() != null) {
               Link link = slot.getLink();
               if (link.closestDistanceToLink(x, y)) {
                  return link;
               }
            }
         }
      }
      return null;
   }

   /////
   ///// I N P U T   /   O U T P U T
   /////
   public abstract void save();

   public abstract void saveAs();

   protected final File askForFilename() {
      return _graphView.askForFilename();
   }

   protected final File getStorageFile() {
      return _graphModel.getStorageFile();
   }

   protected final void save( File file ) {
      try {
         _graphModel.save(file, this);
         Ariane.getPresenter().printSystemMessage(_graphModel.getStorageFile() + ArianeSettings.getString("Workspace.saving.success"));
         setModified(false);
      } catch (IOException ex) {
         Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Workspace.saving.error1") + file.getName() + " " + ex.getLocalizedMessage());
      }
   }

   public final XMLElement convertToXMLTree( File file ) throws IOException {
      return _graphModel.convertToXMLTree(file, this);
   }

   public final void createFromXMLTree( List<XMLElement> xmlTree ) throws IOException {
      _graphModel.createFromXMLTree(xmlTree, this);
      _undoRedoManager.clearCurrentListOfActions();
   }

   /////
   ///// U N D O / R E D O (H I S T O R Y)
   /////
   public final void validateOneAction( UndoableEdit action ) {
      startActionEdition();
      validateAction(action);
      endActionEdition();
   }

   public final void validateAction( UndoableEdit action ) {
      _undoRedoManager.validateAction(action);
   }

   public final void startActionEdition() {
      _undoRedoManager.startActionEdition();
   }

   public final void endActionEdition() {
      _undoRedoManager.endActionEdition(this);
   }

   public final boolean canUndo() {
      return _undoRedoManager.canUndo();
   }

   public final boolean canRedo() {
      return _undoRedoManager.canRedo();
   }

   public final void undo() {
      setModified(true);
      _undoRedoManager.undo();
   }

   public final void redo() {
      setModified(true);
      _undoRedoManager.redo();
   }

   ///////
   /////// S E L E C T I O N
   ///////
   private final class DraggingSelection {

      private final Rectangle _selectionRectangle;
      private final Point _offsetBrweenMousePointerAndRectanglel;
      private final Point _rectangleLocation;

      DraggingSelection( Rectangle rectangle, Point xycoord ) {
         _selectionRectangle = rectangle;
         _offsetBrweenMousePointerAndRectanglel = new Point(rectangle.x - xycoord.x, rectangle.y - xycoord.y);
         _rectangleLocation = rectangle.getLocation();
      }

      public Point getOriginLocation() {
         return _selectionRectangle.getLocation();
      }

      public Point getOffset() {
         return _offsetBrweenMousePointerAndRectanglel;
      }

      public Point getLocation() {
         return _rectangleLocation;
      }

      public Point getTranslation( Point newOrigin ) {
         Point translation = new Point();
         translation.x = _rectangleLocation.x;
         translation.y = _rectangleLocation.y;
         _rectangleLocation.x = newOrigin.x;
         _rectangleLocation.y = newOrigin.y;
         translation.x -= newOrigin.x;
         translation.y -= newOrigin.y;
         return translation;
      }
   }
}
