/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.presenter;

import java.awt.Point;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JMenuBar;
import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.generator.CodeGenerator;
import fr.greyc.pantheon.ariane.model.tree.ItemTreeModel;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.model.xml.XMLFileInputStream;
import fr.greyc.pantheon.ariane.view.OperatorLibraryPathChooser;
import fr.greyc.pantheon.ariane.view.PropertyPanel;
import fr.greyc.pantheon.ariane.view.UserPreferencesChooser;
import fr.greyc.pantheon.ariane.view.Workspace;
import fr.greyc.pantheon.ariane.view.explorer.ItemTreeView;
import fr.greyc.pantheon.ariane.view.utils.ScriptNameChooser;

public final class Presenter {

   private static int _currentReadPasteSession;

   private final Workspace _workspace;

   private final ItemTreeModel _itemTreeModel;

   public Presenter( ItemTreeModel itemTreeModel ) {
      _itemTreeModel = itemTreeModel;
      _workspace = new Workspace(this, new ItemTreeView(this, itemTreeModel));
   }

   /////
   ///// W O R K S P A C E
   /////
   public void generatesScriptCodeFromGraph( Graph graph ) {
      String[] chosenFile = ScriptNameChooser.chooseFilename(graph.getName(), CodeGenerator.AVAILABLE_LANGUAGES);
      if (chosenFile.length > 0) {
         CodeGenerator generator = CodeGenerator.createCodeGeneratorFromLanguage(new File(chosenFile[0]), chosenFile[1]);
         generator.generate(this, graph.getListOfGraphItems());
      }
   }

   public boolean generatesScriptCodeFromCurrentGraph( String scriptFilename, List<String> graphFilenames ) {
      int suffixPosition = scriptFilename.lastIndexOf('.');
      String suffix = suffixPosition > 0 ? scriptFilename.substring(suffixPosition + 1) : "";
      Graph graph = loadGraph(new File(graphFilenames.get(0)));
      CodeGenerator generator = CodeGenerator.createCodeGeneratorFromSuffix(new File(scriptFilename), suffix);
      if (generator != null && graph != null) {
         generator.generate(this, graph.getListOfGraphItems());
         return true;
      }
      return false;
   }

   private void changeOperatorLibraryPath( boolean isUsed, String pathName ) {
      UserPreferences.setOperatorLibraryState(isUsed);
      if (!isUsed) {
         return;
      }
      String libraryFilename = pathName + ArianeSettings.getLibraryConfigFilename();
      try {
         List<XMLElement> configHandler = ArianeSettings.readXMLFile(new File(libraryFilename));
         ItemTreeModel operatorTree = _itemTreeModel;
         operatorTree.setCommonOperators(configHandler);
         ArianeSettings.setLibraryPath(pathName);
      } catch (IOException e) {
         UserPreferences.setOperatorLibraryState(false);
         Ariane.getPresenter().printErrorMessage(e.getLocalizedMessage());
      }
   }

   public void chooseOperatorLibrary() {
      OperatorLibraryPathChooser libraryChooser = new OperatorLibraryPathChooser(UserPreferences.isOperatorLibraryUsed(), ArianeSettings.getLibraryPath());
      libraryChooser.setVisible(true);
      if (libraryChooser.isValidated()) {
         changeOperatorLibraryPath(libraryChooser.isUsed(), libraryChooser.getOperatorLibraryPath());
         if (UserPreferences.updateUserPreferenceFile()) {
            _workspace.repaintAllGraphs();
         }
      }
   }

   public void openPreferencesChooser() {
      UserPreferencesChooser preferenceChooser = new UserPreferencesChooser();
      preferenceChooser.setVisible(true);
      if (preferenceChooser.isValidated()) {
         _workspace.repaintAllGraphs();
         UserPreferences.updateUserPreferenceFile();
      }
   }

   public void restoreSplitPanesLocation() {
      _workspace.restoreSplitPanesLocation();
   }

   public void display( List<String> graphFilenames ) {
      List<Graph> graphs = new ArrayList<>();
      for (String graphFilename : graphFilenames) {
         graphs.add(loadGraph(new File(graphFilename)));
      }
      _workspace.display(graphs);
   }

   public static int getCurrentReadOrPasteSessionId() {
      return _currentReadPasteSession;
   }

   public JMenuBar buildMenuBar() {
      return _workspace.buildMenuBar();
   }

   public void initContent() {
      _workspace.initContent();
   }

   public Workspace getWorkspace() {
      return _workspace;
   }

   /////
   ///// G R A P H
   /////
   public void openFile( String filename ) {
      if (!_workspace.isGraphAlreadyLoaded(filename)) {
         Graph graph = loadGraph(new File(filename));
         if (graph != null) {
            _workspace.addGraph(graph);
            Ariane.getPresenter().printSystemMessage(filename + ArianeSettings.getString("Workspace.load.message"));
         }
      }
   }

   public void createItemsFromClipboard( Graph graph, Transferable contents ) {
      try {
         _currentReadPasteSession++;
         XMLElement result = (XMLElement) contents.getTransferData(XMLElement.FLAVOR);
         graph.pasteSelectedItems(result);
      } catch (UnsupportedFlavorException | IOException e) {
      }
   }

   public Graph loadGraph( File file ) {
      _currentReadPasteSession++;
      try {
         XMLFileInputStream convert = new XMLFileInputStream(file);
         convert.parse();
         convert.close();
         Graph graph = new RootGraph(file);
         graph.createFromXMLTree(convert.getElements());
         addRecentProjects(file.getAbsolutePath());
         return graph;
      } catch (IOException | SAXException | ParserConfigurationException ex) {
         printErrorMessage(ArianeSettings.getString("Workspace.loading.error1") + file.getName() + " " + ex.getLocalizedMessage());
         return null;
      }
   }

   public void addNewItemInGraphAtRelativePosition( TreeItem treeItem, Point coordinate ) throws InstantiationException {
      _workspace.addNewItemAtRelativePosition(treeItem, coordinate);
   }

   /////
   ///// C O N S O L E
   /////
   public void printOutputMessage( String message ) {
      _workspace.printOutputMessage(message);
   }

   public void printSystemMessage( String mesg ) {
      _workspace.printSystemMessage(mesg);
   }

   public void printWarning( String mesg ) {
      _workspace.printWarning(mesg);
   }

   public void printErrorMessage( String mesg ) {
      _workspace.printErrorMessage(mesg);
   }

   public void printErrorMessage( ArianeSettings.ErrorCode errorCode, String item, int itemId, String mesg ) {
      _workspace.printErrorMessage(errorCode, item, itemId, mesg);
   }

   public void displayDocumentationPage( String title, String page, String index ) {
      _workspace.displayHTMLManualPage(title, page, index);
   }

   public void copy() {
      _workspace.copy();
   }

   public void cut() {
      _workspace.cut();
   }

   public void paste() {
      _workspace.paste();
   }

   public ItemTreeModel getTree() {
      return _itemTreeModel;
   }

   public void displayPropertyPanel( Graph graph, PropertyPanel propertyPanel ) {
      _workspace.displayPropertyPanel(propertyPanel);
      _workspace.setCurrentGraph(graph);
   }

   public void addGraph( SubGraph subGraph ) {
      _workspace.addGraph(subGraph);
   }

   public void renameGraph( Graph graph, String name ) {
      _workspace.renameGraph(graph, name);
   }

   public void setExecution( boolean isRunning ) {
      _workspace.setExecution(isRunning);
   }

   public void updateMenu() {
      _workspace.updateMenu();
   }

   public void updateRecentFilesListMenu() {
      _workspace.updateRecentlyClosedWorkspacesListMenu();
   }

   public void removeGraph( Graph graph ) {
      _workspace.removeGraph(graph);
   }

   public void addRecentProjects( String absolutePath ) {
      _workspace.addRecentlyClosedWorkspace(absolutePath);
   }

   public boolean isPastable() {
      return _workspace.isPastable();
   }

   public void deleteSelectedItems() {
      _workspace.deleteSelectedItems();
   }

   public void setModifiedSaveB( boolean value ) {
      _workspace.setModifiedSaveB(value);
   }

   public void changeLanguage() {
      _workspace.changeLanguage();
   }

   public int[] getSplitPaneRatios() {
      return _workspace.getSplitPaneRatios();
   }

   public void displayLibraryDocumentation() {
      displayDocumentationPage("Index", _itemTreeModel.getPathToLibraryDocumentationFile(), null);
   }
}
