/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;
import java.util.ResourceBundle;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ScrollPaneConstants;
import javax.swing.UIManager;
import javax.swing.border.EmptyBorder;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.swingext.AttributeValueLayout;

/**
 * A <code>PropertyViewer</code> provides a dialog box
 * to display graph item's properties.
 */
public final class PropertyPanel extends JPanel {

   private static final int DEFAULT_ROW_HEIGHT = 25;
   private PropertyPanel.TextField _resultLabel;
   private final List<JButton> _inputButtons = new ArrayList<>();
   private final List<JButton> _outputButtons = new ArrayList<>();
   private ResourceBundle _currentLanguage;
   private final List<PropertyCouple> _properties = new ArrayList<>();
   private final List<PropertyCouple> _inputsoutputs = new ArrayList<>();

   public PropertyPanel() {
      setOpaque(false);
   }

   public PropertyPanel( final ResourceBundle language, final GraphItem item ) {
      _currentLanguage = language;
      setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), ArianeSettings.getString("PropertyViewer.title") + " (" + item.getId() + ")"));
      setLayout(new BorderLayout());
      setBackground(getBackground().darker());

      buildPropertyList(item);

      JPanel contents = initPanelContents(item);
      JScrollPane scrollPane = new JScrollPane(contents, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
      scrollPane.setBorder(null);
      add(scrollPane, BorderLayout.CENTER);
   }

   public boolean isCorrectlyLocalized() {
      return _currentLanguage == ArianeSettings.getLanguage();
   }

   public void update( final GraphItem item ) {
      boolean updated = false;
      if (_resultLabel != null) {
         updated = true;
         if (item.getExecutionStatus() == GraphItem.Status.EXECUTED) {
            _resultLabel.setText(item.getOutputResultSlot().getValue());
         } else {
            _resultLabel.setText(ArianeSettings.getString("PropertyViewer.result.unbound"));
         }
      }
      // Input buttons
      if (!_inputButtons.isEmpty()) {
         for (int i = 0; i < item.getInputDataSlots().size(); i++) {
            updated = true;
            boolean enabled = item.getExecutionStatus() == GraphItem.Status.EXECUTED && item.getInputDataSlots().get(i).isValidateValue();
            _inputButtons.get(i).setEnabled(enabled);
         }
      }
      // Output buttons
      if (!_outputButtons.isEmpty()) {
         for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            updated = true;
            boolean enabled = item.getExecutionStatus() == GraphItem.Status.EXECUTED && item.getOutputDataSlots().get(i).isValidate();
            _outputButtons.get(i).setEnabled(enabled);
         }
      }
      if (updated) {
         revalidate();
         repaint();
      }
   }

   private JPanel initPanelContents( final GraphItem item ) {
      JPanel contents = new JPanel();
      contents.setLayout(new BoxLayout(contents, BoxLayout.Y_AXIS));
      contents.setBackground(getBackground().darker());

      Dimension dimension = new Dimension(10000, DEFAULT_ROW_HEIGHT);
      JPanel panel1 = new JPanel(new AttributeValueLayout());
      for (PropertyCouple property : _properties) {
         JLabel title = new JLabel(property._title);
         title.setOpaque(true);
         title.setBorder(new EmptyBorder(0, 2, 0, 0));
         title.setFont(UserPreferences.getDefaultFont());
         panel1.add(title);

         JComponent component = property._value;
         adjustComponentSize(component);
         panel1.add(property._value);

         panel1.setMaximumSize(dimension);
         contents.add(panel1);
      }

      for (PropertyCouple property : _inputsoutputs) {
         JPanel panel = new JPanel(new BorderLayout());
         JLabel title = new JLabel(property._title);
         title.setOpaque(true);
         title.setBorder(new EmptyBorder(0, 2, 0, 0));
         title.setFont(UserPreferences.getDefaultFont());
         panel.add(title, BorderLayout.NORTH);

         JComponent component = property._value;
         adjustComponentSize(component);
         panel.add(property._value, BorderLayout.CENTER);

         panel.setMaximumSize(dimension);
         contents.add(panel);
      }

      JPanel buttonPanel = new JPanel(new GridLayout(1, 2));
      JButton buttonOK = new JButton(ArianeSettings.getString("Graph.Menu.execute"));
      buttonOK.setFont(UserPreferences.getDefaultFont());
      buttonOK.addActionListener(new ActionListener() {
         @Override
         public void actionPerformed( ActionEvent event ) {
            item.getGraph().startExecution();
         }
      });
      buttonPanel.add(buttonOK);

      JButton buttonReset = new JButton(ArianeSettings.getString("Graph.Menu.reset"));
      buttonReset.setFont(UserPreferences.getDefaultFont());
      buttonReset.addActionListener(new ActionListener() {
         @Override
         public void actionPerformed( ActionEvent event ) {
            item.resetAll();
         }
      });
      dimension = buttonPanel.getPreferredSize();
      dimension.width = 10000;
      buttonPanel.setMaximumSize(dimension);
      buttonPanel.add(buttonReset);
      contents.add(buttonPanel);
      contents.add(Box.createVerticalGlue());
      return contents;
   }

   private void adjustComponentSize( final JComponent component ) {
      int height = component.getPreferredSize().height;
      Dimension dimension = new Dimension(0, height);
      component.setPreferredSize(dimension);
   }

   private void buildPropertyList( final GraphItem item ) {
      createNameElement(item);
      createCategoryElement(item);
      createBinaryElement(item);
      createManualElement(item);
      createFormulaElement(item);
      createSpecificElement(item);
      createParametersElement(item);
      createInputElement(item);
      createOutputElement(item);
      createResultElement(item);
   }

   private void createResultElement( final GraphItem item ) {
      if (item.getOutputResultSlots().size() > 0) {
         _resultLabel = new PropertyPanel.TextField();
         if (item.getExecutionStatus() == GraphItem.Status.EXECUTED) {
            _resultLabel.setText(item.getOutputResultSlot().getValue());
         } else {
            _resultLabel.setText(ArianeSettings.getString("PropertyViewer.result.unbound"));
         }
         _resultLabel.setEditable(false);
         _properties.add(new PropertyCouple(ArianeSettings.getString("PropertyViewer.result"), _resultLabel));
      }
   }

   private void createOutputElement( final GraphItem item ) {
      for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
         JButton button = (JButton) item.getOutputDataSlots().get(i).getView().getValueEditor();
         boolean enabled = item.getExecutionStatus() == GraphItem.Status.EXECUTED && item.getOutputDataSlots().get(i).isValidate();
         button.setEnabled(enabled);
         _outputButtons.add(button);
         _inputsoutputs.add(new PropertyCouple(ArianeSettings.getString("PropertyViewer.output")
             + " #"
             + i
             + " (ctrl-"
             + i + ")", button));
      }
   }

   private void createInputElement( final GraphItem item ) {
      for (int i = 0; i < item.getInputDataSlots().size(); i++) {
         JButton button = (JButton) item.getInputDataSlots().get(i).getView().getValueEditor();
         boolean enabled = item.getExecutionStatus() == GraphItem.Status.EXECUTED && item.getInputDataSlots().get(i).isValidateValue();
         button.setEnabled(enabled);
         _inputButtons.add(button);
         _inputsoutputs.add(new PropertyCouple(ArianeSettings.getString("PropertyViewer.input") + " #" + i, button));
      }
   }

   private void createParametersElement( final GraphItem item ) {
      for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
         InputParameterSlot isp = item.getInputParameterSlots().get(i);
         JComponent component = isp.getView().getValueEditor();
         _properties.add(new PropertyCouple(item.getInputParameterSlots().get(i).getName(), component));
      }
   }

   private void createSpecificElement( final GraphItem item ) {
      List<PropertyCouple> additionalProperties = item.getView().getSpecificProperties();
      if (additionalProperties != null) {
         _properties.addAll(additionalProperties);
      }
   }

   private void createFormulaElement( final GraphItem item ) {
      if (item.getFormula() != null && item.isDisplayableFormula()) {
         JComponent component = item.getView().getFormulaEditor();
         _properties.add(new PropertyCouple(item.getFormulaLabel(), component));
      }
   }

   private void createManualElement( final GraphItem item ) {
      JButton jbuttonman;
      jbuttonman = new JButton(item.getName());
      jbuttonman.setFont(UserPreferences.getDefaultFont());
      jbuttonman.addActionListener(new ActionListener() {
         @Override
         public void actionPerformed( ActionEvent event ) {
            item.getView().displayDocumentationManual();
         }
      });
      _properties.add(new PropertyCouple(ArianeSettings.getString("PropertyViewer.documentation"), jbuttonman));
   }

   private void createBinaryElement( final GraphItem item ) {
      JComponent editor = item.getView().getBinaryEditor();
      if (editor != null) {
         _properties.add(new PropertyCouple(ArianeSettings.getString("PropertyViewer.binary"), editor));
      }
   }

   private void createCategoryElement( final GraphItem item ) {
      if (item.getCategory() != null) {
         JTextField component = new JTextField(item.getCategory());
         component.setEnabled(false);
         _properties.add(new PropertyCouple(ArianeSettings.getString("PropertyViewer.category"), component));
      }
   }

   private void createNameElement( final GraphItem item ) {
      _properties.add(new PropertyCouple(ArianeSettings.getString("PropertyViewer.name"), item.getView().getNameEditor()));
   }

   public static final class PropertyCouple {

      private final String _title;
      private final JComponent _value;

      public PropertyCouple( final String title, final JComponent value ) {
         _title = title;
         value.setFont(UserPreferences.getDefaultFont());
         _value = value;
      }
   }

   public static final class TextField extends JTextField {

      public TextField() {
         Color disabledTextColor = UIManager.getColor("TextField.disabledText");
         setForeground(disabledTextColor);
      }

      /**
       * This method was added because Nimbus L&F does not consider the non editable state (only enabled and disabled).
       */
      @Override
      public void setEditable( boolean editable ) {
         super.setEditable(editable);
         if (editable) {
            Color disabledBackground = UIManager.getColor("TextField.background");
            setBackground(disabledBackground);
         } else {
            Color disabledBackground = UIManager.getColor("control");
            setBackground(disabledBackground);
         }
      }
   }
}
