/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.explorer;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.tree.ItemTreeModel;
import fr.greyc.pantheon.ariane.model.tree.MutableTreeItem;
import fr.greyc.pantheon.ariane.model.tree.SubroutineTreeItem;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.model.xml.XMLFileOutputStream;
import fr.greyc.pantheon.swingext.GDialog;
import fr.greyc.pantheon.swingext.HTMLDocumentEditor;
import fr.greyc.pantheon.swingext.MessageBox;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

/**
 * The class <code>SubroutineEditor</code> defines a dialog box where
 * the user can add new subroutine in the explorer.
 */
public final class SubroutineEditor extends GDialog {

   private SubroutineTreeItem _currentItem;
   private MutableTreeItem _currentNode;
   private DefaultTreeModel _treeModel;
   private JComboBox<String> _categoryField;
   private JTextField _nameField;
   private JTextField _signatureField;
   private JTextField _docURLField;
   private boolean _inserted;
   private boolean _validated;
   private static String _currentDirectory = UserPreferences.getPreferredBinaryDirectory();

   public SubroutineEditor( CompositeItem item ) {
      super((JDialog) null, ArianeSettings.getString("SubroutineEditor.title"), true);
      addWindowListener(new WindowAdapter() {
         @Override
         public void windowClosing( WindowEvent event ) {
            cancelAction();
         }
      });

      try {
         XMLElement element;
         element = new XMLElement("Subroutine");
         element.addChildren(new XMLElement("Category", ""));
         element.addChildren(new XMLElement("Name", item.getNickName()));
         element.addChildren(new XMLElement("Documentation", ""));
         // Build the default parameter values list
         String list = "";
         for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            InputParameterSlot parameter = item.getInputParameterSlots().get(i);
            String name;
            if (parameter.hasName()) {
               name = parameter.getName();
            } else {
               name = "";
            }

            String defaultValue = "";
            try {
               boolean hasValue = parameter.hasValue(false);
               if (!parameter.isConnected() && hasValue) {
                  defaultValue = "=" + parameter.getValue();
               }
            } catch (ArianeMissingParameterException e) {
            }
            if (name.length() > 0 || defaultValue.length() > 0) {
               list += "$" + i + ":" + name + defaultValue + ";";
            }
         }
         element.addChildren(new XMLElement("Signature", list));
         element.addChildren(item.convertToXML(null));
         _currentItem = new SubroutineTreeItem(element.getChildren());
         _inserted = true;
      } catch (IOException e) {
      }
      displayContent();
      setLocationRelativeTo(null);
   }

   public SubroutineEditor( JDialog parent, MutableTreeItem currentNode, DefaultTreeModel treeModel ) {
      super(parent, ArianeSettings.getString("SubroutineEditor.title"), true);
      _currentNode = currentNode;
      _currentItem = (SubroutineTreeItem) _currentNode.getUserObject();
      _treeModel = treeModel;
      _inserted = false;
      displayContent();
      setLocationRelativeTo(parent);
   }

   public boolean isValidated() {
      return _validated;
   }

   public SubroutineTreeItem getTreeItem() {
      return _currentItem;
   }

   private void displayContent() {
      setResizable(false);
      JPanel textPanel = new JPanel();
      textPanel.setLayout(new GridBagLayout());
      textPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), ArianeSettings.getString("SubroutineEditor.borderTitle")));
      GridBagConstraints cst = new GridBagConstraints();

      cst.fill = GridBagConstraints.BOTH;
      textPanel.add(new JLabel(ArianeSettings.getString("SubroutineEditor.categoryText")), cst);
      String[] categories = Ariane.getPresenter().getTree().getSubroutineCategories();
      _categoryField = new JComboBox<>(categories);
      _categoryField.setEditable(true);
      _categoryField.setSelectedItem(_currentItem.getCategory());
      cst.gridwidth = GridBagConstraints.REMAINDER; //end row
      textPanel.add(_categoryField, cst);

      cst.gridwidth = 1;
      textPanel.add(new JLabel(ArianeSettings.getString("SubroutineEditor.nameText")), cst);
      _nameField = new JTextField(30);
      _nameField.setText(_currentItem.getName());
      cst.gridwidth = GridBagConstraints.REMAINDER; //end row
      textPanel.add(_nameField, cst);

      // Signature
      cst.gridwidth = 1;
      textPanel.add(new JLabel(ArianeSettings.getString("SubroutineEditor.signature")), cst);
      _signatureField = new JTextField(_currentItem.getSignature());
      cst.gridwidth = GridBagConstraints.REMAINDER; //end row
      textPanel.add(_signatureField, cst);
      textPanel.add(new JLabel(ArianeSettings.getString("SubroutineEditor.signatureExample")), cst);

      cst.gridwidth = 1;
      textPanel.add(new JLabel(ArianeSettings.getString("SubroutineEditor.documentationText")), cst);
      _docURLField = new JTextField(30);
      _docURLField.setText(_currentItem.getDocumentationURL());
      textPanel.add(_docURLField, cst);
      JButton button2 = new JButton(ArianeSettings.getString("SubroutineEditor.load"));
      button2.addActionListener(new ActionListener() {
         @Override
         public void actionPerformed( ActionEvent event ) {
            JFileChooser fileChooser = new JFileChooser(_currentDirectory);
            fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
            fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
            final String prefix = "file://";
            if (_docURLField.getText().length() > prefix.length()) {
               fileChooser.setSelectedFile(new File(_docURLField.getText().substring(prefix.length())));
            }
            int returnVal = fileChooser.showOpenDialog(SubroutineEditor.this);
            if (returnVal == JFileChooser.APPROVE_OPTION) {
               _docURLField.setText(prefix + fileChooser.getSelectedFile().getAbsolutePath());
               _docURLField.setCaretPosition(0);
               _currentDirectory = fileChooser.getSelectedFile().getParentFile().getAbsolutePath();
            }
         }
      });
      JButton button3 = new JButton(ArianeSettings.getString("OperatorEditor.edit"));
      button3.addActionListener(new ActionListener() {
         @Override
         public void actionPerformed( ActionEvent event ) {
            HTMLDocumentEditor editor = new HTMLDocumentEditor(SubroutineEditor.this);
            if (_docURLField.getText().length() > 6) {
               editor.setSelectedFile(new File(_docURLField.getText().substring(7)));
            }
            if (editor.edit() == HTMLDocumentEditor.OK_OPTION && editor.getSelectedFile() != null) {
               _docURLField.setText("file://" + editor.getSelectedFile().getAbsolutePath());
               _docURLField.setCaretPosition(0);
            }
         }
      });
      textPanel.add(button2, cst);
      cst.gridwidth = GridBagConstraints.REMAINDER; //end row
      textPanel.add(button3, cst);

      // cancel button
      JButton cancelButton = new JButton();
      cancelButton.setAction(new AbstractAction(ArianeSettings.getString("SubroutineEditor.cancel"),
          ArianeSettings.getIcon("gtk-cancel.png")) {
         @Override
         public void actionPerformed( ActionEvent event ) {
            cancelAction();
         }
      });

      // Apply button
      JButton applyButton = new JButton(ArianeSettings.getString("SubroutineEditor.apply"), ArianeSettings.getIcon("gtk-apply.png"));

      applyButton.addActionListener(new ActionListener() {
         @Override
         public void actionPerformed( ActionEvent event ) {
            applyAction();
         }
      });

      JPanel buttonPanel = new JPanel();
      buttonPanel.add(applyButton);
      buttonPanel.add(cancelButton);

      setDefaultButton(applyButton);
      setCancelButton(cancelButton);

      getContentPane().setLayout(new BorderLayout());
      getContentPane().add(textPanel, BorderLayout.NORTH);
      getContentPane().add(buttonPanel, BorderLayout.SOUTH);
      pack();
   }

   private void cancelAction() {
      _validated = false;
      setVisible(false);
      dispose();
   }

   private void applyAction() {
      String category = _categoryField.getSelectedItem().toString();
      if (category.trim().isEmpty()) {
         Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("SubroutineEditor.error1"));
         return;
      }
      _currentItem.setCategory(category);
      _currentItem.setName(_nameField.getText());
      _currentItem.setSignature(_signatureField.getText());
      _currentItem.setDocumentationURL(_docURLField.getText());
      if (_inserted) {
         ItemTreeModel treeModel = Ariane.getPresenter().getTree();
         List<TreePath> litems = treeModel.getItemsWithTheSpecifiedNameInSpecifiedSubtree(treeModel.getSubroutineTree(), _currentItem.getName());
         if (!litems.isEmpty()) {
            String[] options = {ArianeSettings.getString("overwriteButtonText"), ArianeSettings.getString("nooverwriteButtonText")};
            if (MessageBox.showConfirmDialog(null,
                _currentItem.getName() + " " + ArianeSettings.getString("Tree.dialog.overwriteMessage"),
                ArianeSettings.getString("Tree.dialog.overWriteTitle"),
                options, options[0]) != 0) {
               return;
            }
            Ariane.getPresenter().printWarning(ArianeSettings.getString("CompositeItem.message2"));
         }
         treeModel.addSubroutine(_currentItem);
         save();
         _validated = true;
      } else {
         _treeModel.nodeChanged(_currentNode);
         _validated = false;
      }
      setVisible(false);
      dispose();
   }

   private boolean save() {
      File file = new File(ArianeSettings.getSubroutineFilename());
      File folder = file.getParentFile();
      try {
         folder.mkdirs();
         XMLFileOutputStream out = new XMLFileOutputStream(file, null);
         out.write(convertToXMLTree());
         out.close();
      } catch (IOException ex) {
         Ariane.getPresenter().printSystemMessage(ArianeSettings.getString("PreferenceChooser.message.saved") + " " + ArianeSettings.getPreferencesFilename());
         return false;
      }
      return true;
   }

   private XMLElement convertToXMLTree() throws IOException {
      List<TreeItem> subRoutineList = Ariane.getPresenter().getTree().getSubroutineTreeAsList();
      if (!subRoutineList.isEmpty()) {
         XMLElement element = new XMLElement("SubroutineLibrary");
         for (TreeItem item : subRoutineList) {
            element.addChildren(item.convertToXMLTree());
         }
         return element;
      } else {
         return null;
      }
   }
}
