/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.explorer;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.tree.MutableTreeItem;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.swingext.GDialog;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.AbstractAction;
import javax.swing.Box;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JButton;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.border.EmptyBorder;
import javax.swing.border.LineBorder;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

/**
 * The class <code>TreeEditor</code> defines a dialog box that allows users
 * to modify the tree of items (subroutine or operator subtree).
 */
public abstract class TreeEditor extends GDialog implements MouseListener {

   private final MutableTreeItem _rootNode;

   private final JTree _jtree;
   private final DefaultTreeModel _treeModel;

   private JButton _newButton;
   private JButton _editButton;
   private JButton _deleteButton;

   public TreeEditor( String title, MutableTreeItem rootNode ) {
      super((JFrame) null, title, true);
      setResizable(false);
      _rootNode = (MutableTreeItem) rootNode.clone();
      _treeModel = new DefaultTreeModel(_rootNode);
      addWindowListener(new WindowAdapter() {
         @Override
         public final void windowClosing( WindowEvent event ) {
            cancelEdition();
         }
      });

      _jtree = new JTree(_treeModel);
      _jtree.getInputMap().put(KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_DELETE, 0), "Delete");
      _jtree.getActionMap().put("Delete", new AbstractAction(null) {
         @Override
         public final void actionPerformed( ActionEvent event ) {
            deleteSelectedItem();
         }
      });

      displayContent();
      TreePath path = new TreePath(_treeModel.getRoot());
      _jtree.setSelectionPath(path);
      selectNode();
      setLocationRelativeTo(Ariane.getFrame());
   }

   private void displayContent() {
      setSize(300, 400);
      _jtree.setEditable(true);
      _jtree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
      _jtree.setShowsRootHandles(true);
      _jtree.addMouseListener(this);
      _jtree.setBorder(new LineBorder(Color.BLACK));

      JPanel scrollPane = new JPanel(new BorderLayout());
      scrollPane.add(new JScrollPane(_jtree), BorderLayout.CENTER);
      scrollPane.setBorder(new EmptyBorder(10, 10, 10, 10));

      _editButton = new JButton(ArianeSettings.getString("TreeEditor.edit"));
      _editButton.addActionListener(new ActionListener() {
         @Override
         public final void actionPerformed( ActionEvent event ) {
            editCurrentItem();
         }
      });
      _editButton.setEnabled(false);

      _newButton = new JButton(ArianeSettings.getString("TreeEditor.new"));
      _newButton.addActionListener(new ActionListener() {
         @Override
         public final void actionPerformed( ActionEvent event ) {
            addNewItemBelowCurrentSelection();
         }
      });
      _newButton.setEnabled(false);

      _deleteButton = new JButton(ArianeSettings.getString("TreeEditor.delete"));
      _deleteButton.addActionListener(new ActionListener() {
         @Override
         public final void actionPerformed( ActionEvent event ) {
            deleteSelectedItem();
         }
      });
      _deleteButton.setEnabled(false);

      JPanel buttonPanel1 = new JPanel();
      buttonPanel1.add(_editButton);
      buttonPanel1.add(_newButton);
      buttonPanel1.add(_deleteButton);

      JButton saveButton = new JButton(ArianeSettings.getString("TreeEditor.save"), ArianeSettings.getIcon("gtk-apply.png"));
      saveButton.addActionListener(new ActionListener() {
         @Override
         public final void actionPerformed( ActionEvent event ) {
            saveCurrentTree();
         }
      });

      JButton cancelButton = new JButton();
      cancelButton.setAction(new AbstractAction(ArianeSettings.getString("TreeEditor.cancel"), ArianeSettings.getIcon("gtk-cancel.png")) {
         @Override
         public final void actionPerformed( ActionEvent event ) {
            cancelEdition();
         }
      });
      Box buttonPanel = Box.createVerticalBox();
      buttonPanel.add(buttonPanel1);
      JPanel buttonPanel2 = new JPanel();
      buttonPanel2.add(saveButton);
      buttonPanel2.add(cancelButton);
      buttonPanel.add(buttonPanel2);
      setDefaultButton(saveButton);
      setCancelButton(cancelButton);

      getContentPane().setLayout(new BorderLayout());
      getContentPane().add(scrollPane, BorderLayout.CENTER);
      getContentPane().add(buttonPanel, BorderLayout.SOUTH);
   }

   @Override
   public final void mouseClicked( MouseEvent event ) {
      if (event.getClickCount() > 1) {
         editCurrentItem();
      } else {
         selectNode();
      }
   }

   @Override
   public final void mouseEntered( MouseEvent event ) {
      /*
       * empty
       */
   }

   @Override
   public final void mouseExited( MouseEvent event ) {
      /*
       * empty
       */
   }

   @Override
   public final void mousePressed( MouseEvent event ) {
      if (event.isPopupTrigger()) {
         openPopupMenu(event.getX(), event.getY());
      }
   }

   @Override
   public final void mouseReleased( MouseEvent event ) {
      if (event.isPopupTrigger()) {
         openPopupMenu(event.getX(), event.getY());
      }
   }

   private void openPopupMenu( int x, int y ) {
      MutableTreeItem selectedNode = (MutableTreeItem) _jtree.getLastSelectedPathComponent();
      if (selectedNode != null) {
         switch (selectedNode.getLevel()) {
            case 0:
               openPopupMenuRoot(x, y);
               break;
            case 1:
               openPopupMenuCategory(_jtree, x, y);
               break;
            case 2:
               openPopupMenuItem(_jtree, x, y);
               break;
            default:
               break;
         }
      }
   }

   private void openPopupMenuRoot( int x, int y ) {
      JPopupMenu popupMenu = new JPopupMenu();
      JMenuItem item = new JMenuItem(ArianeSettings.getString("TreeEditor.addCategory"));
      item.addActionListener(new ActionListener() {
         @Override
         public void actionPerformed( ActionEvent event ) {
            addNewItemBelowCurrentSelection();
         }
      });
      popupMenu.add(item);
      popupMenu.show(_jtree, x, y);
   }

   protected abstract void openPopupMenuCategory( final JTree jtree, int x, int y );

   protected abstract void openPopupMenuItem( final JTree jtree, int x, int y );

   ///// 
   ///// AC T I O N S
   ///// 
   protected abstract boolean isNewEnabled( TreePath path );

   protected final boolean isEditEnabled( TreePath path ) {
      return path != null && path.getPathCount() > 1;
   }

   protected final boolean isdeleteEnabled( TreePath path ) {
      return isEditEnabled(path);
   }

   private void selectNode() {
      TreePath path = _jtree.getSelectionPath();
      _newButton.setEnabled(isNewEnabled(path));
      _editButton.setEnabled(isEditEnabled(path));
      _deleteButton.setEnabled(isdeleteEnabled(path));
   }

   private void startEditingCategory( MutableTreeItem subtree ) {
      TreePath path = new TreePath(_treeModel.getPathToRoot(subtree));
      _jtree.setSelectionPath(path);
      _jtree.startEditingAtPath(path);
   }

   protected abstract TreeItem createItem( String categoryName );

   protected final void addNewItemBelowCurrentSelection() {
      TreePath currentSelection = _jtree.getSelectionPath();
      MutableTreeItem currentNode;
      if (currentSelection == null) {
         currentNode = _rootNode;
      } else {
         currentNode = (MutableTreeItem) currentSelection.getLastPathComponent();
      }
      if (currentNode.equals(_rootNode)) { // Create a category
         MutableTreeItem node = new MutableTreeItem("category", true);
         _treeModel.insertNodeInto(node, currentNode, currentNode.getChildCount());
         _jtree.expandPath(currentSelection);
         _jtree.setExpandsSelectedPaths(true);
         startEditingCategory(node);
      } else {							 // Create an item.
         MutableTreeItem categoryNode;
         if (currentNode.getAllowsChildren()) {
            categoryNode = currentNode;
         } else {
            categoryNode = (MutableTreeItem) currentNode.getParent();
         }
         TreeItem newItem = createItem((String) categoryNode.getUserObject());
         if (newItem != null) {
            MutableTreeItem node = new MutableTreeItem(newItem, false);
            _treeModel.insertNodeInto(node, categoryNode, categoryNode.getChildCount());
            _jtree.expandPath(currentSelection);
            getItemEditor(this, node, _treeModel).setVisible(true);
         }
      }
   }

   protected final void editCurrentItem() {
      TreePath currentSelection = _jtree.getSelectionPath();
      if (currentSelection != null) {
         MutableTreeItem selectedNode = (MutableTreeItem) currentSelection.getLastPathComponent();
         if (selectedNode.getLevel() < 1) {
         } else if (selectedNode.getUserObject() instanceof TreeItem) {
            getItemEditor(this, selectedNode, _treeModel).setVisible(true);
         } else {
            startEditingCategory(selectedNode);
         }
      }
   }

   protected abstract JDialog getItemEditor( GDialog dialog, MutableTreeItem currentNode, DefaultTreeModel treeModel );

   protected final void deleteSelectedItem() {
      TreePath currentSelection = _jtree.getSelectionPath();
      if (currentSelection != null) {
         MutableTreeItem selectedNode = (MutableTreeItem) currentSelection.getLastPathComponent();
         MutableTreeItem parentNode = (MutableTreeItem) selectedNode.getParent();
         if (!selectedNode.equals(_rootNode) && parentNode != null) {
            _treeModel.removeNodeFromParent(selectedNode);
         }
         selectNode();
      }
   }

   protected final void saveCurrentTree() {
      storeInPreferenceFile(_rootNode);
      setVisible(false);
      dispose();
   }

   /**
    * Stores the specified subtree in the current explorer tree and in the related file.
    *
    * @param root the root node of the subtree.
    * @return true if all preferences has been saved.
    */
   protected abstract boolean storeInPreferenceFile( MutableTreeItem root );

   protected final void cancelEdition() {
      setVisible(false);
      dispose();
   }
}
