/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.graph;

import java.awt.Component;
import java.awt.Container;
import java.awt.FocusTraversalPolicy;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

/**
 * Traversal policy of the components in a graph.
 */
public final class GraphFocusTraversalPolicy extends FocusTraversalPolicy {

   private final Comparator<? super Component> _comparator;
   private Container _cachedRoot;
   private List<Component> _cachedCycle;

   public GraphFocusTraversalPolicy() {
      // the order is based on coordinates.
      _comparator = new Comparator<Component>() {
         @Override
         public int compare( Component component1, Component component2 ) {
            if (component1.getBounds().x < component2.getBounds().x) {
               return -1;
            } else if (component1.getBounds().x > component2.getBounds().x) {
               return 1;
            } else if (component1.getBounds().y < component2.getBounds().y) {
               return -1;
            }
            return 1;
         }
      };
   }

   private void enumerateAndSortCycle( Container focusCycleRoot, List<Component> cycle ) {
      if (!focusCycleRoot.isShowing() || !focusCycleRoot.isVisible() || !focusCycleRoot.isDisplayable()) {
         return;
      }
      Component[] components = focusCycleRoot.getComponents();
      for (int i = 0; i < components.length; i++) {
         if (components[i].isFocusable()) {
            cycle.add(components[i]);
         }
      }
      // Sort the Components in the cycle
      Collections.sort(cycle, _comparator);
   }

   @Override
   public Component getComponentAfter( Container aContainer, Component aComponent ) {
      List<Component> cycle = new ArrayList<>();
      enumerateAndSortCycle(aContainer, cycle);

      try {
         int index = Collections.binarySearch(cycle, aComponent, _comparator);

         if (index < 0) {
            int newIndex = cycle.indexOf(aComponent);
            if (newIndex >= 0) {
               index = newIndex;
            } else {
               index = -index - 2;
            }
         }

         while (true) {
            index++;

            if (index >= cycle.size()) {
               if (aContainer.isFocusCycleRoot()) {
                  _cachedRoot = aContainer;
                  _cachedCycle = cycle;

                  Component retval = getFirstComponent(aContainer);

                  _cachedRoot = null;
                  _cachedCycle = null;
                  return retval;
               } else {
                  return null;
               }
            } else {
               return cycle.get(index);
            }
         }
      } catch (ClassCastException e) {
         return getFirstComponent(aContainer);
      }
   }

   @Override
   public Component getComponentBefore( Container aContainer, Component aComponent ) {
      List<Component> cycle = new ArrayList<>();
      enumerateAndSortCycle(aContainer, cycle);

      try {
         int index = Collections.binarySearch(cycle, aComponent, _comparator);

         if (index < 0) {
            index = -index - 2;
         } else {
            index--;
         }

         if (index < 0) {
            _cachedRoot = aContainer;
            _cachedCycle = cycle;

            Component retval = getLastComponent(aContainer);

            _cachedRoot = null;
            _cachedCycle = null;

            return retval;
         } else {
            return cycle.get(index);
         }
      } catch (ClassCastException e) {
         return getLastComponent(aContainer);
      }
   }

   @Override
   public Component getFirstComponent( Container aContainer ) {
      List<Component> cycle;

      if (aContainer.equals(_cachedRoot)) {
         cycle = _cachedCycle;
      } else {
         cycle = new ArrayList<>();
         enumerateAndSortCycle(aContainer, cycle);
      }

      int size = cycle.size();
      if (size == 0) {
         return null;
      }
      return cycle.get(0);
   }

   @Override
   public Component getLastComponent( Container aContainer ) {
      List<Component> cycle;

      if (aContainer.equals(_cachedRoot)) {
         cycle = _cachedCycle;
      } else {
         cycle = new ArrayList<>();
         enumerateAndSortCycle(aContainer, cycle);
      }

      int size = cycle.size();
      if (size == 0) {
         return null;
      }
      return cycle.get(cycle.size() - 1);
   }

   @Override
   public Component getDefaultComponent( Container aContainer ) {
      return getFirstComponent(aContainer);
   }
}
