/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.graph;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.tree.OperatorTreeItem;
import fr.greyc.pantheon.ariane.model.tree.SubroutineTreeItem;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.model.graph.GraphModel;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.Link;
import fr.greyc.pantheon.swingext.FadingEffects;
import fr.greyc.pantheon.swingext.MessageBox;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.TexturePaint;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.Autoscroll;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetAdapter;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.image.BufferedImage;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.io.File;
import java.io.IOException;

import javax.swing.JFileChooser;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;

/**
 * A <i>Graph</i> holds an executable program as a list of linked items. Items are included using Drag&Drop from the explorer, and linked from a Drag&Drop on their slots.
 * <p>
 * Copy/Paste uses XML version of each selected item.</p>
 */
public final class GraphView extends JPanel implements Printable, Autoscroll {

   private static final int AUTO_SCROLL_MARGIN = 50;
   private static final Color DEFAULT_BACKGROUND_COLOR = new Color(240, 240, 240);
   private static final Color DEFAULT_DATA_LINK_COLOR = new Color(102, 255, 102);
   private static final Color DEFAULT_VALUE_LINK_COLOR = new Color(255, 204, 0);
   private static final Color DEFAULT_MASK_LINK_COLOR = new Color(102, 102, 102);

   private static Color _backgroundColor = DEFAULT_BACKGROUND_COLOR;
   private static Color _imageLinkColor = DEFAULT_DATA_LINK_COLOR;
   private static Color _valueLinkColor = DEFAULT_VALUE_LINK_COLOR;
   private static Color _maskLinkColor = DEFAULT_MASK_LINK_COLOR;

   private static GridTexturePaint _gridTexturePaint = new GridTexturePaint();

   private final GraphModel _graphModel;
   private final Graph _graphPresenter;

   private Link _selectedLink;

   private boolean _draggingLink;

   private Point _movementOrigin;

   private RubberBandBox _rubberBandBox;

   private ElasticLink _ConnectionElastic;

   public GraphView( Graph graph, GraphModel graphModel ) {
      _graphPresenter = graph;
      _graphModel = graphModel;
      setLayout(new XYLayout());
      setBackground(_backgroundColor);
      setFocusCycleRoot(true);
      setFocusTraversalPolicyProvider(true);
      setFocusTraversalPolicy(new GraphFocusTraversalPolicy());
      setAutoscrolls(true); // allow dragging outside the window
      addListeners();
   }

   /////
   ///// A C C E S S O R S
   /////
   @Override
   public String getName() {
      if (_graphPresenter == null) {
         return null;
      } else {
         return _graphPresenter.getName();
      }
   }

   /////
   ///// M E N U
   /////
   private void displayPopupMenu( int x, int y ) {
      JPopupMenu popupMenu = new JPopupMenu();
      final boolean emptySelection = _graphPresenter.isSelectedItemsEmpty();

      JMenuItem item = new JMenuItem(ArianeSettings.getString("Graph.Menu.execute"));
      item.addActionListener(new ActionListener() {
         @Override
         public void actionPerformed( ActionEvent event ) {
            _graphPresenter.startExecution();
         }
      });
      item.setEnabled(!emptySelection);
      popupMenu.add(item);

      item = new JMenuItem(ArianeSettings.getString("Graph.Menu.reset"));
      item.addActionListener(new ActionListener() {
         @Override
         public void actionPerformed( ActionEvent event ) {
            _graphPresenter.resetAll();
         }
      });
      popupMenu.add(item);

      item = new JMenuItem(ArianeSettings.getString("Graph.Menu.copy"));
      item.addActionListener(new ActionListener() {
         @Override
         public void actionPerformed( ActionEvent event ) {
            Ariane.getPresenter().copy();
         }
      });
      item.setEnabled(!emptySelection);
      popupMenu.add(item);

      item = new JMenuItem(ArianeSettings.getString("Graph.Menu.cut"));
      item.addActionListener(new ActionListener() {
         @Override
         public void actionPerformed( ActionEvent event ) {
            Ariane.getPresenter().cut();
         }
      });
      item.setEnabled(!emptySelection);
      popupMenu.add(item);

      item = new JMenuItem(ArianeSettings.getString("Graph.Menu.paste"));
      item.addActionListener(new ActionListener() {
         @Override
         public void actionPerformed( ActionEvent event ) {
            Ariane.getPresenter().paste();
         }
      });
      item.setEnabled(Ariane.getPresenter().isPastable());
      popupMenu.add(item);

      item = new JMenuItem(ArianeSettings.getString("Graph.Menu.delete"));
      item.addActionListener(new ActionListener() {
         @Override
         public void actionPerformed( ActionEvent event ) {
            Ariane.getPresenter().deleteSelectedItems();
         }
      });
      item.setEnabled(!emptySelection);
      popupMenu.add(item);

      item = new JMenuItem(ArianeSettings.getString("Graph.Menu.close"));
      item.addActionListener(new ActionListener() {
         @Override
         public void actionPerformed( ActionEvent event ) {
            _graphPresenter.close();
         }
      });
      popupMenu.add(item);

      popupMenu.show(this, x, y);
   }

   public boolean isCloseConfirmed() {
      FadingEffects.fadeIn(Ariane.getFrame());
      String[] options = {ArianeSettings.getString("saveButtonText"),
         ArianeSettings.getString("nosaveButtonText"),
         ArianeSettings.getString("cancelButtonText")};
      int answer = MessageBox.showConfirmDialog(Ariane.getFrame(),
          ArianeSettings.getString("Graph.quit.message.part1") + _graphPresenter.getName() + ArianeSettings.getString("Graph.quit.message.part2"),
          ArianeSettings.getString("Graph.quit.title"),
          options, options[0]);
      FadingEffects.fadeOut(Ariane.getFrame());

      if (answer == 0) {
         _graphPresenter.save();
         return true;
      }
      return answer == 1;
   }

   /////
   ///// P R E F E R E N C E S
   /////
   public static Color getDefaultBackgroundColor() {
      return DEFAULT_BACKGROUND_COLOR;
   }

   public static Color getBackgroundColor() {
      return _backgroundColor;
   }

   public static void setBackgroundColor( Color color ) {
      _backgroundColor = color;
      _gridTexturePaint = new GridTexturePaint();
   }

   public static Color getDefaultImageLinkColor() {
      return DEFAULT_DATA_LINK_COLOR;
   }

   public static Color getImageLinkColor() {
      return _imageLinkColor;
   }

   public static void setImageLinkColor( Color color ) {
      _imageLinkColor = color;
   }

   public static Color getDefaultMaskLinkColor() {
      return DEFAULT_MASK_LINK_COLOR;
   }

   public static Color getMaskLinkColor() {
      return _maskLinkColor;
   }

   public static void setMaskLinkColor( Color color ) {
      _maskLinkColor = color;
   }

   public static Color getDefaultValueLinkColor() {
      return DEFAULT_VALUE_LINK_COLOR;
   }

   public static Color getValueLinkColor() {
      return _valueLinkColor;
   }

   public static void setValueLinkColor( Color color ) {
      _valueLinkColor = color;
   }

   /////
   ///// P A I N T I N G  /  P R I N T I N G
   /////
   /**
    * Prints the current graph in the given page.
    *
    * @param graphicContext the graphics context.
    * @param pageFormat the properties of the page to be printed.
    * @param pageIndex the index of the page to be printed.
    * @return NO_SUCH_PAGE in case of error, or PAGE_EXISTS.
    */
   @Override
   public int print( Graphics graphicContext, PageFormat pageFormat, int pageIndex ) {
      if (pageIndex > 0) {
         return NO_SUCH_PAGE;
      }
      Graphics2D g2d = (Graphics2D) graphicContext;

      double scaleW = pageFormat.getImageableWidth() / getWidth();
      double scaleH = (pageFormat.getImageableHeight() - 24) / getHeight(); // -24 for title.

      if (scaleH < scaleW) {
         scaleW = scaleH;
      }
      if (scaleW > 1.0) {
         scaleW = 1.0;
      }
      g2d.setPaint(Color.black);

      g2d.translate(pageFormat.getImageableX(), pageFormat.getImageableY() + 24);
      g2d.scale(scaleW, scaleW);
      printAll(g2d);
      return PAGE_EXISTS;
   }

   @Override
   public void printComponent( Graphics graphicContext ) {
      drawLinks(graphicContext);
   }

   @Override
   public void paint( Graphics gc ) {
      super.paint(gc);
      if (_rubberBandBox != null) {
         _rubberBandBox.paint(gc);
      }
      if (_ConnectionElastic != null) {
         _ConnectionElastic.paint(gc);
      }
   }

   @Override
   public void paintComponent( Graphics gc ) {
      super.paintComponent(gc);
      Graphics2D g2d = (Graphics2D) gc;
      setBackground(_backgroundColor);
      g2d.setPaint(_gridTexturePaint.getTexture());
      g2d.fillRect(0, 0, getWidth(), getHeight());
      drawLinks(gc);
   }

   public void repaintAll() {
      for (GraphItem item : _graphModel.getListOfGraphItems()) {
         item.getView().draw();
      }
   }

   private void drawLinks( Graphics gc ) {
      for (GraphItem item : _graphModel.getListOfGraphItems()) {
         // Input images.
         for (InputDataSlot slot : item.getInputDataSlots()) {
            slot.getView().drawLink(gc);
         }
         // Mask.
         if (item.getInputMaskSlot() != null) {
            (item.getInputMaskSlot()).getView().drawLink(gc);
         }
         // Parameters.
         for (InputParameterSlot slot : item.getInputParameterSlots()) {
            slot.getView().drawLink(gc);
         }
      }
   }

   public void resetCurrentSelection() {
      _selectedLink = null;
      _rubberBandBox = null;
      repaint();
   }

   /////
   ///// I T E M S   S E L E C T I O N
   /////
   private void addListeners() {
      addMouseListener(new MouseAdapter() {
         @Override
         public void mousePressed( MouseEvent event ) {
            requestFocusInWindow();
            onMousePressed(event);
         }

         @Override
         public void mouseReleased( MouseEvent event ) {
            onMouseReleased(event);
         }
      });
      addMouseMotionListener(new MouseMotionAdapter() {
         @Override
         public void mouseDragged( MouseEvent event ) {
            onMouseDragged(event);
         }
      });
      addKeyListener(new KeyAdapter() {
         @Override
         public void keyPressed( KeyEvent event ) {
            onKeyPressed(event);
         }
      });
      addFocusListener(new FocusAdapter() {
         @Override
         public void focusGained( FocusEvent event ) {
            _graphModel.synchronizeItemFilesWithSystemFiles();
         }
      });
      new DropTarget(this, new DropTargetAdapter() {
         /**
          * Called while a drag operation is ongoing, when the mouse pointer enters the operable part of the drop site for the DropTarget registered with this listener. Accepts the drop action for a move action.
          *
          * @param event the DropTargetDragEvent.
          */
         @Override
         public void dragEnter( DropTargetDragEvent event ) {
            event.acceptDrag(DnDConstants.ACTION_MOVE);
         }

         /**
          * The DropTarget intercepts drop() notifications before the registered DropTargetListener gets them. Decides if this drop is a valid action. Valid drop action can be:
          * <ul>
          * <li>Disconnection of two slots.
          * <li>Addition of a new item if the graph.
          * </ul>
          *
          * @param event the DropTargetDropEvent.
          */
         @Override
         public void drop( DropTargetDropEvent event ) {
            try {
               if (event.isDataFlavorSupported(SubroutineTreeItem.FLAVORS[0])) {
                  DataFlavor flavor = SubroutineTreeItem.FLAVORS[0];
                  addNewItem(event, flavor);
               } else if (event.isDataFlavorSupported(OperatorTreeItem.FLAVORS[0])) {
                  DataFlavor flavor = OperatorTreeItem.FLAVORS[0];
                  addNewItem(event, flavor);
               } else { // Unknown drop action.
                  event.rejectDrop();
                  event.dropComplete(false);
               }
            } catch (UnsupportedFlavorException | IOException e) {
               event.rejectDrop();
               event.dropComplete(false);
            }
         }

         private void addNewItem( DropTargetDropEvent event, DataFlavor flavor ) throws UnsupportedFlavorException, IOException {
            TreeItem incoming = (TreeItem) event.getTransferable().getTransferData(flavor);
            Point coordinates = new Point(event.getLocation().x - 15, event.getLocation().y - 15);
            _graphPresenter.addItemAtAbsolutePosition(incoming, coordinates);
            validate(); // force to display the new component.
            event.acceptDrop(DnDConstants.ACTION_MOVE);
            event.dropComplete(true);
         }
      });
   }

   /**
    * Manages the event "button pressed". If the button is:
    * <ul>
    * <li>the right button: displays the popup menu (UNIX).
    * <li>the left button: selects a link or a point in the link or initialize a new selection rectangle.
    * </ul>
    *
    * @param event the event properties.
    */
   private void onMousePressed( MouseEvent event ) {
      if (event.isPopupTrigger()) {	// Used by Unix/Linux: display popup menu.
         displayPopupMenu(event.getX(), event.getY());
      } else {
         _graphPresenter.unSelectAll();
         Ariane.getPresenter().displayPropertyPanel(_graphPresenter, null);
         Link link = _graphPresenter.detectLink(event.getX(), event.getY());
         if (link == null) {			// begin a selection rectangle.
            _rubberBandBox = new RubberBandBox(event.getX(), event.getY());
         } else {					// if a link is selected then reshape the link.
            _selectedLink = link;
            link.setSelected(true, event.getPoint());
            setCursor(link.getCursor());
            _draggingLink = true;
            _movementOrigin = event.getPoint();
         }
      }
   }

   private void onMouseDragged( MouseEvent event ) {
      if (_rubberBandBox != null) {
         _rubberBandBox.move(event.getPoint());
      }
      if (_draggingLink && _selectedLink != null) {
         _selectedLink.move(event.getPoint());
         repaint();
      }
   }

   /**
    * Manages the event "button released". If the button is :
    * <ul>
    * <li>right button: displays the popup menu (Windows)
    * <li>left button: updates items and points in the selection rectangle.
    * </ul>
    *
    * @param event the event properties.
    */
   private void onMouseReleased( MouseEvent event ) {
      if (event.isPopupTrigger()) { // Used by Windows.
         displayPopupMenu(event.getX(), event.getY());
      } else {
         if (_rubberBandBox != null) {
            for (GraphItem item : _graphModel.getListOfGraphItems()) {
               // add all items in the selection rectangle in the selection.
               if (_rubberBandBox.contains(new Rectangle(item.getView().getBounds()))) {
                  _graphPresenter.addInSelection(item, true);
               }
            }
            _rubberBandBox.erase();
         } else if (_draggingLink && _selectedLink != null) {
            _graphPresenter.moveLink(_selectedLink, _movementOrigin, event.getPoint());
            // restore the cursor.
            setCursor(null);
            _selectedLink.setSelected(false, null);
            _draggingLink = false;
         }
         resetCurrentSelection();
      }
   }

   public void onKeyPressed( KeyEvent event ) {
      switch (event.getKeyCode()) {
         case KeyEvent.VK_RIGHT:
         case KeyEvent.VK_KP_RIGHT:
            _graphPresenter.moveSelection(new Point(-1, 0));
            break;
         case KeyEvent.VK_LEFT:
         case KeyEvent.VK_KP_LEFT:
            _graphPresenter.moveSelection(new Point(1, 0));
            break;
         case KeyEvent.VK_UP:
         case KeyEvent.VK_KP_UP:
            // check if one of the selected items will disappear!!
            _graphPresenter.moveSelection(new Point(0, 1));
            break;
         case KeyEvent.VK_DOWN:
         case KeyEvent.VK_KP_DOWN:
            _graphPresenter.moveSelection(new Point(0, -1));
            break;
         default:
            break;
      }
   }

   /////
   ///// I N P U T   /   O U T P U T
   /////
   public File askForFilename() {
      String currentDir = UserPreferences.getCurrentDirectory();
      File currentFileName = new File(currentDir);

      JFileChooser fileChooser = new JFileChooser(currentFileName);
      fileChooser.setDialogType(JFileChooser.SAVE_DIALOG);
      fileChooser.setDialogTitle(ArianeSettings.getString("FileChooser.titleSaveGraph"));
      fileChooser.setFileFilter(new javax.swing.filechooser.FileFilter() {
         @Override
         public boolean accept( File file ) {
            String fname = file.getName();
            return fname.endsWith(ArianeSettings.DEFAULT_ARIANE_FILE_EXT) || file.isDirectory();
         }

         @Override
         public String getDescription() {
            return "Ariane (*" + ArianeSettings.DEFAULT_ARIANE_FILE_EXT + ")";
         }
      });

      FadingEffects.fadeIn(Ariane.getFrame());
      if (fileChooser.showSaveDialog(this.getParent()) == JFileChooser.APPROVE_OPTION) {
         File theFile = fileChooser.getSelectedFile();
         if (theFile != null) {
            String fullFileName = fileChooser.getSelectedFile().getPath();
            if (!fullFileName.endsWith(ArianeSettings.DEFAULT_ARIANE_FILE_EXT)) {
               fullFileName += ArianeSettings.DEFAULT_ARIANE_FILE_EXT;
            }
            theFile = new File(fullFileName);
            // Overwrite the chosen file?

            String[] options = {ArianeSettings.getString("overwriteButtonText"),
               ArianeSettings.getString("nooverwriteButtonText")};
            if (theFile.exists() && MessageBox.showConfirmDialog(Ariane.getFrame(),
                ArianeSettings.getString("Workspace.dialog.message.part1")
                + theFile.getName()
                + ArianeSettings.getString("Workspace.dialog.message.part2"),
                ArianeSettings.getString("Workspace.dialog.title"),
                options, options[0]) != 0) {
               FadingEffects.fadeOut(Ariane.getFrame());
               return null;
            }
            UserPreferences.setCurrentDirectory(theFile.getParent());
         }
         FadingEffects.fadeOut(Ariane.getFrame());
         return theFile;
      } else {
         FadingEffects.fadeOut(Ariane.getFrame());
         return null;
      }
   }

   /////
   ///// A U T O S C R O L L I N G   F O R   D n D
   /////
   @Override
   public Insets getAutoscrollInsets() {
      Rectangle visible = getVisibleRect();
      Dimension panelSize = getSize();
      Insets insets = new Insets(visible.y + AUTO_SCROLL_MARGIN, visible.x + AUTO_SCROLL_MARGIN,
          panelSize.height - visible.y - visible.height + AUTO_SCROLL_MARGIN,
          panelSize.width - visible.x - visible.width + AUTO_SCROLL_MARGIN);
      return insets;
   }

   @Override
   public void autoscroll( Point location ) {
      Rectangle rect = new Rectangle(
          location.x - AUTO_SCROLL_MARGIN,
          location.y - AUTO_SCROLL_MARGIN,
          2 * AUTO_SCROLL_MARGIN,
          2 * AUTO_SCROLL_MARGIN);
      scrollRectToVisible(rect);
   }

   ///////
   /////// S E L E C T I O N
   ///////
   private final class RubberBandBox {

      private final int _startX;
      private final int _startY;
      private final Rectangle _currentBox = new Rectangle();
      private final Color SHADOW_COLOR = new Color(0.4f, 0.4f, 0.4f, 0.3f);

      protected RubberBandBox( int x, int y ) {
         _startX = x;
         _startY = y;
      }

      private void erase() {
         repaint();
      }

      private void paint( Graphics graphicContext ) {
         Color savedColor = graphicContext.getColor();
         graphicContext.setColor(SHADOW_COLOR);
         graphicContext.fillRect(_currentBox.x, _currentBox.y, _currentBox.width, _currentBox.height);
         graphicContext.setColor(savedColor);
      }

      private boolean contains( Rectangle rectangle ) {
         return _currentBox.contains(rectangle);
      }

      private void move( Point xycoord ) {
         final int width = Math.abs(_startX - xycoord.x);
         final int height = Math.abs(_startY - xycoord.y);

         if (_currentBox.width != width || _currentBox.height != height) {
            _currentBox.x = Math.min(_startX, xycoord.x);
            _currentBox.y = Math.min(_startY, xycoord.y);
            _currentBox.width = width;
            _currentBox.height = height;
            Rectangle rectangle = new Rectangle(xycoord.x - AUTO_SCROLL_MARGIN, xycoord.y - AUTO_SCROLL_MARGIN, 1, 1);
            scrollRectToVisible(rectangle);
            repaint();
         }
      }
   }

   ///////
   /////// C O N N E C T I O N
   ///////
   public void startConnectionElastic( int x, int y ) {
      _ConnectionElastic = new ElasticLink(this, x, y);
      repaint();
   }

   public void stopConnectionElastic() {
      _ConnectionElastic = null;
      repaint();
   }

   public void drawConnectionElastic( Point coordinate ) {
      if (_ConnectionElastic != null) {
         _ConnectionElastic.moveElastic(this, coordinate);
         repaint();
      }
   }

   private static class GridTexturePaint {

      private final TexturePaint _gridTexture;
      private static final int OFFSET = -20;

      protected GridTexturePaint() {
         final int gridSize = 10;
         BufferedImage bi = new BufferedImage(gridSize, gridSize, BufferedImage.TYPE_INT_ARGB);
         Graphics2D big = bi.createGraphics();
         Color color = getBackgroundColor();
         Color ncolor = new Color(Math.abs((color.getRed() + OFFSET) % 255),
             Math.abs((color.getGreen() + OFFSET) % 255),
             Math.abs((color.getBlue() + OFFSET) % 255),
             color.getAlpha());
         big.setColor(ncolor);
         big.drawRect(0, 0, gridSize, gridSize);
         _gridTexture = new TexturePaint(bi, new Rectangle(0, 0, gridSize, gridSize));
      }

      private TexturePaint getTexture() {
         return _gridTexture;
      }
   }

   public void addItem( TreeItem treeItem, Point point ) {
      _graphPresenter.addItemAtAbsolutePosition(treeItem, point);
      validate(); // force to display the new component.

   }
}
