/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.graph;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.LayoutManager2;
import java.awt.Rectangle;
import java.util.HashMap;
import java.util.Map;

/**
 * Defines a class that knows how to lay out a graph.
 * A component is added with a location constraints.
 */
public final class XYLayout implements LayoutManager2 {

   private final Map<Component, XYConstraint> _componentConstraints = new HashMap<>();
   private static final int ICON_WIDTH = 12;
   private static final int ICON_HEIGHT = 12;
   private static final int OFFSET = 60;

   public enum Orientation {
      NORTHWEST,
      NORTHEAST,
      SOUTHWEST,
      SOUTHEAST,
      EASTSOUTH,
      WESTNORTH
   }

   @Override
   public void addLayoutComponent( String name, Component component ) {
      // Do nothing
   }

   @Override
   public void addLayoutComponent( Component component, Object constraints ) {
      Object innerConstraints;
      if (constraints == null) {
         innerConstraints = new XYConstraint();
      } else {
         innerConstraints = constraints;
      }
      synchronized (component.getTreeLock()) {
         if (innerConstraints instanceof XYConstraint) {
            XYConstraint constraint = (XYConstraint) innerConstraints;
            _componentConstraints.put(component, (XYConstraint) innerConstraints);
            if (constraint.getRank() < 0) {
               component.setBounds(constraint.getLocation().x, constraint.getLocation().y, 0, 0);
            }
         }
      }
   }

   @Override
   public void removeLayoutComponent( Component component ) {
      synchronized (component.getTreeLock()) {
         _componentConstraints.remove(component);
      }
   }

   @Override
   public Dimension preferredLayoutSize( Container container ) {
      return minimumLayoutSize(container);
   }

   @Override
   public Dimension minimumLayoutSize( Container container ) {
      synchronized (container.getTreeLock()) {
         Dimension dimension = new Dimension(0, 0);
         final Dimension containerDim = container.getSize();
         for (Component component : container.getComponents()) {
            if (component.isVisible() && _componentConstraints.get(component).getRank() < 0) {
               Rectangle rectangle = getComponentBounds(component, containerDim);
               dimension.width = Math.max(dimension.width, rectangle.x + rectangle.width + 50);
               dimension.height = Math.max(dimension.height, rectangle.y + rectangle.height + 50);
            }
         }
         return dimension;
      }
   }

   @Override
   public Dimension maximumLayoutSize( Container container ) {
      return new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE);
   }

   @Override
   public void layoutContainer( Container container ) {
      synchronized (container.getTreeLock()) {
         final Dimension containerDim = container.getSize();
         for (Component component : container.getComponents()) {
            if (component.isVisible()) {
               component.setBounds(getComponentBounds(component, containerDim));
            }
         }
      }
   }

   @Override
   public float getLayoutAlignmentX( Container container ) {
      return 0.0f; // Prefer centered alignment.
   }

   @Override
   public float getLayoutAlignmentY( Container container ) {
      return 0.0f; // Prefer centered alignment.
   }

   @Override
   public void invalidateLayout( Container container ) {
      // Do nothing
   }

   private Rectangle getComponentBounds( Component component, Dimension size ) {
      XYConstraint constraints = _componentConstraints.get(component);

      if (isConstrainedByOrientation(constraints)) {
         final int rank = constraints.getRank() + 1;
         switch (constraints.getOrientation()) {
            case WESTNORTH:
               return new Rectangle(0, OFFSET + rank * 20, ICON_WIDTH, ICON_HEIGHT);
            case NORTHWEST:
               return new Rectangle(OFFSET + rank * 20, 0, ICON_WIDTH, ICON_HEIGHT);
            case SOUTHEAST:
               return new Rectangle(200 + OFFSET + rank * 20, size.height - ICON_HEIGHT, ICON_WIDTH, ICON_HEIGHT);
            case SOUTHWEST:
               return new Rectangle(OFFSET + rank * 20, size.height - ICON_HEIGHT, ICON_WIDTH, ICON_HEIGHT);
            case NORTHEAST:
               return new Rectangle(size.width - ICON_WIDTH, OFFSET + rank * 20, ICON_WIDTH, ICON_HEIGHT);
            case EASTSOUTH:
               return new Rectangle(size.width - ICON_WIDTH, 200 + OFFSET + rank * 20, ICON_WIDTH, ICON_HEIGHT);
            default:
               return null;
         }
      } else { // Constraints by coordinates.
         Rectangle rectangle = component.getBounds();
         Dimension dimension = component.getPreferredSize();
         rectangle.width = dimension.width;
         rectangle.height = dimension.height;
         return rectangle;
      }
   }

   private boolean isConstrainedByOrientation( XYConstraint constraints ) {
      return constraints.getRank() != -1;
   }
}
