/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.graph.item;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.KeyStroke;
import javax.swing.border.Border;
import javax.swing.border.EtchedBorder;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.Slot;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.PropertyPanel;
import fr.greyc.pantheon.swingext.RoundedBorder;
import fr.greyc.pantheon.swingext.RoundedPanel;

/**
 * A <code>GraphItemView</code> defines the view of <code>GraphItem</code>.
 */
public abstract class GraphItemView extends RoundedPanel implements MouseListener, MouseMotionListener, FocusListener, KeyListener {

   /**
    * The color for selected shadow.
    */
   public static final int ITEM_HEADER_SIZE = 12;
   private static final Color SELECTED_HIGHLIGHT_COLOR = new Color(255, 0, 00);
   private static final Border UNSELECTED_BORDER = new RoundedBorder(EtchedBorder.RAISED);
   private static final Border SELECTED_BORDER = new RoundedBorder(EtchedBorder.LOWERED, SELECTED_HIGHLIGHT_COLOR);
   private static final ImageIcon UNEXECUTED_STATUS_ICON;
   private static final ImageIcon RUNNING_STATUS_ICON;
   private static final ImageIcon EXECUTED_STATUS_ICON;
   private static final ImageIcon ERROR_STATUS_ICON;

   private GraphItem _graphItem;

   private PropertyPanel _propertyPanel;
   private static JPopupMenu _popupMenu;

   private final JLabel _labelStatus;
   private final JTextField _textFieldName;
   private final JLabel _labelName;
   private JComponent _commentPanel;

   static {
      UNEXECUTED_STATUS_ICON = ArianeSettings.getIcon("StatusIcon0.png");
      RUNNING_STATUS_ICON = ArianeSettings.getIcon("StatusIcon1.png");
      EXECUTED_STATUS_ICON = ArianeSettings.getIcon("StatusIcon2.png");
      ERROR_STATUS_ICON = ArianeSettings.getIcon("StatusIcon3.png");
   }

   public GraphItemView() {
      _labelName = new JLabel();
      _labelName.setFont(UserPreferences.getDefaultFont());
      _textFieldName = new JTextField();
      _textFieldName.addFocusListener(new FocusListener() {
         @Override
         public final void focusLost( FocusEvent event ) {
            _graphItem.setNickName(((JTextField) (event.getSource())).getText(), true);
         }

         @Override
         public final void focusGained( FocusEvent event ) {
         }
      });
      _textFieldName.addActionListener((ActionEvent event) -> {
         _graphItem.setNickName(((JTextField) (event.getSource())).getText(), true);
      });
      _labelStatus = new JLabel();
      super.setLayout(new GridBagLayout());
      super.addMouseListener(this);
      super.addMouseMotionListener(this);
      super.addFocusListener(this);
      super.addKeyListener(this);
   }

   public final void dispose() {
      if (_propertyPanel != null) {
         _propertyPanel.setVisible(false);
      }
   }

   public final void setGraphItem( GraphItem graphItem ) {
      _graphItem = graphItem;
      _commentPanel = createCommentEditor();
   }

   protected final GraphItem getModel() {
      return _graphItem;
   }

   protected final Graph getGraph() {
      return _graphItem.getGraph();
   }

   public final JComponent getNameEditor() {
      return _textFieldName;
   }

   protected final String getFormula() {
      return _graphItem.getFormula();
   }

   protected final void setFormula( String formula ) {
      _graphItem.setFormula(formula);
   }

   @Override
   public final String toString() {
      return _graphItem.toString();
   }

   @Override
   public final String getName() {
      if (_graphItem == null) {
         return null;
      } else {
         return _graphItem.getName();
      }
   }

   protected JComponent createCommentEditor() {
      JLabel label = new JLabel();
      label.setMaximumSize(new Dimension(150, 10000));

      int size = UserPreferences.getDefaultFont().getSize();
      label.setFont(UserPreferences.getDefaultFont().deriveFont(size - 1.0f));
      label.setBorder(BorderFactory.createEmptyBorder(2, 2, 2, 2));
      return label;
   }

   public final void setSelected( boolean selected ) {
      setBorder(selected ? SELECTED_BORDER : UNSELECTED_BORDER);
   }

   public final void setComment( String comment ) {
      if (_commentPanel instanceof JLabel) {
         ((JLabel) _commentPanel).setText(comment);
      } else if (_commentPanel instanceof JTextArea) {
         ((JTextArea) _commentPanel).setText(comment);
      }
   }

   /////
   ///// E V E N T
   /////
   @Override
   public final void mouseClicked( MouseEvent event ) {
      if (event.getClickCount() > 1) {
         onMouseDoubleClicked(event);
      }
   }

   protected void onMouseDoubleClicked( MouseEvent event ) {
   }

   @Override
   public final void mousePressed( MouseEvent event ) {
      if (!event.isShiftDown() && !event.isControlDown()) {
         requestFocusInWindow();
      }
      getParent().setComponentZOrder(this, 0); // Put it in front
      if (event.isPopupTrigger()) { // Unix version -> open the popup menu
         getGraph().addInSelection(_graphItem, false);
         drawPopup(event);
      }
   }

   @Override
   public final void mouseReleased( MouseEvent event ) {
      if (event.isShiftDown() || event.isControlDown()) {
         hidePropertyPanel();
      } else {
         showPropertyPanel();
      }
      onMouseReleased(event);
   }

   protected final void onMouseReleased( MouseEvent event ) {
      if (event.isPopupTrigger()) { // Window version -> open the popup menu
         getGraph().addInSelection(_graphItem, false);
         drawPopup(event);
      } else if ((_popupMenu == null || !_popupMenu.isVisible())
          && !getGraph().dropSelection()
          && (event.isShiftDown() || event.isControlDown())) {
         getGraph().addInSelection(_graphItem, true);
      }
   }

   @Override
   public final void mouseDragged( MouseEvent event ) {
      if (_popupMenu == null || !_popupMenu.isVisible()) {
         if (!_graphItem.isSelected()) {
            getGraph().addInSelection(_graphItem, event.isShiftDown() || event.isControlDown());
         }
         event.translatePoint(getLocation().x, getLocation().y);
         getGraph().dragSelection(event);
      }
   }

   @Override
   public final void mouseMoved( MouseEvent event ) {
   }

   @Override
   public final void mouseEntered( MouseEvent event ) {
   }

   @Override
   public final void mouseExited( MouseEvent event ) {
   }

   @Override
   public void focusLost( FocusEvent event ) {
   }

   protected void focusGainedHook() {
   }

   @Override
   public final void focusGained( FocusEvent event ) {
      showPropertyPanel();
      if (_graphItem.isSelected()) {
         focusGainedHook();
      } else {
         getGraph().addInSelection(_graphItem, false);
      }
   }

   @Override
   public final void keyPressed( KeyEvent event ) {
      // redirect events to the related graph.
      getGraph().getView().onKeyPressed(event);
   }

   @Override
   public final void keyReleased( KeyEvent event ) {
   }

   @Override
   public final void keyTyped( KeyEvent event ) {
   }

   /////
   ///// D I S P L A Y I N G    &    P R I N T I N G
   /////
   public abstract Color getBackgroundColor();

   public void draw() {
      removeAll();
      setBackground(getBackgroundColor());
      setBorder(UNSELECTED_BORDER);

      _labelName.setText(_graphItem.getNickName());
      setToolTipText(toString());

      // layout
      GridBagConstraints constraints = new GridBagConstraints();
      constraints.gridheight = 1;

      // 1. parameters slots
      constraints.anchor = GridBagConstraints.NORTHWEST;
      constraints.gridy = 0;
      constraints.insets = new Insets(1, 5, 1, 0);

      constraints.gridx = 0;
      add(Box.createRigidArea(new Dimension(0, ITEM_HEADER_SIZE)), constraints);

      List<InputParameterSlot> inputParameterSlots = _graphItem.getInputParameterSlots();
      final Insets insets = new Insets(1, 0, 1, 0);
      for (int i = 0; i < inputParameterSlots.size(); i++) {
         Slot slot = inputParameterSlots.get(i);
         constraints.gridx = i;
         add(slot.getView(), constraints);
         constraints.insets = insets;
      }

      // 2. The mask slot
      int numberOfParamaters = inputParameterSlots.size();
      if (numberOfParamaters == 0) {
         numberOfParamaters = 1;
      }

      constraints.insets = new Insets(0, 0, 0, 5);
      constraints.anchor = GridBagConstraints.NORTHEAST;
      constraints.gridx = numberOfParamaters;
      InputMaskSlot inputMaskSlot = _graphItem.getInputMaskSlot();

      if (inputMaskSlot != null) {
         add(inputMaskSlot.getView(), constraints);
      }
      constraints.insets = new Insets(0, 0, 0, 0);

      // 3. The name
      constraints.gridwidth = numberOfParamaters + 1;
      constraints.gridy++;
      constraints.gridx = 0;
      constraints.anchor = GridBagConstraints.CENTER;
      add(_labelName, constraints);

      // 4. The lights ramp that shows the current status.
      constraints.gridwidth = GridBagConstraints.REMAINDER;
      constraints.gridy++;
      constraints.gridx = 0;
      displayStatus();
      add(_labelStatus, constraints);

      // 5. Optional Comments
      constraints.gridwidth = GridBagConstraints.REMAINDER;
      constraints.gridy++;
      constraints.gridx = 0;
      if (_commentPanel.getPreferredSize().width > _commentPanel.getMaximumSize().width) {
         Dimension dim = _commentPanel.getPreferredSize();
         dim.width = _commentPanel.getMaximumSize().width;
         _commentPanel.setPreferredSize(dim);
      }
      add(_commentPanel, constraints);

      // 6. The input + output slots
      constraints.insets = new Insets(0, 0, 1, 1);
      constraints.gridy++;
      constraints.gridx = 0;
      List<InputDataSlot> inputDataSlots = _graphItem.getInputDataSlots();
      List<OutputDataSlot> outputDataSlots = _graphItem.getOutputDataSlots();

      int maxSlot = inputDataSlots.size() > outputDataSlots.size() ? inputDataSlots.size() : outputDataSlots.size();
      for (int i = 0; i < maxSlot; i++) {
         if (i < inputDataSlots.size()) {
            constraints.gridx = 0;
            constraints.anchor = GridBagConstraints.WEST;
            Slot slot = inputDataSlots.get(i);
            //				slot.updateBackgroundColor();
            add(slot.getView(), constraints);
         }
         if (i < outputDataSlots.size()) {
            constraints.gridx = numberOfParamaters;
            constraints.anchor = GridBagConstraints.EAST;
            Slot slot = outputDataSlots.get(i);
            add(slot.getView(), constraints);
         }
         constraints.gridy++;
      }

      // 8. The result
      int max = inputDataSlots.size() > outputDataSlots.size() ? inputDataSlots.size() : outputDataSlots.size();
      constraints.gridy += max;
      constraints.insets = new Insets(0, 0, 2, 10);
      List<OutputResultSlot> outputResultSlots = _graphItem.getOutputResultSlots();

      if (!outputResultSlots.isEmpty()) {
         constraints.gridx = numberOfParamaters;
         constraints.gridwidth = 1;
         constraints.gridheight = 1;
         constraints.anchor = GridBagConstraints.SOUTHEAST;
         add(_graphItem.getOutputResultSlot().getView(), constraints);
      }
   }

   public final void displayStatus() {
      switch (_graphItem.getExecutionStatus()) {
         case UNEXECUTED:
            _labelStatus.setIcon(UNEXECUTED_STATUS_ICON);
            break;
         case RUNNING:
            _labelStatus.setIcon(RUNNING_STATUS_ICON);
            break;
         case EXECUTED:
            _labelStatus.setIcon(EXECUTED_STATUS_ICON);
            break;
         case ERROR:
            _labelStatus.setIcon(ERROR_STATUS_ICON);
            break;
         default:
            break;
      }
   }

   protected final void centerText( String text, Graphics graphics, int row, boolean multiline ) {
      if (text == null) {
         return;
      }
      int line = row;
      FontMetrics fontMetrics;
      fontMetrics = graphics.getFontMetrics();
      Dimension size = getSize();
      if (fontMetrics.stringWidth(text) > size.width) {
         if (multiline) {
            String subtext = text;
            // Use several lines and do not center text.
            String tmp;
            while (subtext.length() > 0) {
               int index = subtext.length();
               tmp = subtext;
               while (fontMetrics.stringWidth(tmp) > size.width - 4) {
                  tmp = subtext.substring(0, --index);
               }
               graphics.drawString(tmp, 2, line);
               line += fontMetrics.getHeight();
               subtext = subtext.substring(index);
            }
         }
         graphics.drawString(text, 2, line);
      } else { // Centers the text
         int column = (size.width - fontMetrics.stringWidth(text)) / 2;
         graphics.drawString(text, column, line);
      }
   }

   @Override
   public final void printChildren( Graphics gc ) {
      /*
       * empty
       */ }

   @Override
   public final void printBorder( Graphics graphics ) {
      Dimension size;
      size = new Dimension(getSize());
      Graphics2D g2d = (Graphics2D) graphics;
      g2d.setColor(Color.black);
      BasicStroke lineStroke = new BasicStroke(2.0F, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
      g2d.setStroke(lineStroke);
      g2d.drawRect(0, 0, size.width, size.height);
   }

   @Override
   public void printComponent( Graphics gc ) {
      Dimension size;
      int row;
      size = new Dimension(getSize());
      gc.setColor(Color.white);
      gc.fillRect(0, 0, size.width, size.height);

      gc.setColor(Color.black);
      gc.setFont(UserPreferences.getDefaultFont());
      FontMetrics fontMetrics;
      fontMetrics = gc.getFontMetrics();
      row = fontMetrics.getHeight();
      // For aestheticism purpose, split title and parameters.
      gc.drawLine(0, row + 4, size.width, row + 4);
   }

   public final void displayDocumentationManual() {
      if (getModel().getDocumentationManualURL() != null) {
         Ariane.getPresenter().displayDocumentationPage(ArianeSettings.getString("HTMLViewer.manual") + " " + getName(), getModel().getDocumentationManualURL(), null);
      }
   }

   /////
   ///// P R O P E R T Y    P A N E L
   /////
   /**
    * @return the list of specific properties (other than common properties).
    */
   public List<PropertyPanel.PropertyCouple> getSpecificProperties() {
      return null;
   }

   public JComponent getBinaryEditor() {
      String binary = getModel().getBinary();
      if (binary != null) {
         JTextField textField = new JTextField(binary);
         textField.setEnabled(false);
         textField.setEditable(false);
         return textField;
      }
      return null;
   }

   public JComponent getFormulaEditor() {
      JTextField formulaText = new JTextField(getModel().getFormulaText());
      formulaText.addFocusListener(new FocusListener() {
         @Override
         public final void focusLost( FocusEvent event ) {
            getModel().changeFormula((JTextField) event.getSource(), ((JTextField) event.getSource()).getText(), true);
         }

         @Override
         public final void focusGained( FocusEvent event ) {
            /*
             * empty
             */
         }
      });
      formulaText.addActionListener((ActionEvent event) -> {
         getModel().changeFormula((JTextField) event.getSource(), ((JTextField) event.getSource()).getText(), true);
      });
      return formulaText;
   }

   protected PropertyPanel createPropertyPanel() {
      return new PropertyPanel(ArianeSettings.getLanguage(), _graphItem);
   }

   public final void updatePropertyViewer() {
      if (_propertyPanel != null) {
         _propertyPanel.update(_graphItem);
      }
   }

   protected final void removePropertyViewer() {
      if (_propertyPanel != null) {
         _propertyPanel.setVisible(false);
      }
      _propertyPanel = null;
   }

   private void showPropertyPanel() {
      if (_propertyPanel == null || !_propertyPanel.isCorrectlyLocalized()) {
         Thread thread = new Thread() {
            @Override
            public void run() {
               _propertyPanel = createPropertyPanel();
               _propertyPanel.setVisible(true);
               Ariane.getPresenter().displayPropertyPanel(getGraph(), _propertyPanel);
            }
         };
         thread.start();
      } else {
         _propertyPanel.setVisible(true);
         Ariane.getPresenter().displayPropertyPanel(getGraph(), _propertyPanel);
      }
   }

   private void hidePropertyPanel() {
      Ariane.getPresenter().displayPropertyPanel(getGraph(), null);
   }

   public final void fireConnection() {
      if (_propertyPanel != null) {
         _propertyPanel.validate();
         _propertyPanel.repaint();
      }
   }

   public final void firePropertyName( String name ) {
      _labelName.setText(name);
   }

   public final void firePropertyTextName( String name ) {
      _textFieldName.setText(name);
   }

   /////
   ///// M E N U
   /////
   protected abstract JPopupMenu initPopup();

   protected final void drawPopup( MouseEvent event ) {
      _popupMenu = initPopup();
      _popupMenu.show(this, event.getX(), event.getY());
   }

   protected final JMenuItem createExecuteMenuItem() {
      JMenuItem item;

      item = new JMenuItem(ArianeSettings.getString("PopupMenu.execute"));
      item.addActionListener((ActionEvent event) -> {
         getGraph().startExecution();
      });

      item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
      return item;
   }

   protected final JMenuItem createDeleteMenuItem() {
      JMenuItem item;
      item = new JMenuItem(ArianeSettings.getString("PopupMenu.remove"));
      item.addActionListener((ActionEvent event) -> {
         getModel().delete();
      });

      item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
      return item;
   }

   protected final JMenuItem createCutMenuItem() {
      JMenuItem item;
      item = new JMenuItem(ArianeSettings.getString("Graph.Menu.cut"));
      item.addActionListener((ActionEvent event) -> {
         Ariane.getPresenter().cut();
      });

      item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, KeyEvent.CTRL_MASK));
      return item;
   }

   protected final JMenuItem createCopyMenuItem() {
      JMenuItem item;
      item = new JMenuItem(ArianeSettings.getString("Graph.Menu.copy"));
      item.addActionListener((ActionEvent event) -> {
         Ariane.getPresenter().copy();
      });

      item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, KeyEvent.CTRL_MASK));
      return item;
   }

   protected final JMenuItem createResetMenuItem() {
      JMenuItem item;
      item = new JMenuItem(ArianeSettings.getString("PopupMenu.reset"));
      item.addActionListener((ActionEvent event) -> {
         getModel().resetAll();
      });

      item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
      return item;
   }
}
