/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.graph.slot;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.border.LineBorder;

import fr.greyc.pantheon.ariane.model.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.Slot;

public abstract class SlotView extends JLabel implements MouseListener, MouseMotionListener {

   private static final int LINE_WIDTH = 1;
   private static final transient LineBorder ACTIVE_BORDER = (LineBorder) BorderFactory.createLineBorder(Color.RED, LINE_WIDTH);
   private static final transient LineBorder NORMAL_BORDER = (LineBorder) BorderFactory.createLineBorder(new Color(0, 0, 0, 0), LINE_WIDTH);

   private Slot _slot;

   private static Slot _targetSlot;
   private static Slot _sourceSlot;

   private boolean _selected;

   public SlotView() {
      setNormalBorder();
      setFocusable(false);
      addMouseListener(this);
      addMouseMotionListener(this);
   }

   public final void setModel( Slot slot ) {
      _slot = slot;
   }

   protected final Slot getModel() {
      return _slot;
   }

   public final boolean isSelected() {
      return _selected;
   }

   public final void setSelected( boolean isSelected ) {
      if (getModel().isConnected()) {
         highlightSlot(isSelected);
         for (Slot slot : getModel().getConnectedSlots()) {
            slot.getView().highlightSlot(isSelected);
         }
      }
   }

   public final void highlightSlot( boolean isSelected ) {
      _selected = isSelected;
      if (isSelected) {
         setActiveBorder();
      } else {
         setNormalBorder();
      }
   }

   public final void setActiveBorder() {
      setBorder(ACTIVE_BORDER);
   }

   public final void setNormalBorder() {
      setBorder(NORMAL_BORDER);
   }

   public abstract void setConnectedIcon();

   public abstract void setDisconnectedIcon();

   public abstract Point getAnchor();

   @Override
   public final void print( Graphics gc ) {
      gc.fillRect(0, 0, getWidth(), getHeight());
   }

   protected abstract Color getLinkColor();

   public final void drawLink( Graphics gc ) {
      if (getModel().isConnected()) {
         ((InputSlot) getModel()).getLink().draw(gc, getLinkColor(), isSelected());
      }
   }

   public abstract JComponent getValueEditor();

   /////////
   ///////// MOUSE
   /////////
   private void startConnection() {
      synchronized (this) {
         _targetSlot = null;
         _sourceSlot = getModel();
      }
   }

   private void dropIntoGraph() {
      if (_targetSlot == null) {
         getModel().getGraph().disconnectSlot(getModel());
      } else {
         setNormalBorder();
         getModel().connectWith(_targetSlot);
      }
   }

   @Override
   public final void mouseReleased( MouseEvent event ) {
      setCursor(null);
      getModel().getGraph().getView().stopConnectionElastic();
      setSelected(false);
      dropIntoGraph();
      getModel().getGraph().getView().requestFocusInWindow();
   }

   @Override
   public final void mousePressed( MouseEvent event ) {
      startConnection();
      setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
      getModel().getGraph().getView().startConnectionElastic(getLocationOnScreen().x + 4, getLocationOnScreen().y + 4);
   }

   @Override
   public final void mouseExited( MouseEvent event ) {
      if (_sourceSlot != null) {
         if (!isSelected()) {
            setNormalBorder();
         }
         _targetSlot = null;
      }
   }

   @Override
   public final void mouseEntered( MouseEvent event ) {
      if (_sourceSlot != null) {
         _targetSlot = getModel();
         // Accepts the drop action for a move action.
         if (_sourceSlot.isAcceptableDropSite(_targetSlot)) {
            setActiveBorder();
         }
      }
   }

   @Override
   public final void mouseDragged( MouseEvent event ) {
      getModel().getGraph().getView().drawConnectionElastic(event.getLocationOnScreen());
   }

   @Override
   public final void mouseClicked( MouseEvent event ) {
      /*
       * empty
       */
   }

   @Override
   public final void mouseMoved( MouseEvent event ) {
      /*
       * empty
       */
   }
}
