/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.swingext;

import java.awt.BorderLayout;
import java.awt.Cursor;
import java.awt.Dimension;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.RandomAccessFile;

import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextArea;
import javax.swing.border.BevelBorder;
import javax.swing.border.EmptyBorder;

import fr.greyc.pantheon.imageio.ImageDecoder;

/**
 * An implementation of image previewer that is used in FileChooser.
 * Allows the visualization in thumbnail image for supported image format,
 * such as GIF or JPEG or PANDORE.
 */
public final class ImageFilePreviewer extends JPanel implements PropertyChangeListener {

   private static final int MAXSIZE = 120;
   private static final int PREVIEW_LENGTH = 300;
   private ThumbnailReader _reader;

   public ImageFilePreviewer( JFileChooser fileChooser ) {
      super(new BorderLayout());
      super.setPreferredSize(new Dimension(200, 50));
      super.setBorder(new BevelBorder(BevelBorder.LOWERED));
      fileChooser.addPropertyChangeListener(this);
   }

   private void displayTextPreview( File selectedFile ) {
      add(loadText(selectedFile), BorderLayout.CENTER);
      revalidate();
      repaint();
   }

   private void displayImagePreview( File selectedFile ) {
      if (_reader != null) {
         _reader.kill();
      }
      _reader = new ThumbnailReader(this, selectedFile);
      _reader.start();
   }

   @Override
   public void propertyChange( PropertyChangeEvent event ) {
      if (event.getPropertyName().equals(JFileChooser.SELECTED_FILE_CHANGED_PROPERTY)) {
         File selectedFile = (File) event.getNewValue();
         removeAll();
         if (selectedFile != null) {
            if (isText(selectedFile, false)) {
               displayTextPreview(selectedFile);
            } else {
               displayImagePreview(selectedFile);
            }
         }
      }
   }

   private final class ThumbnailReader extends Thread {

      private final JComponent _component;
      private final File _selectedFile;
      private Boolean _isKilled = false;

      ThumbnailReader( JComponent component, File selectedFile ) {
         _component = component;
         _selectedFile = selectedFile;
      }

      public void kill() {
         _isKilled = true;
      }

      private boolean isKilled() {
         return _isKilled;
      }

      @Override
      public void run() {
         _component.getParent().getParent().setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
         ImageIcon imageIcon = loadImage(_selectedFile);
         if (isKilled()) {
            _component.getParent().getParent().setCursor(null);
            return;
         }
         JLabel label = new JLabel(imageIcon);
         if (isKilled()) {
            _component.getParent().getParent().setCursor(null);
            return;
         }
         _component.add(label, BorderLayout.CENTER);
         if (isKilled()) {
            _component.getParent().getParent().setCursor(null);
            return;
         }
         _component.getParent().getParent().setCursor(null);
         revalidate();
         repaint();
      }
   }

   private ImageIcon loadImage( File selectedFile ) {
      if (selectedFile != null) {
         return ImageDecoder.getThumbnail(selectedFile.getPath(), MAXSIZE);
      }
      return null;
   }

   private JTextArea loadText( File selectedFile ) {
      if (selectedFile != null) {
         try {
            String preview;
            try (FileInputStream fin = new FileInputStream(selectedFile)) {
               final byte[] data = new byte[PREVIEW_LENGTH];
               int bytesRead = 0;
               for (int i = 0; i < PREVIEW_LENGTH; i++) {
                  int character = fin.read();
                  if (character == -1) {
                     break;
                  }
                  bytesRead++;
                  data[i] = (byte) character;
               }
               preview = new String(data, 0, bytesRead);
            }
            JTextArea textArea = new JTextArea(preview);
            textArea.setEditable(false);
            textArea.setBorder(new EmptyBorder(5, 5, 5, 5));
            textArea.setFont(textArea.getFont().deriveFont(8.0f));
            return textArea;
         } catch (IOException ex) {
            return null;
         }
      }
      return null;
   }

   public static boolean isText( final File file, boolean acceptUTF8encoding ) {
      if (!file.exists()) {
         return false;
      }
      int bufferSize = 1024;
      boolean isText = true;
      byte[] buffer = new byte[bufferSize];
      try {
         try (RandomAccessFile fis = new RandomAccessFile(file, "r")) {
            fis.seek(0);
            final int read = fis.read(buffer);
            int lastByteTranslated = 0;
            for (int i = 0; i < read && isText; i++) {
               final byte character = buffer[i];
               int ub = character & 0xff;  // unsigned
               int utf8value = lastByteTranslated + ub;
               lastByteTranslated = ub << 8;

               if (!(ub == 0x09 // tab
                   || ub == 0x0A // line feed
                   || ub == 0x0C // form feed
                   || ub == 0x0D // carriage return
                   || ub >= 0x20 && ub <= 0x7E // Letters, Numbers and other "normal synbols"
                   || ub >= 0xA0 && ub <= 0xEE // Symbols of Latin-1
                   || acceptUTF8encoding && utf8value >= 0x2E2E && utf8value <= 0xC3BF)) {
                  isText = false;
                  break;
               }
            }
         }

         return isText;
      } catch (IOException e) {
         return false;
      }
   }
}
