/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.swingext.multivieweditor;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.Action;
import javax.swing.ActionMap;
import javax.swing.Icon;
import javax.swing.JComponent;
import javax.swing.JTabbedPane;
import javax.swing.SwingUtilities;
import javax.swing.TransferHandler;
import javax.swing.event.MouseInputListener;

/**
 * Customized TabbedPane with closable tabs.
 */
public final class DnDTabbedPane extends JTabbedPane {

   private static final int LINEWIDTH = 3;
   private static final int BUTTON_SIZE = 25;

   private int _dragTabIndex = -1;
   private transient DropLocation _dropLocation;

   protected DnDTabbedPane( TransferHandler handler ) {
      super.setTransferHandler(handler);
      super.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
      Controller listener = new Controller();
      super.addMouseListener(listener);
      super.addMouseMotionListener(listener);
      super.addPropertyChangeListener(listener);
   }

   protected boolean isRightTabbedPane() {
      return equals(((GMultiviewEditor) getParent()).getRightComponent());
   }

   protected void autoScrollTabList( Point pt ) {
      if (getTabCount() == 0) {
         return;
      }
      Rectangle rect = getTabAreaBounds();
      Rectangle leftSliderBound = new Rectangle();
      Rectangle rightSliderBound = new Rectangle();

      if (tabPlacement == TOP || tabPlacement == BOTTOM) {
         leftSliderBound.setBounds(rect.x, rect.y, BUTTON_SIZE, rect.height);
         rightSliderBound.setBounds(rect.x + rect.width - 2 * BUTTON_SIZE, rect.y, 2 * BUTTON_SIZE, rect.height);
      } else {
         leftSliderBound.setBounds(rect.x, rect.y, rect.width, BUTTON_SIZE);
         rightSliderBound.setBounds(rect.x, rect.y + rect.height - 2 * BUTTON_SIZE, rect.width, 2 * BUTTON_SIZE);
      }
      if (leftSliderBound.contains(pt)) {
         clickArrowButton("scrollTabsBackwardAction");
      } else if (rightSliderBound.contains(pt)) {
         clickArrowButton("scrollTabsForwardAction");
      }
   }

   private void clickArrowButton( String actionKey ) {
      ActionMap map = getActionMap();
      if (map != null) {
         Action action = map.get(actionKey);
         if (action != null && action.isEnabled()) {
            action.actionPerformed(new ActionEvent(this, ActionEvent.ACTION_PERFORMED, null, 0, 0));
         }
      }
   }

   protected DropLocation calculateDropLocationForPoint( Point pt ) {
      final boolean isVerticalPlacement = getTabPlacement() == JTabbedPane.TOP || getTabPlacement() == JTabbedPane.BOTTOM;
      Rectangle tabAreaBounds = getTabAreaBounds();
      Rectangle rect;
      for (int i = 0; i < getTabCount(); i++) {
         rect = getBoundsAt(i);
         if (isVerticalPlacement) {
            rect.x = rect.x - rect.width / 2;
            rect.y = tabAreaBounds.y;
            rect.height = tabAreaBounds.height;
         } else {
            rect.x = tabAreaBounds.x;
            rect.y = tabAreaBounds.y - rect.height / 2;
            rect.width = tabAreaBounds.width;
         }
         if (rect.contains(pt)) {
            return new DropLocation(pt, i);
         }
      }
      if (tabAreaBounds.contains(pt)) {
         return new DropLocation(pt, getTabCount());
      }
      return new DropLocation(pt, -1);
   }

   protected int getDragTabIndex() {
      return _dragTabIndex;
   }

   protected DropLocation getDropLocation() {
      return _dropLocation;
   }

   protected Object initDropLocation( TransferHandler.DropLocation location, Object state, boolean forDrop ) {
      if (location instanceof DropLocation) {
         DropLocation old = _dropLocation;
         _dropLocation = (DropLocation) location;
         firePropertyChange("dropLocation", old, _dropLocation);
      }
      return null;
   }

   protected void exportTab( int dragIndex, JTabbedPane target, int targetIndex ) {
      if (targetIndex < 0) {
         return;
      }
      Component cmp = getComponentAt(dragIndex);
      Container parent = target;
      while (parent != null) {
         if (cmp.equals(parent)) {
            return;
         }
         parent = parent.getParent();
      }

      Component tab = getTabComponentAt(dragIndex);
      String title = getTitleAt(dragIndex);
      Icon icon = getIconAt(dragIndex);
      String tip = getToolTipTextAt(dragIndex);
      boolean isEnabled = isEnabledAt(dragIndex);
      remove(dragIndex);
      target.insertTab(title, icon, cmp, tip, targetIndex);
      target.setEnabledAt(targetIndex, isEnabled);
      target.setTabComponentAt(targetIndex, tab);
      target.setSelectedIndex(targetIndex);
      if (tab instanceof JComponent) {
         ((JComponent) tab).scrollRectToVisible(tab.getBounds());
      }
   }

   protected void convertTab( int prev, int next ) {
      if (next < 0 || prev == next) {
         return;
      }
      Component cmp = getComponentAt(prev);
      Component tab = getTabComponentAt(prev);
      String str = getTitleAt(prev);
      Icon icon = getIconAt(prev);
      String tip = getToolTipTextAt(prev);
      boolean isEnabled = isEnabledAt(prev);
      int tgtindex = prev > next ? next : next - 1;
      remove(prev);
      insertTab(str, icon, cmp, tip, tgtindex);
      setEnabledAt(tgtindex, isEnabled);
      if (isEnabled) {
         setSelectedIndex(tgtindex);
      }
      setTabComponentAt(tgtindex, tab);
   }

   protected void paintDropLine( Graphics graphics ) {
      DropLocation dropLocation = getDropLocation();
      if (dropLocation == null || !dropLocation.isDropable()) {
         return;
      }
      Graphics2D g2 = (Graphics2D) graphics;
      g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.5f));
      g2.setColor(Color.RED);
      Rectangle dropIndicator = getGraphicalDropIndicatorBounds(dropLocation);
      Rectangle rect = SwingUtilities.convertRectangle(this, dropIndicator, getRootPane().getGlassPane());
      g2.fill(rect);
   }

   private Rectangle getGraphicalDropIndicatorBounds( DropLocation dropLocation ) {
      int idx = dropLocation.getTabIndex();
      if (idx < 0) {
         return new Rectangle(0, 0, 0, 0);
      } else if (getTabCount() == 0) {
         Rectangle rect = getTabAreaBounds();
         rect.setSize(LINEWIDTH, rect.height);
         return rect;
      } else {
         boolean isZero = idx == 0;
         Rectangle rect = getBoundsAt(isZero ? 0 : idx - 1);
         if (getTabPlacement() == JTabbedPane.TOP || getTabPlacement() == JTabbedPane.BOTTOM) {
            return new Rectangle(rect.x - LINEWIDTH / 2 + rect.width * (isZero ? 0 : 1), rect.y, LINEWIDTH, rect.height);
         } else {
            return new Rectangle(rect.x, rect.y - LINEWIDTH / 2 + rect.height * (isZero ? 0 : 1), rect.width, LINEWIDTH);
         }
      }
   }

   protected Rectangle getTabAreaBounds() {
      Rectangle tabbedRect = getBounds();
      Component component = getSelectedComponent();
      if (component == null) {
         tabbedRect.x = 0;
         tabbedRect.y = 0;
         return tabbedRect;
      }
      int xx = tabbedRect.x;
      int yy = tabbedRect.y;
      final Rectangle compRect = getSelectedComponent().getBounds();
      switch (tabPlacement) {
         case TOP:
            tabbedRect.height = tabbedRect.height - compRect.height;
            break;
         case BOTTOM:
            tabbedRect.y = tabbedRect.y + compRect.y + compRect.height;
            tabbedRect.height = tabbedRect.height - compRect.height;
            break;
         case LEFT:
            tabbedRect.width = tabbedRect.width - compRect.width;
            break;
         default:
            tabbedRect.x = tabbedRect.x + compRect.x + compRect.width;
            tabbedRect.width = tabbedRect.width - compRect.width;
            break;
      }
      tabbedRect.translate(-xx, -yy);
      return tabbedRect;
   }

   protected static final class DropLocation extends TransferHandler.DropLocation {

      private final int _index;
      private boolean _dropable = true;

      public DropLocation( Point point, int index ) {
         super(point);
         _index = index;
      }

      public int getTabIndex() {
         return _index;
      }

      public void setDropable( boolean flag ) {
         _dropable = flag;
      }

      public boolean isDropable() {
         return _dropable;
      }
   }

   private class Controller implements MouseInputListener, PropertyChangeListener {

      private Point _startPt;

      @Override
      public void propertyChange( PropertyChangeEvent event ) {
         String propertyName = event.getPropertyName();
         if ("dropLocation".equals(propertyName)) {
            repaintDropLocation();
         }
      }

      private void repaintDropLocation() {
         Component glassPane = getRootPane().getGlassPane();
         if (glassPane instanceof GhostGlassPane) {
            GhostGlassPane ghostGlassPane = (GhostGlassPane) glassPane;
            ghostGlassPane.setTargetTabbedPane(DnDTabbedPane.this);
            ghostGlassPane.repaint();
         }
      }

      @Override
      public void mousePressed( MouseEvent event ) {
         DnDTabbedPane src = (DnDTabbedPane) event.getSource();
         if (src.getTabCount() < 1) {
            _startPt = null;
         } else {
            Point mousePt = event.getPoint();
            int index = src.indexAtLocation(mousePt.x, mousePt.y);
            boolean regularTab = index >= 0 && src.isEnabledAt(index) && src.getComponentAt(index) != null;
            _startPt = regularTab ? mousePt : null;
            ((GMultiviewEditor) src.getParent()).setSelectedComponent(src.getSelectedComponent());
         }
      }

      @Override
      public void mouseDragged( MouseEvent event ) {
         if (_startPt != null) {
            Point mousePt = event.getPoint();
            DnDTabbedPane src = (DnDTabbedPane) event.getSource();
            TransferHandler th = src.getTransferHandler();
            _dragTabIndex = src.indexAtLocation(mousePt.x, mousePt.y);
            th.exportAsDrag(src, event, TransferHandler.MOVE);
            src.getRootPane().getGlassPane().setVisible(true);
            src.initDropLocation(new DropLocation(mousePt, -1), null, true);
            _startPt = null;
         }
      }

      @Override
      public void mouseClicked( MouseEvent event ) {
      }

      @Override
      public void mouseEntered( MouseEvent event ) {
      }

      @Override
      public void mouseExited( MouseEvent event ) {
      }

      @Override
      public void mouseMoved( MouseEvent event ) {
      }

      @Override
      public void mouseReleased( MouseEvent event ) {
      }
   }
}
